#
# file.tcl - file load and save
# 
# Copyright (C) 2005  Toni Ronkko
# 
# Peach Gardens is free software; you can redistribute it and/or modify 
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Peach Gardens is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranty of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with Peach Gardens; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# Aug 23, 2004  Toni Ronkko <toni at users.playstation2-linux.com>
# 
# $Id: file.tcl,v 1.3 2005/01/16 23:37:21 tronkko Exp $
#
# $Log: file.tcl,v $
# Revision 1.3  2005/01/16 23:37:21  tronkko
# supports pcx file format
#
# Revision 1.2  2005/01/09 20:53:03  tronkko
# file format 2 added accessibility, depth coordinate and background chip
#
# Revision 1.1  2004/12/09 22:16:57  tronkko
# Initial revision
#
#


#
# Load chipset, palettes and maps from a file.
#
proc load_data { fn } { 
  global edit_status
  global chipset_present

  # cannot undo after load
  reset_undo

  # open file for reading
  set ok 1  
  set fp [open $fn "r"]
  if { $fp != "" } {

    # clear all data
    destroy_chipset
    destroy_workspaces

    # read file into memory
    set ok [internal_read $fp $fn]
    if { $ok } {
      # no unsaved data in memory
      set edit_status 0
      set chipset_present 1
    } else {
      # erase all workspaces
      set edit_status 0
      destroy_chipset
      destroy_workspaces
    }
    close $fp
  } else {
    # cannot open file
    modal_dialog "Error" "Cannot open file $fn" error
    set ok 0
  }

  return $ok
}


#
# Read data from file fp.  All existing data must have been destroyed
# before using the function.
#
# File format 1 was used prior to Jan 7, 2005.
#
# File format 2 used since Jan 7, 2005 added support for chip depth,
# background chip and accessibility.
#
proc internal_read { fp fn } {
  global min_file_version
  global max_file_version
  global chipset_fn
  global chip_x1
  global chip_y1
  global workspace_chips
  global chip_bg
  global chip_depth
  global chip_terrain
  global cur_file_version
  
  # read version number
  if { [gets $fp line] >= 0 } {
    set col [split $line " "]
    if { [lindex $col 0] == "!PGardens" } {

      # check version
      set ver [lindex $col 1]
      if { $ver < $min_file_version  ||  $ver > $max_file_version } {
        # unsupported version
        modal_dialog "Error" "Unsupported file format" error
        return 0
      }

      # tell user about old version
      if { $cur_file_version != $ver } {
        modal_dialog "Warning" "$fn has been created using\
older version.  The file will be still imported, but all saves\
will use a later file format." warning
      }
      
    } else {
      # not a peach gardens map file
      modal_dialog "Error" "Not a Peach Gardens map file" error
      return 0
    }
      
  } else {
    # unexpected eof
    modal_dialog "Error" "Unexpected end of file" error
    return 0
  }

  # read file one line at the time
  while { [gets $fp line] >= 0 } {
    set col [split $line " "]
    set tag [lindex $col 0]
    switch $tag {
      "!chipset" {
        # get chipset file and chip dimensions
        set n [expr int([lindex $col 1])]
        set w [expr int([lindex $col 2])]
        set h [expr int([lindex $col 3])]
        set chipset_fn [file join [file dirname $fn] [lindex $col 4]]

        # Load chipset image.  Be ware of init_chipset, which sets global
        # nchips variable to zero.
        set img [read_image $chipset_fn]
        if { $img != "" } {
          init_chipset $img $w $h
          image delete $img
        } else {
          modal_dialog "Error" "Cannot find image: $chipset_fn" error
          return 0
        }
        
        # read chips
        for {set i 0} {$i < $n} {incr i} {
          if { [gets $fp line] >= 0 } {
            set col [split $line " "]
            set i [expr int([lindex $col 0])]
            set chip_x1($i) [expr int([lindex $col 1])]
            set chip_y1($i) [expr int([lindex $col 2])]

            # additions of file format 2
            if { [llength $col] > 3 } {
              set chip_bg($i) [expr int([lindex $col 3])] 
              set chip_depth($i) [expr int([lindex $col 4])]

              for {set j 1} {$j < 10} {incr j} {
                set t [lindex $col [expr 4+$j]]
                set chip_terrain($i,$j) $t
              }
            }
            
          } else {
            modal_dialog "Error" "Unexpected end of file" error
            return 0
          }
        }

        # Now that background chips have been read, re-create images for
        # display.  Doing this makes ugly violet regions go away.
        refresh_chipset
      }
      
      "!work" {
        # create new workspace
        set w    [expr int([lindex $col 1])]
        set h    [expr int([lindex $col 2])]
        set type [lindex $col 3]
        set name [lindex $col 4]
        set i [new_workspace $type $name $w $h]

        # read chips to workspace
        set x 0
        set y 0
        set j 0
        set n [expr $w*$h]
        while {$j < $n} {
          if { [gets $fp line] >= 0 } {
            # got a line
            set col [split $line " "]
            set ncols [llength $col]
            incr j $ncols

            # save chips in a line
            for {set k 0} {$k < $ncols} {incr k} {
              # assign chip to workspace array
              set workspace_chips($i,$x,$y) [lindex $col $k]

              # update coordinate for next chip
              incr x
              if { $x >= $w } {
                incr y
                set x 0
              }
            }
          } else {
            modal_dialog "Error" "Unexpected end of file" error
            return 0
          }
        }
      }
      
      default {
        # unknown tag
        modal_dialog "Error" "Unrecognized tag: $tag" error
        return 0
      }
    }
  }
  
  return 1
}


#
# Save chipset, palettes and maps to a file.
#
proc save_data { fn } {
  global cur_file_version
  global chipset_fn
  global nchips
  global chipw
  global chiph
  global chip_x1
  global chip_y1
  global chip_bg
  global chip_depth
  global chip_terrain
  global nworkspaces
  global workspace_used
  global workspace_width
  global workspace_height
  global workspace_chips
  global workspace_type
  global workspace_name
  
  # open file for writing
  set ok 1  
  set fp [open $fn "w"]
  if { $fp != "" } {

    # header and version info
    puts $fp "!PGardens $cur_file_version"

    # output chipset_fn that is relative to fn
    set relpath [make_relative [file dirname $fn] $chipset_fn]
    puts $fp "!chipset $nchips $chipw $chiph $relpath"
    
    # chips
    for {set i 0} {$i < $nchips} {incr i} {
      puts -nonewline $fp "$i $chip_x1($i) $chip_y1($i)"
      puts -nonewline $fp " $chip_bg($i) $chip_depth($i)"
      for {set j 1} {$j < 10} {incr j} {
        puts -nonewline $fp " $chip_terrain($i,$j)"
      }
      puts $fp ""
    }

    # workspaces
    for {set i 0} {$i < $nworkspaces} {incr i} {
      if { $workspace_used($i) } {
        # the workspace is used
        set w $workspace_width($i)
        set h $workspace_height($i)
        set t $workspace_type($i)

        # header
        puts $fp "!work $w $h $t $workspace_name($i)"

        # chips
        for {set y 0} {$y < $h} {incr y} {
          for {set x 0} {$x < $w} {incr x} {
            # get chip id
            set chip $workspace_chips($i,$x,$y)

            # separator
            if { $x != 0 } {
              puts -nonewline $fp " "
            }

            # print chip id
            puts -nonewline $fp "$chip"
          }
          puts $fp ""        
        }
      }
    }
    
    close $fp
  } else {
    # cannot open file
    modal_dialog "Error" "Cannot write file $fn" error
    set ok 0
  }
  
  return $ok
}


#
# Make file name fn relative to directory dir.  The function does not
# probably work on Microsoft systems if the chipset image is on a
# different drive than the map file.  However, I do not care to
# fix the problem, as the file would be broken anyway.  There is no
# way a file using absolute paths could work under unix.  Hence,
# it is better to ignore the bug, and assume that the chipset image
# is always stored next to the map file, in which case the problem
# does not exits.
#
proc make_relative { dir fn } {
  set result ""

  # make arguments absolute
  set dir [file nativename $dir]
  if { [file pathtype $dir] == "relative" } {
    set dir [file join [pwd] $dir]
  }
  set fn [file nativename $fn]
  if { [file pathtype $fn] == "relative" } {
    set fn [file join [pwd] $fn]
  }
  
  # make both arguments absolute, and split the components
  set dcomp [file split $dir]
  set fcomp [file split $fn ]

  # erase similar components at the start
  set i 0
  set j 0
  set imax [llength $dcomp]
  set jmax [llength $fcomp]
  while { $i < $imax  &&  $j < $jmax } {
    if { [lindex $dcomp $i] == [lindex $fcomp $j] } {
      # components are the same
      incr i
      incr j
    } else {
      # found the very first differing component
      break;
    }
  }

  # ascend from the initial directory
  while { $i < $imax } {
    set result [file join $result ".."]
    incr i
  }

  # descend to file
  while { $j < $jmax } {
    set result [file join $result [lindex $fcomp $j]]
    incr j
  }
  return $result
}
