#
# import.tcl - import chipset
# 
# Copyright (C) 2005  Toni Ronkko
# 
# Peach Gardens is free software; you can redistribute it and/or modify 
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Peach Gardens is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with Peach Gardens; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# Aug 23, 2004  Toni Ronkko <toni at users.playstation2-linux.com>
# 
# $Id: import.tcl,v 1.5 2005/01/16 23:38:53 tronkko Exp $
#
# $Log: import.tcl,v $
# Revision 1.5  2005/01/16 23:38:53  tronkko
# support for pcx file format, returns image as is so that caller does
# not need to re-read image file
#
# Revision 1.4  2005/01/09 20:58:46  tronkko
# set zoom to 2 to speed up import
#
# Revision 1.3  2004/12/09 22:12:06  tronkko
# do not confirm overwrite of existing chipset as features new and
# import can handle that more intelligently, deal with errors that have
# occured while reading image file
#
# Revision 1.2  2004/11/30 20:30:30  tronkko
# drag canvas with button 3 as in palette.tcl
#
# Revision 1.1  2004/11/28 11:08:38  tronkko
# Initial revision
#
#
#

# set of file name patterns for image formats
set import_types {
  {{Image Files} {.gif .pcx}}
  {{GIF Image}   {.gif}}
  {{PCX Image}   {.pcx}}
  {{All Files}   *}
}

# current chip size (in import dialog)
set import_xsize ""
set import_ysize ""

# return value of chip size dialog
set import_dialog_result ""

# width and height of enlarged chipset image (in import dialog)
set import_width  0
set import_height 0

# name of canvas object holding enlarge chipset image (for import)
set import_canvas 0

# zoom factor in import chipset dialog
set import_zoom 2


#
# Ask for image file name, and show import dialog asking for input options.
# Returns list of filename, chip_width, chip_height and image unless user
# aborts the operation in which case an empty list is returned.
#
proc import_dialog {} {
  global import_types
  global import_dialog_result
  global import_xsize
  global import_ysize
  global import_width
  global import_height
  global import_canvas
  global import_zoom
  global have_spinbox
  global default_dir
  
  # ask the user to name an image file to load
  set result [list]
  set fn [tk_getOpenFile -parent . -filetypes $import_types \
      -title "Import chipset..." -initialdir $default_dir]
  if { $fn != "" } {
    # user gave us a file name
    status "Loading file, please wait..."
    update idletasks
    
    # import chipset from file
    set img [read_image $fn]
    if { $img == "" } {
      status "Image not loaded"
      return [list]
    }
    set zoomed [image create photo]
    $zoomed copy $img -zoom $import_zoom $import_zoom

    # get width and height of the enlarged chipset image
    set import_width  [image width $zoomed]
    set import_height [image height $zoomed]

    # create dialog window to define the chip size
    set win [dialog_create {Import Chipset...}]
    set info     [dialog_info $win]
    set controls [dialog_controls $win]

    # default chip size
    set import_xsize 16
    set import_ysize 16

    # width and height buttons
    frame $info.row
    label $info.row.title -text "Chip Size:"
    if { $have_spinbox } {
      # tk 8.4
      spinbox $info.row.xspin -command "refresh_import" \
          -from 8 -to 64 -increment 1 -width 3 \
          -textvariable import_xsize
      label $info.row.x -text "x"
      spinbox $info.row.yspin -command "refresh_import" \
          -from 8 -to 64 -increment 1 -width 3 \
          -textvariable import_ysize
    } else {
      # tk 8.0
      entry $info.row.xspin -width 3 -textvariable import_xsize
      label $info.row.x -text "x"
      entry $info.row.yspin -width 3 -textvariable import_ysize
      button $info.row.update -text "Update" -command "refresh_import"
    }

    # canvas with scrollbars
    set import_canvas $info.scroll.canvas
    set c $import_canvas
    frame $info.scroll
    canvas $c -width 320 -height 256 \
        -xscrollcommand "$info.scroll.xbar set" \
        -yscrollcommand "$info.scroll.ybar set" \
         -takefocus 1 -cursor left_ptr
    $c create image 1 1 -anchor nw -image $zoomed
    scrollbar $info.scroll.xbar -orient horizontal \
        -command "$c xview" -takefocus 0
    scrollbar $info.scroll.ybar -orient vertical \
        -command "$c yview" -takefocus 0
    $c configure -scrollregion [$c bbox all]

    # title
    pack $info.row.title -side left

    # width and height controls
    pack $info.row.xspin -side left
    pack $info.row.x -side left -padx 4
    pack $info.row.yspin -side left
    if { ! $have_spinbox } {
      pack $info.row.update -side left -padx 4
    }
    pack $info.row -fill x -anchor nw

    # canvas and scrollbars
    grid $c -column 0 -row 0 -sticky nsew
    grid $info.scroll.ybar -row 0 -column 1 -sticky ns
    grid $info.scroll.xbar -row 1 -column 0 -sticky ew
    grid rowconfigure $info.scroll 0 -weight 1
    grid columnconfigure $info.scroll 0 -weight 1    
    pack $info.scroll -fill both -anchor nw -expand yes -padx 4 -pady 4

    # button row
    frame $controls.row
    pack $controls.row -expand yes
    
    button $controls.row.ok -text "Import" -default active \
        -command "set import_dialog_result ok"
    pack $controls.row.ok -side left -padx 4 
    bind $controls.row.ok <Return> "%W invoke"
    
    button $controls.row.cancel -text "Cancel" \
        -command "set import_dialog_result cancel"
    pack $controls.row.cancel -side left -padx 4 
    bind $controls.row.cancel <Return> "%W invoke"

    # what to do when the user closes the dialog through window manager
    wm protocol $win WM_DELETE_WINDOW "set import_dialog_result cancel"

    # abort by esc
    bind $win <Escape> "set import_dialog_result cancel"

    # drag canvas by button 3
    bind $c <3> "$c scan mark %x %y; $c configure -cursor fleur"
    bind $c <B3-Motion> "$c scan dragto %x %y"
    bind $c <ButtonRelease-3> "$c configure -cursor left_ptr"

    # scroll canvas by cursor keys
    bind $c <Left>   "$c xview scroll -1 units"
    bind $c <Right>  "$c xview scroll  1 units"
    bind $c <Up>     "$c yview scroll -1 units"
    bind $c <Down>   "$c yview scroll  1 units"

    # update grid size by enter
    bind $info.row.xspin <Return> "refresh_import"
    bind $info.row.yspin <Return> "refresh_import"
    bind $info.row.xspin <FocusOut> "refresh_import"
    bind $info.row.yspin <FocusOut> "refresh_import"

    # hints
    bind $controls.row.ok <Enter> \
        "status {Import chipset into map editor}"
    bind $controls.row.ok <Leave> \
        "status {}"

    bind $info.row.xspin <Enter> \
        "status {Width of a tile}"
    bind $info.row.xspin <Leave> \
        "status {}"

    bind $info.row.yspin <Enter> \
        "status {Height of a tile}"
    bind $info.row.yspin <Leave> \
        "status {}"

    bind $c <Enter> \
        "status {Preview of the generated chipset}"
    bind $c <Leave> \
        "status {}"

    # move cursor to spinbox
    focus $info.row.xspin

    # draw default grid
    refresh_import
    
    # show dialog window and wait for user's response
    set import_dialog_result ""
    status "Please define chip size"
    dialog_wait $win import_dialog_result

    # close the dialog
    destroy $win

    # import file if conformed
    if { $import_dialog_result == "ok" } {
      # set default directory (for loading map)
      set default_dir [file dirname $fn]

      # for returning import parameters
      set result [list $fn $import_xsize $import_ysize $img]
    }

    # delete image resources
    image delete $zoomed
  }
  return $result;
}


#
# Draw grid over the chipset in the chipset import window.  Also check the
# validity of grid size parameters.  This is needed if user types directly
# to the grid size boxes.
#
proc refresh_import {} {
  global import_xsize
  global import_ysize
  global import_zoom
  global import_width
  global import_height
  global import_canvas

  # confine width and height to acceptable limits.
  if { $import_xsize < 8 } {
    set import_xsize 8
  }
  if { $import_xsize > 64 } {
    set import_xsize 64
  }
  if { $import_ysize < 8 } {
    set import_ysize 8
  }
  if { $import_ysize > 64 } {
    set import_ysize 64
  }

  # fetch temporary variables
  set xi [expr $import_xsize * $import_zoom]
  set yi [expr $import_ysize * $import_zoom]
  set w $import_width
  set h $import_height
  set c $import_canvas

  # delete existing lines
  $c delete "line"

  # insert new grid
  if { $xi > 0  &&  $yi > 0 } {
    for {set x $xi} {$x < $w} {incr x $xi} {
      $c create line $x 0 $x $h -fill black -width 1 -tag "line"
    }
    for {set y $yi} {$y < $h} {incr y $yi} {
      $c create line 0 $y $w $y -fill black -width 1 -tag "line"
    }
  }
}
