#
# palette.tcl - workspaces of selector widget
# 
# Copyright (C) 2005  Toni Ronkko
# 
# Peach Gardens is free software; you can redistribute it and/or modify 
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Peach Gardens is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with Peach Gardens; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# Aug 23, 2004  Toni Ronkko <toni at users.playstation2-linux.com>
# 
# $Id: palette.tcl,v 1.3 2005/04/03 16:41:27 tronkko Exp $
#
# $Log: palette.tcl,v $
# Revision 1.3  2005/04/03 16:41:27  tronkko
# bug fix: include chips at right and bottom edges on automatically
# generated chipsets (such as one generated by reset)
#
# Revision 1.2  2005/01/09 20:54:42  tronkko
# use select_palette instead of refresh_view to ensure that all chips
# are re-drawn
#
# Revision 1.1  2004/12/09 22:12:32  tronkko
# Initial revision
#
#

# current workspace of selector (for highlighting menu entries)
set cur_palette 0


#
# Show palette creation dialog, and create new palette.
#
proc create_palette_dialog {} {
  global selector_widget
  global write_protect  
  global chipset_present
  global workspace_dialog_size
  
  # Refuse to create palette unless there is a chipset.  This is needed as
  # disabling menus does not disable keyboard shortcuts that we have
  # added.
  if { ! $chipset_present } {
    return
  }

  # offer to create small workspace by default
  set workspace_dialog_size 0

  # create workspace
  set i [create_workspace_dialog "Create Palette..." palette]
  if { $i >= 0 } {
    # file has now been edited
    set_edited 1
    
    # switch to the newly created palette
    select_palette $i
    refresh_menu

    # warn user if the cannot possibly edit the workspace
    if { $write_protect($selector_widget) } {
      modal_dialog "Info" "Palettes are currently protected against accidental editing. Please switch off the protection to edit the palette." info
    }
  }
}


#
# Edit name or resize palette.
# 
proc edit_palette_dialog {} {
  global cur_workspace
  global selector_widget

  # find out current palette
  set cur $cur_workspace($selector_widget)

  # edit current workspace
  if { $cur >= 0 } {
    if { [edit_workspace_dialog "Palette Properties..." $cur] } {
      # file has now been edited
      set_edited 1
      
      # update screen
      select_palette $cur
      refresh_menu
      
      # erase undo buffer as the undo events may refer to non-existent area
      reset_undo
    }
  } else {
    modal_dialog "Error" "No palettes present" error
  }
}


#
# Refresh list of workspaces in main menu.  The function must be called
# after inserting or deleting workspaces.
#
proc refresh_palette_menu {} {
  global palette_menu
  global nworkspaces
  global workspace_name
  global selector_widget
  global workspace_used
  global workspace_type
  global chipset_present
  
  # empty the menu completely
  $palette_menu delete 0 last

  # re-insert workspaces
  set count 0
  for {set i 0} {$i < $nworkspaces} {incr i} {
    if { $workspace_used($i)  &&  $workspace_type($i) == "palette" } {
      $palette_menu add radiobutton -underline 0 \
          -label $workspace_name($i) -variable cur_palette \
          -value $i -command "select_palette $i"
      incr count
    }
  }

  # allow editing?
  if { $chipset_present } {
    set new_state normal
    if { $count != 0 } {
      set edit_state normal
    } else {
      set edit_state disabled
    }
  } else {
    # no chipset open
    set new_state disabled
    set edit_state disabled
  }

  # re-insert actions
  $palette_menu add separator
  $palette_menu add command -label "New..." -state $new_state \
      -command create_palette_dialog
  $palette_menu add command -label "Reset..." -state $edit_state \
      -command reset_palette  
  $palette_menu add command -label "Properties..." -state $edit_state \
      -command edit_palette_dialog
  $palette_menu add command -label "Delete..." -state $edit_state \
      -command delete_palette  
}


#
# Select ith palette to selector widget on the left.  The function is
# typically invoked when the user selectes a workspace from the
# workspace menu.
#
proc select_palette { i } {
  global selector_widget
  global cur_palette

  # highlight palette in menu
  set cur_palette $i

  # show palette in widget
  select_workspace $selector_widget $i
}


#
# Delete current palette.
#
proc delete_palette {} {
  global workspace_name
  global cur_workspace
  global selector_widget

  # get index of current palette
  set cur $cur_workspace($selector_widget)
  if { $cur >= 0 } {
    # confirm delete
    set name $workspace_name($cur)
    set ok [modal_dialog \
        "Delete Palette?" \
        "The operation cannot be undone.  Really delete $name?" \
        question {Yes No} 0 ]
    if { $ok == 0 } {
      # erase workspace
      set pal [delete_workspace $cur]

      # file has now been edited
      set_edited 1

      # reset undo buffers so that deleted buffer cannot be edited
      reset_undo

      # select some other palette
      select_palette $pal
      refresh_menu
    } else {
      status "Not deleted"
    }
  } else {
    modal_dialog "Error" "No palettes present" error
  }
}


#
# Create initial palette of w chips wide and h chips tall, and insert all
# chips to the palette.  This is used for creating the very first
# palette after loading a new chipset image.
#
# Returns the number of the palette.
#
proc init_palette { w h } {
  global workspace_chips
  global nchips

  # create void workspace
  set i [new_workspace "palette" "default" $w $h]

  # fill the workspace with chips
  set id 1
  for {set y 0} {$y < $h} {incr y} {
    for {set x 0} {$x < $w} {incr x} {
      # replace void chip with a valid one
      if { $id < $nchips } {
        set workspace_chips($i,$x,$y) $id
        incr id
      }
    }
  }
  
  return $i
}


#
# Destroy existing chipset, and re-insert each chip
#
proc reset_palette {} {
  global nchips
  global chip_x1
  global chip_y1
  global cur_workspace
  global selector_widget
  global workspace_chips

  # get current workspace
  set cur $cur_workspace($selector_widget)
  if { $cur < 0 } {
    modal_dialog "Error" "No palettes present" error
    return
  }

  # confirm destruction of existing palette
  set button [modal_dialog \
      "Reset Palette..." \
      "Current palette will be destroyed, and the \
operation cannot be undone.  Really reset palette?" \
      question {Yes No} 0 ]
  if { $button == 0 } {

    # find maximum width
    set w 0
    for {set i 1} {$i < $nchips} {incr i} {
      if { $chip_x1($i) > $w } {
        set w [expr $chip_x1($i) + 1]
      }
    }

    # find maximum height
    set h 0
    for {set i 1} {$i < $nchips} {incr i} {
      if { $chip_y1($i) > $h } {
        set h [expr $chip_y1($i) + 1]
      }
    }

    # resize palette to new width
    resize_workspace $cur $w $h

    # fill workspace with void chips
    for {set y 0} {$y < $h} {incr y} {
      for {set x 0} {$x < $w} {incr x} {
        set workspace_chips($i,$x,$y) 0
      }
    }

    # insert every chip to workspace
    for {set i 1} {$i < $nchips} {incr i} {
      set x $chip_x1($i)
      set y $chip_y1($i)
      set workspace_chips($cur,$x,$y) $i
    }

    # file has now been edited
    set_edited 1

    # prevent user from un-doing the operation
    reset_undo

    # paint new palette to screen
    select_palette $cur
  }
}
