#
# tabnotebook.tcl - tabbed notebook widget
#
# This very module is based on a code taken from book "Effective Tcl/Tk
# Programming" by Mark Harrison and Michael McLennan.  The code has been
# further modified by Toni Ronkko.
#
#  Effective Tcl/Tk Programming
#    Mark Harrison, DSC Communications Corp.
#    Michael McLennan, Bell Labs Innovations for Lucent Technologies
#    Addison-Wesley Professional Computing Series
# 
# ======================================================================
#  Copyright (c) 1996-1997  Lucent Technologies Inc. and Mark Harrison
# ======================================================================
#
#
# Dec 39, 2004  Toni Ronkko <toni at users.playstation2-linux.com>
# 
# $Id: tabnotebook.tcl,v 1.1 2005/01/09 20:54:46 tronkko Exp $
#
# $Log: tabnotebook.tcl,v $
# Revision 1.1  2005/01/09 20:54:46  tronkko
# Initial revision
#
#
#

option add *Tabnotebook.tabs.background #666666 widgetDefault
option add *Tabnotebook.margin 6 widgetDefault
option add *Tabnotebook.tabColor #a6a6a6 widgetDefault
option add *Tabnotebook.activeTabColor #d9d9d9 widgetDefault
option add *Tabnotebook.tabFont \
    -*-helvetica-bold-r-normal--*-120-* widgetDefault


#
# Create tabbed notebook widget.
#
proc tabnotebook_create { win } {
  global tn_info

  frame $win -class Tabnotebook
  canvas $win.tabs -highlightthickness 0
  pack $win.tabs -fill x

  notebook_create $win.notebook
  pack $win.notebook -expand yes -fill both

  set tn_info($win-tabs) ""
  set tn_info($win-current) ""
  set tn_info($win-pending) ""

  # switch pages with keyboard
  bind $win <Control-Tab> "tabnotebook_next $win; break"
  return $win
}


#
# Get name of currently selected page.
#
proc tabnotebook_current { win } {
  global tn_info
  return $tn_info($win-current)
}


#
# Select next page.
#
proc tabnotebook_next { win } {  
  global tn_info

  # get number of pages
  set pages [llength $tn_info($win-tabs)]

  # get current page
  set curpage [lsearch -exact $tn_info($win-tabs) $tn_info($win-current)]

  # compute index of next page
  if { [expr $curpage+1]  < $pages } {
    set i [expr $curpage + 1]
  } else {
    set i 0
  }

  # switch to next page
  tabnotebook_display $win [lindex $tn_info($win-tabs) $i]
}


#
# Create named page in tabbed notebook win.  Returns name of frame widget
# that contains the page.  Use the path name to insert components to
# the page.
#
proc tabnotebook_page { win name } {
  global tn_info

  set page [notebook_page $win.notebook $name]
  lappend tn_info($win-tabs) $name
  
  if { $tn_info($win-pending) == "" } {
    set id [after idle [list tabnotebook_refresh $win]]
    set tn_info($win-pending) $id
  }
  return $page
}


#
# internal: update tabs after inserting a new page.
#
proc tabnotebook_refresh { win } {
  global tn_info

  # erase all tabs from display
  $win.tabs delete all

  set margin [option get $win margin Margin]
  set color [option get $win tabColor Color]
  set font [option get $win tabFont Font]
  set x 2
  set maxh 0

  # re-insert each tab
  foreach name $tn_info($win-tabs) {
    set id [$win.tabs create text \
        [expr $x+$margin+2] [expr -0.5*$margin] \
        -anchor sw -text $name -font $font \
        -tags [list $name]]
    
    set bbox [$win.tabs bbox $id]
    set wd [expr [lindex $bbox 2]-[lindex $bbox 0]]
    set ht [expr [lindex $bbox 3]-[lindex $bbox 1]]
    if { $ht > $maxh } {
      set maxh $ht
    }

    # tab around the text
    $win.tabs create polygon 0 0  $x 0 \
        [expr $x+$margin*0.3] [expr -$ht-$margin*0.6] \
        [expr $x+$margin] [expr -$ht-$margin] \
        [expr $x+$margin+$wd] [expr -$ht-$margin] \
        [expr $x+$margin+$wd+$margin*0.7] [expr -$ht-$margin*0.6] \
        [expr $x+$wd+2*$margin] 0 \
        2000 0  2000 10  0 10 \
        -outline black -fill $color \
        -tags [list $name tab tab-$name] 
    
    $win.tabs raise $id
    
    $win.tabs bind $name <ButtonPress-1> \
        [list tabnotebook_display $win $name]
    
    set x [expr $x+$wd+2*$margin]
  }
  set height [expr $maxh+2*$margin]
  $win.tabs move all 0 $height
  
  $win.tabs configure -width $x -height [expr $height+4]

  # re-select current tab
  if { $tn_info($win-current) != "" } {
    tabnotebook_display $win $tn_info($win-current)
  } else {
    tabnotebook_display $win [lindex $tn_info($win-tabs) 0]
  }
  set tn_info($win-pending) ""
}


#
# Switch pages in tabbed notebook.
#
proc tabnotebook_display {win name} {
  global tn_info

  # show new page
  notebook_display $win.notebook $name

  # normalize color of old tab
  set normal [option get $win tabColor Color]
  $win.tabs itemconfigure tab -fill $normal

  # activate tab
  set active [option get $win activeTabColor Color]
  $win.tabs itemconfigure tab-$name -fill $active
  $win.tabs raise $name
  
  # remember current page
  set tn_info($win-current) $name
}
