/*
 * birds.c - birds flying all over screen
 * 
 * Copyright (C) 2005  Toni Ronkko
 * 
 * Peach Gardens is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * Peach Gardens is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with Peach Gardens; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * 
 * Dec 6 2004, Toni Ronkko <toni at users.playstation2-linux.com>
 *
 * $Id: birds.c,v 1.3 2005/06/22 13:08:53 tronkko Exp $
 *
 * $Log: birds.c,v $
 * Revision 1.3  2005/06/22 13:08:53  tronkko
 * disable birds on maemo platform
 *
 * Revision 1.2  2005/04/03 16:28:39  tronkko
 * move birds with view
 *
 * Revision 1.1  2005/01/09 21:10:27  tronkko
 * Initial revision
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>

#include "options.h"
#include "feat.h"
#include "graphics.h"
#include "image.h"
#include "locate.h"
#include "ticks.h"
#include "birds.h"
#include "view.h"
#include "debug.h"


/* disable birds on Maemo due to poor bitblt performance */
#ifdef MAEMO
# undef ENABLE_BIRDS
# define ENABLE_BIRDS 0
#endif

/* distance beyound screen boundaries at which to remove bird */
#define BIRD_DIST 100

/* number of animation phases */
#define ANIM_PHASES 6

/* name of pcx image file containing birds */
#define IMAGE_FILE "birds.pcx"

/* width and height of a single bird image in the birds image file */
#define BIRDW 24
#define BIRDH 32

/* number of animations per flying direction in the image file */
#define MAX_PHASES 3

/* number of direction for which there is an animation in the image file */
#define MAX_DIRECTIONS 4

/* number of distinct birds per image row */
#define BIRDS_PER_ROW 3

/* number of distinct birds per image column */
#define BIRDS_PER_COLUMN 2

/* number of distinct birds in image file */
#define IMAGES (BIRDS_PER_COLUMN*BIRDS_PER_ROW)



/*
 * Information on a bird.  The structure is used internally for registering
 * birds on the sky.
 *
 * The speed of the bird flying through the sky is based in ticks.  Speed 
 * 1 means that bird advances about one pixels per tick.  That is, the bird
 * flies about 20 pixels per frame.
 */
struct bird
{
  int used;                                    /* is bird in air? */
  float x;                                     /* location of bird */
  float y;
  float dx;                                    /* direction and speed */
  float dy;
  int id;                                      /* image index */
  int phase;                                   /* wing phase */
  int tick;                                    /* when flapped wings */
};


/* find vacant slot in birds array */
#if ENABLE_BIRDS
static int find_free (void);
#endif


/* animation sequence */
static struct image img;

/* has the module been initialized? */
static int initialized = 0;

/* when were birds last animated */
static ticks_t animated_last = 0;

/* when were birds generated last */
static ticks_t generated_last = 0;

/* odds of a bird appearing */
static long int generation_odds = BIRD_GENERATION;

/* odds of new birds coming from other direction */
static long int direction_odds = BIRD_DIRECTION;

/* flying direction of new birds */
#if ENABLE_BIRDS
static int swarm_angle = 0;
#endif

/* information on birds flying */
static struct bird birds[MAX_BIRDS];

/*
 * Pixels offsets added to bird location.  The coordinates adjust the
 * position of a bird depending on the animation phase.  This makes the
 * birds accelerate and decelerate.
 */
#if ENABLE_BIRDS
static int offset[MAX_DIRECTIONS][ANIM_PHASES][2] =
{
  /* north */
  {{0,1}, {0,0}, {0,-1}, {0,-2}, {0,-1}, {0,0}},

  /* east */
  {{1,0}, {2,0}, {1,0}, {0,0}, {-1,0}, {0,0}},

  /* south */
  {{0,1}, {0,2}, {0,1}, {0,0}, {0,-1}, {0,0}},

  /* west */
  {{1,0}, {0,0}, {-1,0}, {-2,0}, {-1,0}, {0,0}}
};
#endif

/* images corresponding to animation phases */
#if ENABLE_BIRDS
static int anim_image[MAX_DIRECTIONS][ANIM_PHASES] =
{
  {0, 1, 2, 1, 1, 0},                          /* north */
  {0, 1, 1, 2, 2, 1},                          /* east */
  {0, 1, 1, 2, 2, 1},                          /* south */
  {0, 1, 2, 1, 1, 0}                           /* west */
};
#endif


/*
 * Insert new bird to screen coordinate (x,y), and make the bird fly to
 * direction (dx,dy) at speed s.  After inserting the bird to screen,
 * call functions draw_birds() and animate_birds() repeatedly to make the
 * bird fly across the screen.
 *
 * The bird speed is measured in pixels per tick.  Typical flying speed
 * varies between 0.1 and 0.2 pixels per tick.
 *
 * The function insert_bird() returns true if the bird was inserted and zero
 * if there was enough birds at the sky already.  A return value of zero
 * should not be considered fatal, as the limit on the number of birds is
 * rather small and birds are generated automatically to fill the sky.
 */
int
insert_bird(
    struct view *vp, float x, float y, float dx, float dy, float s)
{
#if ENABLE_BIRDS
  
  int i;
    
  /* find a vacant slot */
  i = find_free ();

  if (i >= 0) {
    /* there is room for one more bird */
    float len;
    float vx, vy;
    
    /* get view coords */
    vx = viewx(vp);
    vy = viewy(vp);

    /* mark the slot used */
    birds[i].used = 1;
    
    /* multiply normalized direction with speed to get movement vector */
    len = sqrt (dx*dx + dy*dy);
    birds[i].dx = s*dx / len;
    birds[i].dy = s*dy / len;      
        
    /* move bird into view */
    birds[i].x = x + vx;
    birds[i].y = y + vy;

    /* image of the bird */
    birds[i].id = random() % IMAGES;

    /* reset animation */
    birds[i].phase = 0;
    birds[i].tick = 0;

    /* succeed */
    return 1;
  } else {
    /* all slots taken */
    return 0;
  }

#else

  /* no birds */
  (void)vp; (void) x; (void) y; (void) dx; (void) dy; (void) s;
  return 1;
  
#endif
}


/*
 * Draw birds to current buffer.  To move the birds, also call
 * animate_birds().
 */
void
draw_birds(
    struct view *vp)
{
#if ENABLE_BIRDS
  
  int i;
  float vx, vy, vs;
  assert (initialized);
  
  /* get view coords and scale */
  vx = viewx(vp);
  vy = viewy(vp);
  vs = scale (vp);
  
  for (i = 0; i < MAX_BIRDS; i++) {
    if (birds[i].used) {
      /* draw the bird */
      int sx1, sy1, sx2, sy2;
      int dx1, dy1, dx2, dy2;
      int direction;
      int row;
      int col;
      int anim;
      int phase;

      /* compute base row and column of bird image */
      col = (birds[i].id % BIRDS_PER_ROW) * BIRDW * MAX_PHASES;
      row = (birds[i].id / BIRDS_PER_ROW) * BIRDH * MAX_DIRECTIONS;
      assert (birds[i].id < IMAGES);
      assert (col <= (BIRDS_PER_ROW - 1) * MAX_PHASES * BIRDW);
      assert (row <= (BIRDS_PER_COLUMN - 1) * MAX_DIRECTIONS * BIRDH);
      
      /* determinate bird image depending on the flying direction */
      if (birds[i].dx > 0  &&  birds[i].dx > fabs (birds[i].dy)) {
        /* right */
        direction = 1;
      } else if (birds[i].dx < 0  &&  -birds[i].dx > fabs (birds[i].dy)) {
        /* left */
        direction = 3;
      } else if (birds[i].dy < 0) {
        /* up */
        direction = 0;
      } else {
        /* down */
        direction = 2;
      }
      assert (direction < MAX_DIRECTIONS);

      /* wing flapping phase */
      anim = birds[i].phase;
      assert (anim < ANIM_PHASES);

      /* image corresponding to animation phase */
      phase = anim_image[direction][anim];
      assert (phase < MAX_PHASES);
      
      /* compute location of image within the larger one */
      sx1 = col + (phase) * BIRDW;
      sy1 = row + (direction) * BIRDH;
      sx2 = sx1 + BIRDW;
      sy2 = sy1 + BIRDH;

      /* compute position of image on screen */
      dx1 = (int) (birds[i].x - vx) * vs + offset[direction][anim][0];
      dy1 = (int) (birds[i].y - vy) * vs + offset[direction][anim][1];
      dx2 = dx1 + BIRDW * BIRD_SCALE * vs;
      dy2 = dy1 + BIRDH * BIRD_SCALE * vs;

      /* paint bird to sky */
      bitblt (&img,
          sx1, sy1, sx2, sy2, 
          dx1, dy1, dx2, dy2);
    }
  }
  
#else

  /* birds are disabled */
  (void) vp;

#endif
}


/*
 * Animate existing birds and remove birds that fly outside the screen
 * boundaries.  To have new birds generated, also call generate_birds().
 */
void
animate_birds(
    struct view *vp)
{
#if ENABLE_BIRDS
  
  int i;
  ticks_t now;
  int elapsed = 0;
  float vx, vy;
  assert (initialized);
  
  /* get view coords */
  vx = viewx(vp);
  vy = viewy(vp);
  
  /* compute the number of ticks elapsed since last time */
  now = ticks ();
  if (animated_last < now) {
    elapsed = now - animated_last;
  }
  
  /* remember the time when birds were animated */
  animated_last = now;
  
  /* animate birds unless being called twice in a row */
  if (elapsed) {
    for (i = 0; i < MAX_BIRDS; i++) {

      if (birds[i].used) {
        float x, y;
        int flaps;

        /* move bird */
        birds[i].x += birds[i].dx * elapsed;
        birds[i].y += birds[i].dy * elapsed;

        /* compute the number of wing flaps since last animation */
        flaps = (now - birds[i].tick) / BIRD_FLAP_INTERVAL;
          
        /* flap wings */
        birds[i].phase = (birds[i].phase + flaps) % ANIM_PHASES;
        birds[i].tick += flaps * BIRD_FLAP_INTERVAL;

        /* remove bird when it flies outside the screen */
        x = birds[i].x - vx;
        y = birds[i].y - vy;
        if (x < -(BIRD_DIST + BIRDW*BIRD_SCALE) 
            ||  y < -(BIRD_DIST + BIRDH*BIRD_SCALE) 
            ||  x > screenwidth() + BIRD_DIST
            ||  y > screenheight() + BIRD_DIST) {
          /* mark the bird unused */
          birds[i].used = 0;
        }        
      }
      
    }
  }

#else

  /* birds are disabled */
  (void) vp;

#endif
}


/*
 * Generate more birds to the sky at random intervals.  Birds are generated
 * in swarms that fly to very same direction.  Generating birds this way
 * makes their flying pattern look more natural.
 */
void
generate_birds(
    struct view *vp)
{
#if ENABLE_BIRDS
  
  ticks_t now;
  int elapsed = 0;
  assert (initialized);
    
  /* compute the number of ticks elapsed since last time */
  now = ticks ();
  if (generated_last < now) {
    elapsed = now - generated_last;
  }
  
  /* remember the time when birds were generated */
  generated_last = now;

  /* change direction of new birds at random intervals */
  if (elapsed  &&  direction_odds > 0) {
    if ((random() % direction_odds) < elapsed) {
      
      /* update flying direction little at a time */
      swarm_angle = (swarm_angle + (random () % 200) - 100) % 1000;
    }
  }
  
  /* generate new birds at random intervals */
  if (elapsed  &&  generation_odds > 0) {    
    /* generate new bird if only there is a slot vacant */
    if (find_free() >= 0) {
      if ((random() % generation_odds) < elapsed) {
        float s;
        float x, y;
        float dx, dy;
        float angle;
        
        /* generate random speed between 0.1 and 0.2 (pixels per tick) */
        s = ((float) (random() % 100)) * 0.001 + 0.1;
        assert (0.1 <= s  &&  s <= 0.2);

        /* compute random flying direction */
        angle = swarm_angle * M_PI * 2 * 0.001;
        dx = cosf (angle);
        dy = sinf (angle);
        
        /* avoid diagonal movement as we do not have images for them */
        if (fabs (dx) > fabs (dy)) {
          dy = dy * 0.2;
        } else {
          dx = dx * 0.2;
        }
        
        /* generate random location outside the screen space */
        if (fabs (dy) < fabs (dx)) {
          /* flying from east to west or vice versa */
          if (dx < 0) {
            x = screenwidth() + BIRDW*BIRD_SCALE;
          } else {            
            x = -BIRDW*BIRD_SCALE*2;
          }
          y = random() % screenheight();
        } else {
          /* flying from north to south or vice versa */
          if (dy < 0) {
            y = screenheight() + BIRDH*BIRD_SCALE;
          } else {
            y = -BIRDH*BIRD_SCALE;
          }
          x = random() % screenwidth();
        }

        /* activate bird */
        insert_bird (vp, x,y, dx,dy, s);
      }
    }
  }

#else
  
  /* birds are disabled */
  (void) vp;
  
#endif
}


/*
 * Set odds of new birds being generated and new birds appearing from
 * another direction.  The bigger the argument, the less frequently a new
 * bird is generated or the direction is changed.  To disable generation of
 * new birds or change of directions completely, pass a zero as an argument
 * or stop calling generate_birds() function.
 *
 * The odds are tied to the tick counter.  Bird generation odds of 20 means
 * that a new bird is generated once per frame, which is about 20 ticks.
 * Likewise, number 5000 means that a new bird is born about once every
 * fifth second.
 */
void
bird_frequency (long int generation, long int change)
{
  generation_odds = generation;
  direction_odds = change;
}


/*
 * Remove all birds from the sky instantly and reset bird generation odds.
 * The function is typically called after switching from a screen that does
 * not have birds.  For example, if the player enters an open space after
 * being in a cave, then you may want reset the birds so that they do not
 * appear at the exact same positions in which they were when the player
 * entered the cave.
 */
void
reset_birds (void)
{
  int i;

  /* mark all birds free */
  for (i = 0; i < MAX_BIRDS; i++) {
    birds[i].used = 0;
  }

  /* reset counters just in case */
  generated_last = 0;
  animated_last = 0;

  /* reset intervals */
  generation_odds = BIRD_GENERATION;
  direction_odds = BIRD_DIRECTION;
}


/*
 * Initialize birds by loading the associated pcx image files.  Note that
 * the module only controls birds that fly on the sky.  Birds on the
 * ground are not handled by this very module.
 */
void 
init_birds (void)
{
  assert (!initialized);
  
  /* load images */
  init_image (&img, locate(IMAGE_FILE));
  
  /* image loaded */
  initialized = 1;

  /* no birds are currently in the air */
  reset_birds ();
}


/*
 * De-initialize module.
 */
void
done_birds (void)
{
  if (initialized) {
    done_image (&img);
  }
  initialized = 0;
}


/*
 * Find a vacant slot in the birds array, and return its index.  Returns -1
 * if all the slots are taken.  Note that return value of zero means that
 * slot 0 was available.  A return value of zero is not an error!
 */
#if ENABLE_BIRDS
static int
find_free (void)
{
  int i;
  for (i = 0; i < MAX_BIRDS; i++) {
    if (!birds[i].used) {
      /* vacant slot found */
      return i;
    }
  }
  
  /* no slots available */
  return -1;
}
#endif
