/*
 * chipset.c - manage chipsets
 * 
 * Copyright (C) 2005  Toni Ronkko
 * 
 * Peach Gardens is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * Peach Gardens is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with Peach Gardens; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * 
 * Dec 23 2004, Toni Ronkko <toni at users.playstation2-linux.com>
 *
 * $Id: chipset.c,v 1.3 2005/04/03 16:20:18 tronkko Exp $
 *
 * $Log: chipset.c,v $
 * Revision 1.3  2005/04/03 16:20:18  tronkko
 * using abort_msg
 *
 * Revision 1.2  2005/01/09 21:10:54  tronkko
 * simplified resource management
 *
 * Revision 1.1  2004/12/29 00:08:07  tronkko
 * Initial revision
 *
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "config.h"
#include "debug.h"
#include "chip.h"
#include "chipset.h"

#ifdef HAVE_ALLOCA_H
# include <alloca.h>
#endif


/*
 * Get pointer to a chip within chipset.  The function is only used when
 * NDEBUG in not set.  Otherwise, get_chip_macro() is used instead.
 */
struct chip *
get_chip_function(
    struct chipset *chp, int i)
{
  struct chip *cp;
  
  /* check input args */
  assert (chp != NULL);
  assert (chp->chips != NULL);
  assert (0 <= i  &&  i < chp->nchips);

  /* get chip */
  cp = get_chip_macro (chp, i);

  /* check chip */
  assert (cp->id == i);
  return cp;
}


/*
 * Allocate memory for a chipset with n chips and read chipset image fn into
 * memory.  The function requires a working graphics subsystem, and does not
 * return on error.
 */
void 
init_chipset(
    struct chipset *chp, int n, int w, int h, const char *fn)
{
  struct chip *p;
  struct image *imp;

  assert (chp != NULL);
  assert (n > 0);
  assert (w > 0  &&  h > 0);
  
  /* allocate memory for chips */
  p = malloc (n * sizeof (struct chip));
  DEBUG (("%p: alloc %d chips", p, n)); 
  if (p != NULL) {
    int i;
    char *buffer;
    int len;
    
    /* initialize general properties */
    chp->chips  = p;
    chp->nchips = n;
    chp->chipw  = w;
    chp->chiph  = h;

    /* open image file */
    imp = &chp->img;
    init_image (imp, NULL);
    open_image (imp, fn);
    
    /* allocate temporary buffer for image data */
    len = imp->width * imp->height;
    buffer = alloca (len);
    assert (buffer != NULL);
    
    /* read image data into buffer */
    if (read_image (imp, buffer, len)) {
      /* prepare image for drawing */
      if (register_image (imp, buffer, len)) {
        int x, y;
        
        /* generate default chipset */
        i = 1;
        for (y = 0; y < (imp->height + h - 1) / h; y++) {
          for (x = 0; x < (imp->width + w - 1) / w; x++) {
            if (i < n) {
              p[i].id = i;
              p[i].x  = x;
              p[i].y  = y;
              p[i].z  = 0;
              p[i].bg = 0;
              i++;
            }
          }
        }

        /* create void chip */
        p[0].id = 0;
        p[0].x  = -1;
        p[0].y  = -1;        
        p[0].z  = -1;        
        p[0].bg = 0;
        
      } else {
        abort_msg ("cannot register %s", fn);
      }
    } else {
      abort_msg ("error reading %s", fn);
    }
    close_image (imp);
    
  } else {
    /* not enough memory for chipset data */
    out_of_memory ();
  }
}


/*
 * Release a chipset immediately from memory.
 */
void
done_chipset(
    struct chipset *chp)
{
  /* check inputs */
  assert (chp != NULL);
  assert (chp->chips != NULL);
  
  /* release chipset image */
  done_image (&chp->img);
  
  /* release memory allocated for chips */
  if (chp->chips != NULL) {
    DEBUG (("%p: free chips", chp->chips)); 
    free (chp->chips);
  }

  /* de-initialize member variables */
  chp->chipw  = 0;
  chp->chiph  = 0;
  chp->nchips = 0;
  chp->chips  = NULL;
}
