/*
 * clouds.c - cloud animation
 * 
 * Copyright (C) 2005  Toni Ronkko
 * 
 * Peach Gardens is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * Peach Gardens is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with Peach Gardens; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * 
 * Dec 27 2004, Toni Ronkko <toni at users.playstation2-linux.com>
 *
 * $Id: clouds.c,v 1.3 2005/06/22 13:04:51 tronkko Exp $
 *
 * $Log: clouds.c,v $
 * Revision 1.3  2005/06/22 13:04:51  tronkko
 * disable clouds on maemo platform
 *
 * Revision 1.2  2005/04/03 16:29:07  tronkko
 * move with view
 *
 * Revision 1.1  2005/01/09 21:10:58  tronkko
 * Initial revision
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>

#include "config.h"
#include "feat.h"
#include "clouds.h"
#include "graphics.h"
#include "image.h"
#include "view.h"
#include "ticks.h"
#include "locate.h"
#include "debug.h"


/* Relative speed of clouds compared to ground speed.  Value of 1 means
 * that clouds move the same speed as ground when player is moving, and
 * thus create an illusion of fog.  Increasing this constant affects
 * illusion of cloud altitude.  (Clouds move faster than ground as if clouds
 * were closer to the viewer.) */
#define CLOUD_ALT 1.5

/* Height of cloud layer.  Creates illusion of depth when moving. */
#define CLOUD_HEIGHT 0.2

/* Allowed deviation of cloud layer from general wind direction (degrees).
 * This settings adds a bit of randomness to clouds. */
#define WIND_DEVIATION 16

/* maximum number of cloud layers */
#define MAX_LAYERS 3

/* number of decimal bits in fixed point numbers */
#define FIXED_BITS 10

/* disable clouds if not using textures on x11, or using Maemo */
#if defined(X11) && defined(OPENGL) && !defined(PS2DEV)
# if !USE_TEXTURES
#   undef ENABLE_CLOUDS
#   define ENABLE_CLOUDS 0
# endif
#endif
#ifdef MAEMO
# undef ENABLE_CLOUDS
# define ENABLE_CLOUDS 0
#endif


/* fixed point data type */
typedef int fixed;


/* draw cloud layer */
#if ENABLE_CLOUDS
static void draw_layer (int sx, int sy, int scale);
#endif


/* image of a cloud (single layer) */
static struct image cloud;

/* layer positions */
static fixed layer_x[MAX_LAYERS];
static fixed layer_y[MAX_LAYERS];

/* relative speed of layer */
static float layer_speed[MAX_LAYERS];

/* scale layer image by factor */
static int layer_scale[MAX_LAYERS];

/* wind direction */
static float wind_dir = 0.5;

/* wind speed in pixels per tick, 0.05 means pixel every 20 ticks */
static float wind_speed = 0.05;

/* current number of layers */
static int nlayers = MAX_LAYERS;

/* when were clouds last animated */
static ticks_t animated_last = 0;

/* module initialized? */
static int initialized = 0;



/*
 * Draw clouds over current scene.  If global feature ENABLE_CLOUDS is set
 * to zero, no clouds are drawn.  Drawing clouds requires at least three
 * passes over whole screen space, and this can be time-consuming unless the
 * process is optimized in hardware.  Moreover, enlarged textures do not
 * work properly in X11 graphics interface unless textures are used.
 */
void
draw_clouds(
    struct view *vp)
{
#if ENABLE_CLOUDS 
  int i;
  int x, y;
  int vx, vy;
  assert (initialized);
  assert (nlayers <= MAX_LAYERS);
  
  /* paint all layers */
  for (i = 0; i < nlayers; i++) {

    /*
     * Get view offset for cloud layer.  Multiplying the offset by
     * CLOUD_ALT helps to create illusion of depth; when moving the
     * player, clouds look like being above the ground.  Without this,
     * clouds remain fixed to environment and this look more like fog.
     * Multiplying the offset with CLOUD_HEIGHT creates illusion of cloud
     * height.  If the variable is zero, all layers are tightly packed.  If
     * the value is about 0.2, clouds have thickness that can be viewed when
     * moving player.
     */
    vx = (int) (viewx(vp) * CLOUD_ALT * scale(vp) * (1 + i * CLOUD_HEIGHT));
    vy = (int) (viewy(vp) * CLOUD_ALT * scale(vp) * (1 + i * CLOUD_HEIGHT));
    
    /* compute upper-left position of layer */
    x = ((layer_x[i] >> FIXED_BITS) + vx) & (256 * layer_scale[i] - 1);
    y = ((layer_y[i] >> FIXED_BITS) + vy) & (256 * layer_scale[i] - 1);

    /* paint layer */
    draw_layer (-x, -y, layer_scale[i]);
  }
#else
  (void) vp;
#endif
}


/*
 * Animation should be reset every now and then to ensure that animation is
 * smooth.  Invoke reset_clouds() whenever the player enters a scene that
 * has clouds.
 */
void
animate_clouds (void)
{
  ticks_t now;
  int elapsed = 0;
  int i;
  fixed dx, dy;
  fixed x, y;
  fixed max;
  float s, dev;
  assert (initialized);
  assert (nlayers <= MAX_LAYERS);

  /* compute the number of ticks elapsed since last time */
  now = ticks ();
  if (animated_last < now) {
    elapsed = now - animated_last;
  }

  /* remember the time when birds were animated */
  animated_last = now;  
    
  /* update layer positions */
  for (i = 0; i < nlayers; i++) {
    /* get current position */
    x = layer_x[i];
    y = layer_y[i];
    
    /* compute deviation from wind dir */
    dev = ((float)i / nlayers - 0.5) * 2 * M_PI * 2 / 360 * WIND_DEVIATION;

    /* compute wind direction vector */
    s = wind_speed * layer_speed[i] * elapsed;
    dx = (fixed) ((cos(wind_dir + dev) * s) * (1<<FIXED_BITS));
    dy = (fixed) ((sin(wind_dir + dev) * s) * (1<<FIXED_BITS));
    
    /* update position */
    x += dx;
    y += dy;

    /* compute max coordinate */
    max = (256 * layer_scale[i]) << FIXED_BITS;
    assert (max > 0);
    
    /* ensure that position stays in range */
    x = x % max;
    if (x < 0) x = max + x;
    y = y % max;
    if (y < 0) y = max + y;

    /* save modified position */
    layer_x[i] = x;
    layer_y[i] = y;
  }
}



/*
 * Draw cloud layer at (sx,sy) with given scale.  The coordinate must be
 * zero or negative to overlay the whole scene with clouds.  Scale must be
 * 1, 2, 4 or 8.
 */
#if ENABLE_CLOUDS 
static void
draw_layer(
    int sx, int sy, int scale)
{
  int x, y;
  int a, b;
  int scrw, scrh;

  assert (initialized);
  assert (sx <= 0);
  assert (sy <= 0);
  assert (scale > 0);
  assert (scale == 1  ||  scale == 2  ||  scale == 4  ||  scale == 8);
  
  /* get screen dimensions */
  scrw = screenwidth();
  scrh = screenheight();
  
  /* draw layer in blocks */
  for (y = sy; y < scrh; y += 256*scale) {
    for (x = sx; x < scrw; x += 256*scale) {
      
      /* compute width of sub-block */
      int block = 256/scale;
      
      /* draw one block as several 256x256 sub-blocks */
      for (b = 0; b < scale; b++) {
        for (a = 0; a < scale; a++) {
          
          /* compute position on screen */
          int tx = x + a*block*scale;
          int ty = y + b*block*scale;
          
          /* compute position in cloud image */
          int x1 = a * block;
          int y1 = b * block;
          int x2 = x1 + block;
          int y2 = y1 + block;
          
          /* draw block unless it would be clipped */
          if (-256 < tx  &&  tx < scrw  &&  -256 < ty  &&  ty < scrh) {
            bitblt (&cloud,
                x1, y1, x2, y2, 
                tx, ty, tx+256, ty+256);
          }
        }
        
      }
    }
  }
}
#endif


/*
 * Reset position of cloud layers and generate random wind direction.  The
 * function is often used when the player enters a scene with clouds.
 *
 * The function is called at module initialization.  Hence, variable
 * initialized may be zero at the time of the function call.
 */
void
reset_clouds (void)
{
  int i;
  assert (nlayers <= MAX_LAYERS);
  
  /* use all layers */
  nlayers = MAX_LAYERS;
  
  /* reset layers */
  for (i = 0; i < nlayers; i++) {
    /* generate pseudo-random position */
    layer_x[i] = ((131*i) % 256) << FIXED_BITS;
    layer_y[i] = (( 51*i) % 256) << FIXED_BITS;

    /* generate pseudo-random speed */
    switch (i % 3) {
    case 0:  layer_scale[i] = 1; break;
    case 1:  layer_scale[i] = 2; break;
    case 2:  layer_scale[i] = 4; break;
    default: layer_scale[i] = 1; break;
    }

    /* move scaled layers faster */
    layer_speed[i] = 0.5 + ((float) layer_scale[i] - 1) / 5;
  }
  
  /* generate random wind direction */
  wind_dir = (random() % 1000) * 0.001 * 3.14 * 2;
  wind_speed = 0.05;

  /* reset animation */
  animated_last = ticks ();
}


/*
 * Initialize clouds module.
 */
void
init_clouds (void)
{
  int i;
  const char *fn;
  char buffer[256*256];
  assert (!initialized);

  /* find true location of the image file */
  fn = locate("cloud.pcx");
  
  /* open image file */
  init_image (&cloud, NULL);
  open_image (&cloud, fn);
  
  /* read image into memory buffer */
  if (read_image (&cloud, buffer, sizeof (buffer))) {
      
    /* convert red channel into alpha */
    for (i = 0; i < (int) cloud.ncolors; i++) {
      int alpha = cloud.palette[i].r;

      /* store color */
      cloud.palette[i].a = alpha;
      cloud.palette[i].r = 0xff;
      cloud.palette[i].g = 0xff;
      cloud.palette[i].b = 0xff;
    }
    
    /* prepare image (and new color table) for drawing */
    if (register_image (&cloud, buffer, sizeof (buffer))) {
      /* generate random wind direction */
      reset_clouds ();
    } else {
      abort_msg ("cannot register %s", fn);
    }
  } else {
    abort_msg ("error reading %s", fn);
  }
  close_image (&cloud);
  
  /* mark the module initialized */
  initialized = 1;
}


/*
 * De-initialize clouds.
 */
void
done_clouds (void)
{
  assert (initialized);

  /* release cloud image */
  done_image (&cloud);

  /* mark the module un-initialized */
  initialized = 0;
}
