/*
 * font.c - bitmap font
 * 
 * Copyright (C) 2005  Toni Ronkko
 * 
 * Peach Gardens is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * Peach Gardens is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with Peach Gardens; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * 
 * Nov 16 2004, Toni Ronkko <toni at users.playstation2-linux.com>
 *
 * $Id: font.c,v 1.3 2005/06/22 13:10:10 tronkko Exp $
 *
 * $Log: font.c,v $
 * Revision 1.3  2005/06/22 13:10:10  tronkko
 * support for non-bitmap fonts
 *
 * Revision 1.2  2005/04/03 16:21:51  tronkko
 * using abort_msg
 *
 * Revision 1.1  2005/01/09 21:11:22  tronkko
 * Initial revision
 *
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "config.h"
#include "font.h"
#include "graphics.h"
#include "text.h"
#include "image.h"
#include "debug.h"

#ifdef HAVE_ALLOCA_H
# include <alloca.h>
#endif


/* current text font */
struct font *cur_font = NULL;

/* default info function for bitmap fonts */
static const struct glyph *info_bitmap(int c);

/* default draw function for bitmap fonts */
static int draw_bitmap (int c);


/*
 * Get glyph for character c.  This very function can be re-defined by
 * various fonts by overwriting info field in font structure.
 */
static const struct glyph*
info_bitmap(
    int c)
{
  assert (cur_font != NULL);
  assert (0 <= c  &&  c < MAX_GLYPHS);
  return &cur_font->glyphs[c];
}

/*
 * Draw glyph c to current text pos, and returns character width.  This very
 * function is used by default for bitmap fonts, but can be overwritten by
 * modifying draw field in font structure.
 */
static int 
draw_bitmap(
    int c)
{
  int w = 0;                                   /* character width */
  const struct glyph *gp;                      /* glyph data */
  
  /* print char */
  gp = cur_font->info(c);
  if (gp != NULL) {
    /* character must be found within image */
    assert (0 <= gp->x  &&  gp->x+gp->w <= cur_font->data.width);
    assert (0 <= gp->y  &&  gp->y+gp->h <= cur_font->data.height);
    
    /* paint character to screen */
    if (c > ' ') {
      textblt(&cur_font->data,
          gp->x, gp->y, gp->w, gp->h,
          textx() + gp->lead, texty() - gp->ascent);
    }
    
    /* compute width of character */
    w = gp->lead + gp->w + gp->follow;
  }
  return w;
}

/*
 * Compute width of string s in pixels, assuming current font.  The
 * function can handle variable width fonts, and it should be used instead
 * of trying to deduce width from string lenght.
 *
 * The function does not consider line feeds.
 */
int
textwidth(
    const char *s)
{
  int w = 0;
  while (*s != '\0') {

    /* locate character */
    const struct glyph *gp = getglyph (*s);
    if (gp != NULL) {
      /* add character width */
      w += gp->w + gp->lead + gp->follow;
    }
    
    ++s;
  }
  return w;
}


/*
 * Get height of string s in pixels in the current font.  The height
 * may include empty space above or below the font.
 */
int
textheight(
    const char *s)
{
  int ascent = 0;
  int descent = 0;
  while (*s != '\0') {
    
    /* locate character */
    const struct glyph *gp = getglyph (*s);
    if (gp != NULL) {
      /* find tallest glyph */
      if (gp->ascent > ascent) {
        ascent = gp->ascent;
      }
      if (gp->descent > descent) {
        descent = gp->descent;
      }
    }
    
    ++s;
  }
  return ascent + descent;
}


/*
 * Set text font used for further output.  To invalidate the current font,
 * set fptr to NULL.
 *
 * The font structure passed to the function should be allocated from static
 * or otherwise permanent memory area.  Otherwise, be sure to invalidate the
 * font when the structure is destroyed.
 */
void
setfont(
    struct font *fptr)
{
  /* set/invalidate current font */
  cur_font = fptr;
}


/*
 * Get pointer to current font.  The function is only used in debug mode.
 * Otherwise, getfont_macro() is used instead.
 */
struct font *
getfont_function (void)
{
  /* get current font */
  struct font *fnt = getfont_macro ();

  /* must have font */
  assert (fnt != NULL);
  return fnt;
}


/*
 * Get information on a glyph.
 */
const struct glyph *
getglyph_function(
    int c) 
{
  assert (0 <= c  &&  c < MAX_GLYPHS);
  assert (getfont() != NULL);
  return getglyph_macro(c);
}


/*
 * Initialize generic font.
 */
void
init_font(
    struct font *fptr)
{
  assert (fptr != NULL);

  /* reset all field */
  memset (fptr, 0, sizeof (*fptr));

  /* initialize with empty image bitmap */
  init_image (&fptr->data, NULL);
}


/*
 * Initialize font structure and read font image from file fn.  Also sets
 * character width and height into the font structure.  To adjust the space
 * between two successive characters, you may want to modify hspace and
 * vspace variables in the resulting font structure.
 *
 * The newly loaded font becomes the current font for successive printing
 * as a result of the function.
 */
void 
init_bitmap_font(
    struct font *fptr, const char *fn, const struct glyph *glyphs)
{
  struct image *imp;
  char *buffer;
  int len;
  
  assert (fptr != NULL);
  assert (fn != NULL);

  /* general properties of a font */
  init_font (fptr);
  init_glyphs (fptr->glyphs, glyphs, MAX_GLYPHS);

  /* set up default handlers */
  fptr->info = info_bitmap;
  fptr->draw = draw_bitmap;
  
  /* open image file */
  imp = &fptr->data;  
  init_image (imp, NULL);
  open_image (imp, fn);
      
  /* allocate temporary buffer for image data */
  len = imp->width * imp->height;
  buffer = alloca (len);
  assert (buffer != NULL);
  
  /* read image data into buffer */
  if (read_image (imp, buffer, len)) {
    int i;
    
    /*
     * Insert partially transparent colors to the palette.  Since many
     * file formats cannot handle transparency, derive transparency from
     * rgb colors.  To make this work, the letters should be drawn white
     * and the background should be drawn black.  The characters should
     * be anti-aliased with shades of gray.
     *
     * Note the shades of gray in the color palette will be replaced by
     * color #7f7f7f.  This is required to preserve the colors in PS2.
     * OpenGL implementation does not seem to care about the color we
     * specify here.
     */
    for (i = 0; i < imp->ncolors; i++) {
      imp->palette[i].a = imp->palette[i].r;
#ifdef PS2DEV          
      imp->palette[i].r = 0x7f;
      imp->palette[i].g = 0x7f;
      imp->palette[i].b = 0x7f;
#else /*OPENGL*/
      imp->palette[i].r = 0xff;
      imp->palette[i].g = 0xff;
      imp->palette[i].b = 0xff;
#endif /*OPENGL*/
    }  
    
    /*
     * Prepare image for drawing.  This has to be done after fixing the
     * palette or else the changes made to the palette have no effect.
     */
    if (register_image (imp, buffer, len)) {
      /* start using the font */
      setfont (fptr);
    } else {
      abort_msg ("cannot register %s", fn);
    }
  } else {
    abort_msg ("error reading %s", fn);
  }
    
  /* close image file */
  close_image (imp);
}


/*
 * Release all resources tied to a font.  The function also invalidates the
 * current font, should it be the active one at the moment.  Hence, if you
 * use this very function at the middle of a program, be sure to set a new
 * font prior to drawing any further text.
 */
void
done_font(
    struct font *fptr)
{
  assert (fptr != NULL);
  
  /* invalidate current font */
  if (cur_font == fptr) {
    cur_font = NULL;
  }

  /* get rid of the image data */
  done_image (&fptr->data);

  /* invalidate structure to notice invalid use */
#ifndef NDEBUG
  memset (fptr, 0xcc, sizeof (*fptr));
#endif
}


/*
 * Initialize indexed array of glyphs.  This is used with many bitmap
 * fonts.
 */
void
init_glyphs(
    struct glyph *glyphs, const struct glyph *gp, int nglyphs)
{
  /* zero all glyphs */
  memset (glyphs, 0, sizeof (glyphs[0]) * nglyphs);

  /* index glyphs */
  while (gp->code != '\0') {
    assert (0 <= gp->code  &&  gp->code < nglyphs);
    
    /* copy glyph */
    memcpy (&glyphs[gp->code], gp, sizeof (*gp));
    
    ++gp;
  }
}
