/*
 * image.c - I/O of image files
 * 
 * Copyright (C) 2005  Toni Ronkko
 * 
 * Peach Gardens is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * Peach Gardens is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with Peach Gardens; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * 
 * Nov  6 2004, Toni Ronkko <toni at users.playstation2-linux.com>
 *
 * $Id: image.c,v 1.3 2005/04/03 16:22:31 tronkko Exp $
 *
 * $Log: image.c,v $
 * Revision 1.3  2005/04/03 16:22:31  tronkko
 * shadow pixel
 *
 * Revision 1.2  2005/01/09 21:06:30  tronkko
 * find_color(), do not return on failure
 *
 * Revision 1.1  2004/11/20 10:57:05  tronkko
 * Initial revision
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "config.h"
#include "debug.h"
#include "image.h"

#ifdef HAVE_ALLOCA_H
# include <alloca.h>
#endif


/* size of PCX file header in disk */
#define PCX_HEADER_SIZE 128


/* initialize for pcx file format */
static int init_pcx (struct image *imp);


/*
 * Header of pcx file in disk.  Note that you must not insert your very
 * own fields here, as the contents of the structure are read from disk.
 * For more information on pcx file format and the fields, please see:
 * http://courses.ece.uiuc.edu/ece390/books/labmanual/graphics-pcx.html
 */
struct pcx_header {
  char manufacturer;                           /* 0x0a = ZSoft PCX */
  char version;
  char encoding;                               /* 1 = RLE */
  char nbits;                                  /* bits per pixel */
  unsigned char xmin[2];
  unsigned char ymin[2];
  unsigned char xmax[2];
  unsigned char ymax[2];
  unsigned char hres[2];                       /* dots per inch */
  unsigned char vres[2];
  char palette[16*3];
  char _filler1;
  char nplanes;                                /* # color planes */
  unsigned char rowlen[2];                     /* bytes per scan line */
  unsigned char palettetype[2];                /* 1=color, 2=grayscale */
};


/*
 * Open image file and read header.  Upon return, image width, height and
 * palette are set.  The function does not return on error.
 */
void 
open_image(
    struct image *imp, const char *fn)
{
  assert (imp != NULL);
  assert (imp->FP == NULL);

  imp->FP = fopen (fn, "r");
  if (imp->FP != NULL) {    
    /* read header and palette */
    if (!init_pcx (imp)) {
      abort_msg ("error reading file: %s", fn);
    } 
  } else {
    abort_msg ("cannot open file: %s", fn);
  }
}


/*
 * Close image file without destroying image structure.
 */
void
close_image(
    struct image *imp)
{
  assert (imp != NULL);
  
  /* close image file */
  if (imp->FP != NULL) {
    fclose (imp->FP);
    imp->FP = NULL;
  }
}


/*
 * Read image file fn to memory and perform all necessary initializations to
 * use the image for further drawing.  Currently, only pcx file format
 * is supported for indexed RGB images with 8 bites per pixel.
 *
 * The function prepares images for drawing.  Thus, any graphics libraries
 * must be initialized at the time of the function call.
 */
void 
init_image(
    struct image *imp, const char *fn)
{
  char *buffer;
  int len;
  
  /* initialize image structure */
  assert (imp != NULL);
  memset (imp, 0, sizeof (*imp));
  
  /* read image */
  if (fn != NULL) {
    open_image (imp, fn);
    
    /* allocate temporary buffer for image data */
    len = imp->width * imp->height;
    buffer = alloca (len);
    assert (buffer != NULL);
    
    /* read image data into buffer */
    if (read_image (imp, buffer, len)) {
      /* prepare image for drawing */
      if (!register_image (imp, buffer, len)) {
        abort_msg ("cannot register %s", fn);
      }
    } else {
      abort_msg ("error reading %s", fn);
    }
      
    /* image is now in memory, and we do not need the file */
    close_image (imp);
  }
}


/*
 * Release image from memory.
 */
void
done_image (struct image *imp)
{
  assert (imp != NULL);

  /* ensure that image file is closed properly */
  close_image (imp);

  /* release hardware specific data */
  unregister_image (imp);
}


/*
 * Read image from file to memory buffer.  The image is read in indexed mode 
 * where each pixel is actually a reference to color lookup table.  Image
 * files with 24 bits per pixel are not currently supported at all.
 */
int
read_image(
    struct image *imp, void *buffer, int len)
{
  int ncols;
  char *buffermax = (char*)buffer + len;

  assert (imp != NULL);
  assert (imp->FP != NULL);
  assert (buffer != NULL);

  /* ensure that there is enough space in buffer */
  assert (len >= imp->width * imp->height);
  if (len < imp->width * imp->height) {
    DEBUG (("buffer too small"));
    return 0;
  }

  /* seek to beginning of image data */
  if (fseek (imp->FP, PCX_HEADER_SIZE, SEEK_SET) == /*OK*/0) {
    int y;
    int used[256] = { 0 };                     /* which colors are used */
    char *dest = buffer;
    
    /* read all scan lines */
    for (y = 0; y < imp->height; y++) {
      int pixel, repeat, c, i;
      void *scanmax;
      
      /* read one scan line */
      scanmax = dest + imp->width;
      while (dest < (char*)scanmax) {
        /* get byte from file */
        c = fgetc (imp->FP);
        if (c == EOF) {
          DEBUG (("unexpected EOF"));
          return 0;
        }
        
        /* repeat pixel? */
        if ((c & 0xc0) == 0xc0) {
          /* do repeat, pixel value follows */
          repeat = c & 0x3f;
          pixel = fgetc (imp->FP);
          if (pixel == EOF) {
            DEBUG (("unexpected EOF"));
            return 0;
          }
        } else {
          /* no repeat */
          repeat = 1;
          pixel = c;
        }
        assert (repeat > 0);

        /* remember which pixel values have been used */
        assert (0 <= pixel  &&  pixel < 256);
        used[pixel]++;
        
        /* save pixel to buffer */
        for (i = 0; i < repeat; i++) {
          assert (dest < buffermax);
          *dest++ = pixel;
        }
      }
      
      /* skip a filler byte (rows must be even size) */
      if ((imp->width & 1) != 0) {
        c = fgetc (imp->FP);
      }
    }

    /* fix number of colors */
    for (ncols = 256; ncols > 0; ncols--) {
      if (used[ncols-1] != 0) break;
    }
    imp->ncolors = ncols;
    
  } else {
    DEBUG (("cannot seek to beginning of image data"));
    return 0;
  }
  return 1;
}


/*
 * Search color table of an image for best matching color.
 */
struct color*
find_color(
    struct image *imp, int r, int g, int b)
{
  int i;
  int rd, gd, bd;                              /* difference of r,g,b */
  int best = -1;                               /* idx with smallest dist */
  int mindist = 0x7fffffff;

  /* check parameters */
  assert (imp != NULL);
  assert (imp->ncolors > 0);
  assert (0 <= r  &&  r <= 0xff);
  assert (0 <= g  &&  g <= 0xff);
  assert (0 <= b  &&  b <= 0xff);

  /* find best matching color */
  for (i = 0; i < imp->ncolors; i++) {
    int d;
    
    /* compute distance to desired color */
    rd = imp->palette[i].r - r;
    gd = imp->palette[i].g - g;
    bd = imp->palette[i].b - b;
    d = 3*rd*rd + 3*gd*gd + 2*bd*bd;

    /* remember color with shortest distance */
    if (d < mindist) {
      best = i;
      mindist = d;
    }
  }
  assert (0 <= best  &&  best < imp->ncolors);
  return &imp->palette[best];
}


/*
 * Read header of a pcx image file and prepare to read some data from the
 * file.  Note that the number of colors cannot be reliably computed at this
 * point as pcx file format extension always creates a palette of 256 colors
 * despite of the actual number of the colors.  The true amount of colors
 * can only be computed by analysing the image data using a histogram.  So,
 * the palette always contains 256 colors after initialization, and the
 * number of colors is decreased after loading the image.
 */
static int
init_pcx(
    struct image *imp)
{
  struct pcx_header h;
  assert (imp != NULL);
  assert (imp->FP != NULL);

  /* read header from file */
  if (fread (&h, sizeof (h), 1, imp->FP) == 1) {
    
    /* is the file supported? */
    if (h.manufacturer != 0x0a) {
      DEBUG(("not a ZSoft PCX file"));
      return 0;
    }
    if (h.version < 5) {
      DEBUG(("obsolete file format"));
      return 0;
    }
    if (h.nbits != 8) {
      DEBUG(("color planes not supported (%d)", h.nbits));
      return 0;
    }
    if (h.encoding != 1) {
      DEBUG(("unsupported encoding scheme"));
      return 0;
    }
    if (h.palettetype[1] != 0) {
      DEBUG(("unsupported color presentation (%d)", h.palettetype[1]));
      return 0;
    }
    
    /* locate palette at the end of the file */
    if (fseek (imp->FP, -(256*3+1), SEEK_END) == /*OK*/0) {
      
      /* is the palette there? */
      if (fgetc (imp->FP) == 12) {
        /* palette found */
        int i, r, g, b, a;
        int rowlen;
        
        /* read palette at the end of the file */
        for (i = 0; i < 256; i++) {
          /* read single color */
          r = fgetc (imp->FP);
          g = fgetc (imp->FP);
          b = fgetc (imp->FP);
          a = 0xff;
          
          /* fix alpha component for transparent pixels */
#ifdef TRANSPARENT_PIXEL
          if (((r<<16) | (g<<8) | b)  == TRANSPARENT_PIXEL) {
            /* fully transparent color */
            a = 0x00;
          }
#endif

#ifdef SHADOW_PIXEL
          if (((r<<16) | (g<<8) | b)  == SHADOW_PIXEL) {
            /* half transparent color */
            r = 0x00;
            g = 0x00;
            b = 0x00;
            a = 0x80;
          }
#endif
          
          /* save color to lookup table */
          imp->palette[i].r = r;
          imp->palette[i].g = g;
          imp->palette[i].b = b;
          imp->palette[i].a = a;
        }
        
        /* check for I/O errors */
        if (ferror (imp->FP) != /*OK*/0) {
          DEBUG (("I/O error while reading palette"));
          return 0;
        }
        
        /* initialize image structure */
        imp->ncolors= i;
        imp->width  = (h.xmax[1]<<8)+h.xmax[0]-((h.xmin[1]<<8)+h.xmin[0])+1;
        imp->height = (h.ymax[1]<<8)+h.ymax[0]-((h.ymin[1]<<8)+h.ymin[0])+1;

        /*
         * We do not survive extra bytes in scan lines so quit now if there
         * are these filler bytes present.
         */
        rowlen = (h.rowlen[1] << 8) + h.rowlen[0];
        if (rowlen != ((imp->width+1) & ~1)) {
          DEBUG (("odd-sized image (%d)", rowlen));
          return 0;
        }
        
      } else {
        /* wrong magic number at the beginning of the color table */
        DEBUG (("missing color table"));
        return 0;
      }      
    } else {
      /* cannot seek to beginning of palette */
      DEBUG (("cannot seek"));
      return 0;
    }    
  } else {
    /* could not read header */
    DEBUG(("unexpected EOF"));
    return 0;
  }

  /* initialization succeeded */
  return 1;
}
