/*
 * input.c - input devices
 * 
 * Copyright (C) 2005  Toni Ronkko
 * 
 * Peach Gardens is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * Peach Gardens is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with Peach Gardens; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * 
 * Nov 20 2004, Toni Ronkko <toni at users.playstation2-linux.com>
 *
 * $Id: input.c,v 1.2 2005/04/03 16:22:58 tronkko Exp $
 *
 * $Log: input.c,v $
 * Revision 1.2  2005/04/03 16:22:58  tronkko
 * produce less debug info
 *
 * Revision 1.1  2005/01/09 21:06:34  tronkko
 * Initial revision
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "config.h"
#include "input.h"
#include "ticks.h"
#include "debug.h"


/*
 * Information on an event.  The structure is used for maintaining button
 * status and for creating synthetic events.
 */
struct event_info {
  int pressed;                                 /* is pressed */
  int repeats;                                 /* number of repeats */
  ticks_t time;                                /* when was triggered */
  int repeating;                               /* is auto-repeating */
};


/* previous 8 events */
static unsigned int combos[MAX_PORTS];

/* button status */
static struct event_info events[MAX_PORTS][MAX_EVENTS];

/* read-ahead event */
static int next_event = NOEVENT;

/* module initialized already? */
static int initialized = 0;


/*
 * Get button or keyboard event from an input device or generate a synthetic
 * event.  The function creates synthetic press events for buttons that are
 * marked as auto-repeating and synthetic release events for directional
 * buttons.  For example, left and right directional buttons cannot be held
 * down at the same time, but input handlers such as js.c cannot see the
 * release events when the direction is changed.  Hence, we must prepare
 * some synthetic events to fix the problem.
 *
 * To use this very function properly it is important to keep reading events
 * as long as the function returns events.  Especially in multi-player games
 * several events can be triggered in a frame, and a program that only reads
 * one event per frame cannot stay responsive under intense action.
 */
int
get_event (void)
{
  int ev, port, i;
  assert (initialized);
  
get_another:

  /* get event from a source */
  if (next_event == NOEVENT) {
    /* read event from joystick or keyboard */
    ev = joy_event();
    if (ev == NOEVENT) ev = key_event();    
  } else {
    /* get read-ahead event */
    ev = next_event;
    next_event = NOEVENT;
  }

  /* update pressed/released status */
  if (ev != NOEVENT) {
    /* which port */
    port = (ev & PORTMASK) >> PORTOFFSET;
    assert (0 <= port  &&  port < MAX_PORTS);
    
    /* get index to event table */
    i = EVENTID(ev);
    assert (0 <= i  &&  i < MAX_EVENTS);

    switch (ev & EVENTMASK) {

    /* movement along x axis */
    case RIGHT:
    case LEFT: {
      int j;
      assert (ev != NOEVENT);
      
      /* get the other button */
      if ((ev & EVENTMASK) == RIGHT) j = EVENTID(LEFT);
      else j = EVENTID(RIGHT);

      /* update status */
      if ((ev & PRESSMASK) == PRESS) {
        /*
         * Received an event telling about the press of a directional
         * button.  Regular joystick handlers such as js.c do not generate
         * release events when changing direction so we need to generate
         * synthetic release events to maintain button  press/release
         * status. 
         */
        if (!events[port][j].pressed) {
          /* the other direction is not pressed */
          if (events[port][i].pressed == 0) {
            events[port][i].pressed = 1;
            events[port][i].time    = ticks();
            events[port][i].repeats = 0;
          } else {
            /* excess ress event */
            goto get_another;
          }
        } else {
          /* save press event for next time */
          next_event = ev;
          
          /* generate synthetic release event for other button */
          ev = j | (port << PORTOFFSET) | RELEASE;
          events[port][j].pressed = 0;
        }
      } else {
        /*
         * Release left/right button.  Regular joystick drivers only
         * generate axis movement events without generating release events
         * for left/right button separately.  Thus, we generate synthetic
         * release events despite of the direction.  The input handlers such
         * as js.c may generate release event for right button even though
         * left button was pressed.
         */
        if (events[port][EVENTID(RIGHT)].pressed) {
          ev = RIGHT | (port << PORTOFFSET) | RELEASE;
          events[port][EVENTID(RIGHT)].pressed = 0;
        } else if (events[port][EVENTID(LEFT)].pressed) {
          ev = LEFT | (port << PORTOFFSET) | RELEASE;
          events[port][EVENTID(LEFT)].pressed = 0;
        } else {
          /* excess release event */
          goto get_another;
        }
      }
      break;}

    /* movement in y-axis */
    case UP:
    case DOWN:  {
      int j;
      
      /* get the other button */
      if ((ev & EVENTMASK) == UP) j = EVENTID(DOWN);
      else j = EVENTID(UP);

      /* update status */
      if ((ev & PRESSMASK) == PRESS) {
        /* press */
        if (!events[port][j].pressed) {
          /* the other direction is not pressed */
          if (events[port][i].pressed == 0) {
            events[port][i].pressed = 1;
            events[port][i].time    = ticks();
            events[port][i].repeats = 0;
          } else {
            /* excess press event */
            goto get_another;
          }
        } else {
          /* save press event for next time */
          next_event = ev;
          
          /* generate synthetic release event for other button */
          ev = j | (port << PORTOFFSET) | RELEASE;
          events[port][j].pressed = 0;
        }
      } else {
        /* release */
        if (events[port][EVENTID(UP)].pressed) {
          ev = UP | (port << PORTOFFSET) | RELEASE;
          events[port][EVENTID(UP)].pressed = 0;
        } else if (events[port][EVENTID(DOWN)].pressed) {
          ev = DOWN | (port << PORTOFFSET) | RELEASE;
          events[port][EVENTID(DOWN)].pressed = 0;
        } else {
          /* excess release event */
          goto get_another;
        }
      }
      break;}
      
    /* regular on/off event generators */
    case MENU:
    case ACTION:
    case EXIT:
    case PAUSE:
    case ERASE:
      /* update pressed status */
      if ((ev & PRESSMASK) == PRESS) {
        /* button/key pressed */
        if (events[port][i].pressed == 0) {
          events[port][i].pressed = 1;
          events[port][i].time    = ticks();
          events[port][i].repeats = 0;
        } else {
          /* excess press event */
          goto get_another;
        }
      } else {
        if (events[port][i].pressed) {
          /* released */
          events[port][i].pressed = 0;
        } else {
          /*
           * Event indicates that the button was released, but the event
           * table tells that the button was not depressed at all.  Perhaps
           * the very same event has been triggered both from keyboard and
           * joystick, for example, and the the release event was already
           * registered from other device.  Anyway, sending two release
           * events for a button/key may confuse the program logic so act as
           * if the event had never happened.
           */
          goto get_another;
        }
      }
      break;

    default:
      DEBUG (("unknown event: %d", ev));
      goto get_another;
    }
    assert (ev != NOEVENT);

    /* update combos */
    if ((ev & PRESSMASK) == PRESS) {
      combos[port] = combos[port] << 4 | EVENTID(ev);
    }
  }
  
  /* auto-repeat */
  if (ev == NOEVENT) {
    for (port = 0; port < MAX_PORTS  &&  ev == NOEVENT; port++) {
      for (i = 0; i < MAX_EVENTS; i++) {
        /* is the event an auto-repeating one? */
        if (events[port][i].repeating  &&  events[port][i].pressed) {
          ticks_t timeout;
          ticks_t now;

          /* get milliseconds to wait before creating a synthetic event */
          if (events[port][i].repeats == 0) {
            /* the event has not been repeated so far */
            timeout = TIMEOUT_FIRST;
          } else {
            /* the event has been repeated n times */
            timeout = TIMEOUT_SECOND;
          }

          /* create synthetic event */
          now = ticks ();
          if (now - events[port][i].time > timeout) {
            events[port][i].repeats++;
            events[port][i].time = now;
            ev = i | (port << PORTOFFSET) | PRESS;
            break;
          }
        }
      }
    }
  }
  return ev;
}


/*
 * See if a button or a key pressed at the moment.  Note that even though
 * you can check button status by this very function, you still need to read
 * events.  The function pressed() does not update its internal records
 * unless events are read properly.
 *
 * Also note that directional buttons left and right or up and down cannot
 * be pressed at the same time.
 */
int
pressed(
    int ev)
{
  int i;
  int port;
  assert (initialized);
  assert (ev != NOEVENT);

  /* which port */
  port = (ev & PORTMASK) >> PORTOFFSET;
  assert (0 <= port  &&  port < MAX_PORTS);

  /* get index to event table */
  i = EVENTID(ev);
  assert (0 <= i  &&  i < MAX_EVENTS);

  /* return cached information from the event table */
  return events[port][i].pressed;
}


/*
 * Enable auto-repeat of a button.  That is, make an event repeat
 * automatically when the equivalent key/button is pressed long enough.
 */
void
set_autorepeat(
    int ev, int flag)
{
  int i, port;
  assert (initialized);
  assert (ev != NOEVENT);
  assert (flag == 0  ||  flag == 1);

  /* get index to event table */
  i = EVENTID(ev);
  assert (0 <= i  &&  i < MAX_EVENTS);
  
  /* remember which events to repeat automatically */
  for (port = 0; port < MAX_PORTS; port++) {
    events[port][i].repeating = flag;
  }
}


/*
 * Initialize input interfaces.
 */
void
init_input (void)
{
  assert (initialized == 0);

  /* reset event states */
  memset (events, 0, sizeof (events));
  memset (combos, 0, sizeof (combos));
  next_event = NOEVENT;

  /* generic input part now initialized */
  initialized = 1;
      
  /* initialize interfaces */
  init_key ();
  init_joy ();

  /* initial set of autorepeated events */
  set_autorepeat (LEFT,  1);
  set_autorepeat (RIGHT, 1);
  set_autorepeat (UP,    1);
  set_autorepeat (DOWN,  1);
  set_autorepeat (ACTION,1);
  set_autorepeat (ERASE, 1);
}


/*
 * Release resources held by the input readers.
 */
void
done_input (void)
{
  assert (initialized != 0);

  /* de-initialize interfaces */
  done_joy ();
  done_key ();

  initialized = 0;
}
