/*
 * js.c - input module for linux joystick
 * 
 * Copyright (C) 2005  Toni Ronkko
 * 
 * Peach Gardens is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * Peach Gardens is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with Peach Gardens; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * 
 * Nov 20 2004, Toni Ronkko <toni at users.playstation2-linux.com>
 *
 * $Id: js.c,v 1.2 2005/04/03 16:24:17 tronkko Exp $
 *
 * $Log: js.c,v $
 * Revision 1.2  2005/04/03 16:24:17  tronkko
 * specify joystick device, button definition stored in array
 *
 * Revision 1.1  2005/01/09 21:06:37  tronkko
 * Initial revision
 *
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "config.h"
#include "input.h"
#include "debug.h"

#ifdef HAVE_LINUX_JOYSTICK_H
#include <linux/joystick.h>

#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif

#ifdef HAVE_FCNTL_H
# include <fcntl.h>
#endif

#if HAVE_SYS_IOCTL_H
# include <sys/ioctl.h>
#endif


/* max length of device path */
#define MAX_DEVICE_NAME 64

/* max number of joystick buttons */
#define MAX_BUTTONS 16

/* default joystick device */
static char joydev[MAX_DEVICE_NAME] = "";

/* scan joystick ports? */
static int scan_ports = 1;

/* port handles */
static int ports[MAX_PORTS];

/* number of open ports */
static int nports = 0;

/* default buttons assignments */
static int buttons[MAX_BUTTONS] =
{
#ifdef PS2DEV
  /* Playstation 2 DualShock */
  0,                                           /* button 0 */
  ACTION,                                      /* cross */
  EXIT,                                        /* triangle */
  ERASE,                                       /* circle */
  0,                                           /* button 4 */
  0,                                           /* button 5 */
  0,                                           /* button 6 */
  0,                                           /* button 7 */
  MENU,                                        /* select */
  PAUSE,                                       /* start */
  0,                                           /* button 10 */
  0,                                           /* button 11 */
  0,                                           /* button 12 */
  0,                                           /* button 13 */
  0,                                           /* button 14 */
  0                                            /* button 15 */
#else
  /* Logitech DualAction Pad */
  0,                                           /* button 0 */
  ACTION,                                      /* button 1 */
  ERASE,                                       /* button 2 */
  EXIT,                                        /* button 3 */
  0,                                           /* button 4 */
  0,                                           /* button 5 */
  0,                                           /* button 6 */
  0,                                           /* button 7 */
  MENU,                                        /* button 8 */
  PAUSE,                                       /* button 9 */
  0,                                           /* button 10 */
  0,                                           /* button 11 */
  0,                                           /* button 12 */
  0,                                           /* button 13 */
  0,                                           /* button 14 */
  0                                            /* button 15 */
#endif
};

/* has the module been initialized? */
static int initialized = 0;


/*
 * Get event from linux joystick device.  The joystick device typically
 * holds at most 64 events in a circular list.  If the events are not
 * intercepted in a timely manner, then some events will be lost and the
 * state of buttons in input.c will become flawed.
 *
 * Button mappings for Playstation 2 Linux Kit with DualShock-2 controller:
 *
 * Event      Button
 * LEFT       Analog left
 * RIGHT      Analog right
 * UP         Analog up
 * DOWN       Analog down
 * ACTION     Cross
 * EXIT       Triangle
 * ERASE      Circle
 * MENU       Select
 * PAUSE      Start
 * 
 * Button mappings for Linux joystick:
 * 
 * Event      Button
 * LEFT       Axis 1
 * RIGHT      Axis 1
 * UP         Axis 2
 * DOWN       Axis 2
 * ACTION     Button 1
 * EXIT       Button 3
 * ERASE      Button 2
 * MENU       Button 4
 * PAUSE      Unassigned
 */
int
joy_event (void)
{
  int i;
  const int threshold = 0x4000;
  assert (initialized);

  /* test all ports */
  for (i = 0; i < nports; i++) {
    if (ports[i] != -1) {
      struct js_event e;

reread:
      
      /* read event from device */
      if (read (ports[i], &e, sizeof(struct js_event)) > 0) {
        /* translate raw event code to event defined in input.h */
        if (e.type == JS_EVENT_BUTTON) {

          /* button pressed/released */
          if (e.number < MAX_BUTTONS  &&  buttons[e.number] != 0) {
            return buttons[e.number]
                | (i << PORTOFFSET)
                | (e.value ? PRESS : RELEASE);;
          }
          /* press of unassigned button */
          goto reread;
          
        } else if (e.type == JS_EVENT_AXIS) {

          /*
           * Joystick moved.  Note that we do not care to generate release
           * events for previously returned directional buttons here.  It is
           * left for the input.c module that records the button statuses.
           */
          switch (e.number) {
            
          /* x-axis */
          case 0:
            if (e.value < -threshold) {
              return LEFT | (i << PORTOFFSET) | PRESS;
            } else  if (e.value > threshold) {
              return RIGHT | (i << PORTOFFSET) | PRESS;
            } else {
              return LEFT | (i << PORTOFFSET) | RELEASE;
            }
            break;

          /* y-axis */
          case 1:
            if (e.value < -threshold) {
              return UP | (i << PORTOFFSET) | PRESS;
            } else  if (e.value > threshold) {
              return DOWN | (i << PORTOFFSET) | PRESS;
            } else {
              return UP | (i << PORTOFFSET) | RELEASE;
            }
            break;

          default:
            DEBUG (("ignoring axis %d, value %d", e.number, e.value));
            goto reread;
          }
        }
      }
    } /* if port open */
  }
  return NOEVENT;
}


/*
 * Open linux joystick driver.  The function will try to open at most
 * MAX_PORTS joystick devices, but does not fail if no joysticks are
 * detected.
 */
void
init_joy (void)
{
  int i;
  int max;

  /* reset file descriptors */
  for (i = 0; i < MAX_PORTS; i++) {
    ports[i] = -1;
  }

  /* maximum number of ports to try */
  if (scan_ports) {
    /* default device */
    max = MAX_PORTS;
  } else {
    /* use my device */
    max = 1;
  }
    
  /* open ports */
  for (i = 0; i < MAX_PORTS; i++) {
    char devnam[MAX_DEVICE_NAME];
    char *p;
    
    /* prepare device name */
    if (scan_ports) {
      /* scan default ports */
      strncpy (devnam, "/dev/js", sizeof (devnam));
      devnam[MAX_DEVICE_NAME-1] = '\0';
      p = strchr (devnam, '\0');
      
      /* append device number */
      assert (MAX_PORTS < 10);
      *p++ = '0' + i;
      *p = '\0';
      assert (p - devnam < MAX_DEVICE_NAME);
    } else {
      /* use my device */
      strncpy (devnam, joydev, MAX_DEVICE_NAME);
      devnam[MAX_DEVICE_NAME - 1] = '\0';
    }
    
    /* try to open device in non-blocking mode */
    ports[i] = open (devnam, O_RDONLY | O_NONBLOCK);
    if (ports[i] != -1) {

      /* print info on joystick */
#if 0
      char name[128];
      int version;
      
      /* read name of joystick */
      if (ioctl(ports[i], JSIOCGNAME(sizeof(name)), name) < 0) {
        strncpy(name, "Unknown", sizeof(name));
      }

      /* interface version */
      if (ioctl(ports[i], JSIOCGVERSION, &version) < 0) {
        version = 0;
      }
      
      DEBUG(("%s: %s, version %x", devnam, name, version));
#endif

    } else {
      /* failed to open port */
      break;
    }
  }  
  if (i == 0) {
    DEBUG (("no joysticks found"));
  }
  
  /* remember number of open ports */
  nports = i;

  /* module has now been initialized */
  initialized = 1;
}


/*
 * Close joystick interface.
 */
void
done_joy (void)
{
  int i;
  assert (initialized);

  /* close all devices */
  for (i = 0; i < MAX_PORTS; i++) {
    if (ports[i] != 0) {
      close (ports[i]);
      ports[i] = 0;
    }
  }

  initialized = 0;
}


/*
 * Specify path of joystick device, if other than /dev/js0.  Note that the
 * function is intended to be called before initializing this very module.
 */
void
joy_device(
    const char *dev)
{
  assert (!initialized);
  assert (strlen (dev) < MAX_DEVICE_NAME);

  /* copy device name */
  strncpy (joydev, dev, MAX_DEVICE_NAME);
  joydev[MAX_DEVICE_NAME - 1] = '\0';
}

#else /* no joystick */

/* dummy event functions to be used when there is no joystick interface */
int joy_event (void) {
  return NOEVENT;
}

void init_joy (void) {
  DEBUG(("dummy joystick interface"));
}
  
void done_joy (void) {
  /*NOP*/
}

void joy_device (const char *dev) {
  DEBUG(("ignoring joystick device %s", dev));
}

#endif /* no joystick*/
