/*
 * locate.c - locate data files
 * 
 * Copyright (C) 2005  Toni Ronkko
 * 
 * Peach Gardens is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * Peach Gardens is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with Peach Gardens; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * 
 * Nov 2004, Toni Ronkko <toni at users.playstation2-linux.com>
 *
 * $Id: locate.c,v 1.3 2005/04/03 16:31:09 tronkko Exp $
 *
 * $Log: locate.c,v $
 * Revision 1.3  2005/04/03 16:31:09  tronkko
 * not needing feat.h anymore
 *
 * Revision 1.2  2005/01/09 21:12:21  tronkko
 * look closer from source tree
 *
 * Revision 1.1  2004/11/20 10:59:21  tronkko
 * Initial revision
 *
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <limits.h>

#include "config.h"
#include "locate.h"
#include "debug.h"

/* for access */
#if HAVE_UNISTD_H
# include <unistd.h>
#endif

/* length of longest path */
#ifndef PATH_MAX
# define PATH_MAX 1024
#endif


/* convert macro arg to string */
#define XSTRINGIZE(x) #x
#define STRINGIZE(x) XSTRINGIZE(x)

/* look for file in given directory */
static int check_dir (const char *dir, const char *fn);

/* temporary room for returned file name */
static char path[PATH_MAX];


/*
 * Locate file fn and return full path to it.  The function searches the
 * file from (1) directory "data", (2) installation directory specified by
 * compilation time, (3) directory specified by PGDATA variable,
 * (4) directory /usr/local/share/pgardens.  As a last resort the function
 * returns the name of the data file as is.
 *
 * The data directory is searched first to ease debugging, and to allow
 * users to play the game without installing it first.  However, as this
 * procedure may interfere the game if the user invokes the game in an other
 * directory also containing "data" subdirectory, the directory is used if
 * only there is a file called "pgdemo" or "pgdemo.c" in the current
 * directory.
 *
 * The file name returned by the function is allocated from a static memory
 * area.  The successive use of locate() will overwrite the file name
 * returned by the previous call.
 *
 * The function returns a pointer to the input file name in case of
 * failure.  Typically you do not need to check for failure.  Just try to
 * open the fail and let the fopen do the checking.  If the data file
 * happens to reside at current working directory, then this final scheme
 * can even work.
 */
const char*
locate(
    const char *fn)
{
  int ok = 0;

  /* look from installation directory */
#ifdef DATADIR
  if (!ok) {
    ok = check_dir (STRINGIZE(DATADIR), fn);
  }
#endif

  /* look from PGDATA variable */
  if (!ok) {
    char *root = getenv (PGDATA_VAR);
    if (root != NULL) {
      /* variable is defined */
      ok = check_dir (root, fn);
    }
  }

  /* check source tree */
  if (!ok) {
    ok = check_dir ("../data", fn);
  }
  if (!ok) {
    ok = check_dir ("data", fn);
  }
  if (!ok) {
    ok = check_dir ("pgdemo", fn);
  }
  
  /* look from fixed paths */
  if (!ok) {
    ok = check_dir ("/usr/share/pgardens", fn);
  }

  /* return file name */
  if (ok) {
    return path;
  } else {
    /* not found */
    return fn;
  }
}


/*
 * Look for file fn from directory dir.  The function formats full path to
 * the file name to variable path, and checks if the file exists.  Returns
 * true if the file was found.
 */
static int
check_dir(
    const char *dir, const char *fn)
{
  int dirlen, fnlen;
  assert (dir != NULL);
  assert (fn != NULL  &&  *fn != '\0');
  
  /* is file name absolute? */
  if (*fn != '/') {
    /* relative file name */
    dirlen = strlen (dir);
    fnlen  = strlen (fn);

    /* is there enough room in the buffer? */
    if (dirlen + fnlen + strlen ("/") + 1 <= sizeof (path)) {
      char *p;

      /* copy directory name to path and remove excess slashes */
      strncpy (path, dir, dirlen);
      p = path + dirlen;
      while (p > path  &&  p[-1] == '/') {
        p--;
      }
      *p++ = '/';

      /* append file name */
      strncpy (p, fn, fnlen);
      p += fnlen;
      *p = '\0';
      assert (p - path < (int) sizeof (path));
      
      /* is the file there? */
      return access (path, R_OK) == /*OK*/0;
      
    } else {
      DEBUG (("buffer too small"));
      return 0;
    }    
  } else {
    /* file name is absolute so we should not perform directory search */
    return 0;
  }
}
