/*
 * menu.c - simple menu system
 * 
 * Copyright (C) 2005  Toni Ronkko
 * 
 * Peach Gardens is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * Peach Gardens is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with Peach Gardens; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * 
 * Mar  9 2005, Toni Ronkko <toni at users.playstation2-linux.com>
 *
 * $Id: menu.c,v 1.1 2005/04/03 16:31:58 tronkko Exp $
 *
 * $Log: menu.c,v $
 * Revision 1.1  2005/04/03 16:31:58  tronkko
 * Initial revision
 *
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "menu.h"
#include "ticks.h"
#include "debug.h"


/*
 * Initialize menu from given set of options and set paint function.
 */
void
init_menu(
    struct menu *mp, const struct menu_option *op, int options,
    paint_option *paint)
{
  assert (mp != NULL);
  assert (op != NULL  &&  options > 0);
  assert (paint != NULL);

  /* initialize menu */
  memset (mp, 0, sizeof (struct menu));
  mp->paint = paint;
  mp->options = options;
  mp->blink_state = 1;
  mp->blink_interval = MENU_BLINK_NORMAL;
  mp->blink_time = ticks() + mp->blink_interval;
  
  /* select first option */
  mp->selected = 0;
  
  /* allocate memory for options */
  mp->op = malloc (sizeof (struct menu_option) * options);
  // DEBUG (("%p: alloc options", mp->op));
  if (mp->op != NULL) {
    /* duplicate options */
    memcpy (mp->op, op, sizeof (struct menu_option) * options);
  } else {
    out_of_memory();
  }
}


/*
 * De-initialize menu.
 */
void
done_menu(
    struct menu *mp)
{
  assert (mp != NULL);
  assert (mp->op != NULL);

  /* release menu options */
  // DEBUG (("%p: free options", mp->op));
  free (mp->op);

  /* reset structure to detect dangling pointers */
#ifndef NDEBUG
  memset (mp, 0xcc, sizeof (*mp));
#endif
}


/*
 * Check if an option has been accepted and application should exit from
 * menu loop.
 *
 * Example of a menu loop:
 *
 *   init_menu (&mymenu, opts, MENU_OPTIONS, &paintmenu);
 *   while (!menu_finished (&mymenu)) {
 *     // paint screen
 *     clear ();
 *     draw_menu (&mymenu);
 *     swap_buffers ();
 * 
 *     // read events, see menu_select() for an example
 *   }
 *   // get selection, see menu_selection() for an example
 *   done_menu (&mymenu);
 *
 */
int
menu_finished(
    struct menu *mp)
{
  assert (mp != NULL);
  return mp->finished;
}


/*
 * Paint menu options to screen with the help of a custom paint function.
 * Also sets finish flag after a timeout.  This is used to highlight
 * accepted option before exiting from menu loop.
 */
void
draw_menu(
    struct menu *mp)
{
  int i;
  ticks_t now;
  
  assert (mp != NULL);
  assert (mp->paint != NULL);
  assert (-1 <= mp->selected  &&  mp->selected < mp->options);
  assert (mp->blink_state == 0  ||  mp->blink_state == 1);

  /* blink selection */
  now = ticks ();
  if (mp->blink_time < now) {
    mp->blink_time = now + mp->blink_interval;
    mp->blink_state = 1 - mp->blink_state;
  }

  /* exit after timeout */
  if (mp->exit_time != 0  &&  mp->exit_time < now) {
    mp->finished = 1;
  }
  
  /* paint all options */
  for (i = 0; i < mp->options; i++) {
    int state = MENU_ACTIVE;

    /* is option currently selected? */
    if (mp->selected == i) {
      /* currently selected menu option */
      if (mp->blink_state) {
        state = MENU_BLINKON;
      } else {
        state = MENU_BLINKOFF;
      }
    } else {
      /* not a selected menu option */
      if (mp->selected >= 0  &&  mp->exit_time == 0) {
        state = MENU_ACTIVE;
      } else {
        /* exiting from menu */
        state = MENU_INACTIVE;
      }
    }

    /* paint options */
    mp->paint (&mp->op[i], state);
  }
}


/*
 * Get pointer to currently selected menu item or NULL if emnu was canceled
 * with abort_menu().
 *
 * BEWARE: calling done_menu() will destroy the menu option returned by the
 * function.  Never store the pointer returned by menu_selection() to be
 * used later, but copy the option identifier instead.
 *
 * Example:
 * 
 *   struct menu_option *op = menu_selection (&mymenu);
 *   if (op != NULL) {
 *     // print identifier of selected option
 *     printf ("select %d\n", op->id);
 *   } else {
 *     // no option selected
 *     printf ("abort\n");
 *   }  
 *
 */
struct menu_option *
menu_selection(
    struct menu *mp)
{
  assert (mp != NULL);

  if (0 <= mp->selected  &&  mp->selected < mp->options) {
    /* option selected */
    return &mp->op[mp->selected];
  } else {
    /* menu was canceled */
    return NULL;
  }
}


/*
 * Select option by identifier or move selection back or forward.  To move
 * selection back or forward, specify MENU_PREV or MENU_NEXT as an option. 
 * Note that selection cannot be changed after end_menu() or abort_menu()
 * has been called.  This behaviour lets us present a visual hint on
 * selected option without user being able to cancel or change selection
 * before falling out of menu loop.
 *
 * The function is often called from inside an event loop:
 *
 *   struct menu *mymenu;
 *   int ev;
 *   while ((ev = get_event()) != NOEVENT) {
 *     switch (ev) {
 *     case UP:
 *       menu_select (mymenu, MENU_PREV);
 *       break;
 *          
 *     case DOWN:
 *       menu_select (mymenu, MENU_NEXT);
 *       break;
 * 
 *     case ACTION:
 *       end_menu (mymenu);
 *       break;
 *       
 *     case EXIT:
 *       abort_menu (mymenu);
 *       break;
 *     }
 *   }
 * 
 */
void
menu_select(
    struct menu *mp, int option)
{
  int i;
  int found;
  assert (mp != NULL);
  assert (-1 <= mp->selected  &&  mp->selected < mp->options);

  /* only select options when menu is still active */
  if (mp->selected >= 0  &&  mp->exit_time == 0) {

    /* reset blinking */
    mp->blink_state = 1;
    mp->blink_time = ticks() + mp->blink_interval;

    /* change selection */
    switch (option) {
    /* select previous option */
    case MENU_PREV:
      if (mp->selected > 0) {
        mp->selected--;
      }
      break;
      
    /* select next option */
    case MENU_NEXT:
      if (mp->selected < mp->options - 1) {
        mp->selected++;
      }
      break;
      
    /* select by identifier */
    default:
      found = 0;
      for (i = 0; i < mp->options; i++) {
        if (mp->op[i].id == option) {
          found = 1;
          mp->selected = i;
          break;
        }
      }
      
      /* no such option */
      if (!found) {
        abort_msg ("missing option %d", option);
      }
    }
  } else {
    DEBUG (("ignoring excess menu events"));
  }
}


/*
 * Accept currently selected option and post exit.  After returning from
 * this very function menu selection cannot be changed and menu cannot be
 * aborted.  Moreover, the selected option will flash a few times to
 * highlight the selection.
 */
void
end_menu(
    struct menu *mp)
{
  assert (mp != NULL);
  assert (0 <= mp->selected  &&  mp->selected < mp->options);

  /* post exit if not done that before */
  if (mp->selected >= 0  &&  mp->exit_time == 0) {
    ticks_t now = ticks ();
    mp->exit_time = now + MENU_CONFIRM_TICKS;
    mp->blink_interval = MENU_BLINK_CONFIRM;
    mp->blink_time = now + mp->blink_interval;
  } else {
    DEBUG (("already exiting menu"));
  }
}


/*
 * Exit menu without selecting an option.  After calling abort_menu(),
 * function menu_finished() returns true and menu_selection() retruns NULL
 * to signal that no selection was made.
 */
void
abort_menu(
    struct menu *mp)
{
  assert (mp != NULL);

  /* abort menu if not yet exiting */
  if (mp->selected >= 0  &&  mp->exit_time == 0) {
    /* reset selection */
    mp->selected = -1;
    
    /* exit from menu loop after a short while */
    mp->exit_time = ticks() + MENU_ABORT_TICKS;
  } else {
    DEBUG (("already exiting menu"));
  }
}
