/*
 * resource.c - global resources for Peach Gardens
 * 
 * Copyright (C) 2005  Toni Ronkko
 * 
 * Peach Gardens is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * Peach Gardens is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with Peach Gardens; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * 
 * Dec 28 2004, Toni Ronkko <toni at users.playstation2-linux.com>
 *
 * $Id: resource.c,v 1.3 2005/06/22 13:10:49 tronkko Exp $
 *
 * $Log: resource.c,v $
 * Revision 1.3  2005/06/22 13:10:49  tronkko
 * fonts are initialized individually
 *
 * Revision 1.2  2005/04/03 16:25:25  tronkko
 * home town
 *
 * Revision 1.1  2005/01/09 21:07:08  tronkko
 * Initial revision
 *
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "resource.h"
#include "locate.h"
#include "image.h"
#include "text.h"
#include "chip.h"
#include "chipset.h"
#include "map.h"
#include "basename.h"
#include "debug.h"


/* minimum and maximum supported map editor file format */
#define MIN_MAP_VERSION 2
#define MAX_MAP_VERSION 2

/* max length of header line in map editor file (not used for data lines) */
#define MAX_HEADER 1024


/* workspace classification */
enum workspace_t {
  Map, Palette
};


/* read map file */
static void load_map(
    struct chipset *chp, const char *fn);

/* read chipset from map file */
static int read_chipset(
    struct chipset *chp, char *buffer, FILE *FP);

/* read workspace from map file and assign it to global resource */
static int read_workspace(
    struct chipset *chp, char *buffer, FILE *FP);


/* chipsets */
struct chipset *cavewall = NULL;
static struct chipset _cavewall;
struct chipset *town = NULL;
static struct chipset _town;

/* maps */
struct map *mine = NULL;
static struct map _mine;
struct map *hometown = NULL;
static struct map _hometown;

/* version of map file that is being read */
static int file_version = 0;

/* has the module been initialized? */
static int initialized = 0;


/*
 * Load global resources that are needed by a number of modules for the full
 * duration of the game.  The function does not return on failure.
 */
void
init_resource (void)
{
  assert (!initialized);

  /* load chipsets and maps */
  load_map (&_cavewall, locate("caves.map"));
  cavewall = &_cavewall;
  load_map (&_town, locate("towns.map"));
  town = &_town;
  
  /* ensure that we got all maps */
  assert (mine != NULL);
  
  /* module now initialized */
  initialized = 1;
}


/*
 * De-initialize global resources.
 */
void
done_resource (void)
{
  assert (initialized);

  /* release maps */
  assert (mine == &_mine);
  done_map (&_mine);
  mine = NULL;
  
  assert (hometown == &_hometown);
  done_map (&_hometown);
  hometown = NULL;
  
  /* release chipsets */
  assert (cavewall == &_cavewall);
  done_chipset (&_cavewall);
  cavewall = NULL;
  
  assert (town == &_town);
  done_chipset (&_town);
  town = NULL;
  
  /* module succesfully de-initialized */
  initialized = 0;
}


/*
 * Load named map editor file, and assign maps to global structures.
 */ 
static void
load_map(
    struct chipset *chp, const char *fn)
{
  FILE *FP;
  
  /* open file */
  FP = fopen (fn, "r");
  if (FP) {
    char buffer[MAX_HEADER];
    int version;
    file_version = 0;
    
    /* get file version */
    if (fgets (buffer, sizeof(buffer), FP) != NULL) {
      if (sscanf (buffer, "!PGardens %d", &version) == 1) {
        if (MIN_MAP_VERSION <= version  &&  version <= MAX_MAP_VERSION) {
          file_version = version;
        } else {
          abort_msg ("unsupported file format on %s", fn);
        }
      } else {
        /* incorrect header */
        DEBUG (("unexpected header: %s", buffer));
        abort_msg ("file format error on %s", fn);
      }
    } else {
      /* empty file */
      abort_msg ("unexpected end of file while reading %s", fn);
    }

    /* read chipset */
    if (fgets (buffer, sizeof(buffer), FP) != NULL) {
      if (strncmp (buffer, "!chipset", strlen ("!chipset")) == 0) {
        
        /* chipset */
        if (!read_chipset (chp, buffer, FP)) {
          abort_msg ("cannot read chipset from %s", fn);
        }
      } else {
        /* incorrect header */
        DEBUG (("unexpected header: %s", buffer));
        abort_msg ("file format error on %s", fn);
      }
    } else {
      /* empty file */
      abort_msg ("unexpected end of file while reading %s", fn);
    }
      
    /* read workspaces */
    while (fgets (buffer, sizeof(buffer), FP) != NULL) {
      
      if (strncmp (buffer, "!work", strlen ("!work")) == 0) {
        
        /* palette or map */
        if (read_workspace (chp, buffer, FP)) {
          /*NOP*/;
        } else {
          abort_msg ("cannot read workspace from %s", fn);
        }
            
      } else {
        DEBUG (("unsupported section: %s", buffer));
        abort_msg ("cannot read file %s", fn);
      }
    }
      
    fclose (FP);
  } else {
    abort_msg ("cannot open file %s", fn);
  }
}
 

/*
 * Read !chipset tag out of map editor file.
 *
 * Variable buffer is assumed to contain the header line on function entry.
 * Moreover, buffer is assumed to hold at least MAX_HEADER bytes.
 */ 
static int 
read_chipset(
    struct chipset *chp, char *buffer, FILE *FP)
{
  int i;
  int nchips = 0;
  int chipw = 0;
  int chiph = 0;
  char *imgnam = NULL;
  char *p;

  /* extract !chipset from buffer */
  p = strtok (buffer, " \t");
  assert (p != NULL);
  
  /* get number of chips */
  p = strtok (NULL, " \t");
  if (p != NULL) {
    nchips = atoi (p);
  } else {
    DEBUG (("missing number of chips"));
    return 0;
  }
  
  /* chip width */
  p = strtok (NULL, " \t");
  if (p != NULL) {
    chipw = atoi (p);
  } else {
    DEBUG (("missing chip width"));
    return 0;
  }
  
  /* chip height */
  p = strtok (NULL, " \t");
  if (p != NULL) {
    chiph = atoi (p);
  } else {
    DEBUG (("missing chip height"));
    return 0;
  }
  
  /* chip image file */
  p = strtok (NULL, " \t");
  if (p != NULL) {
    /* strip away path */
    imgnam = basename (p);
    
    /* change file name suffix to .pcx */
    p = strrchr (imgnam, '.');
    if (p == NULL) {
      /* no suffix */
      p = strchr (imgnam, '\0');
    }
    assert (p + strlen (".pcx") + 1 <= &buffer[MAX_HEADER]);
    strcpy (p, ".pcx");
  } else {
    DEBUG (("missing chip image"));
    return 0;
  }

  /* check parameters just read */
  assert (nchips > 0);
  assert (chipw > 0);
  assert (chiph > 0);
  assert (imgnam != NULL);
  
  /* read chipset image into memory */
  init_chipset (chp, nchips, chipw, chiph, locate (imgnam));

  /* read individual chips */
  for (i = 0; i < nchips; i++) {
    int id, x, y, bg, z;
    char terrain[9];
    int fields;

    /* read data on one chip */
    fields = fscanf (FP, "%d %d %d %d %d %c %c %c %c %c %c %c %c %c\n",
        &id, &x, &y, &bg, &z,
        &terrain[0], &terrain[1], &terrain[2],
        &terrain[3], &terrain[4], &terrain[5],
        &terrain[6], &terrain[7], &terrain[8]);

    /* did we read it ok? */
    if (fields == 14) {
      int j;
      struct chip *cp;

      /* check data */
      assert (id == 0  ||  (0 <= id  &&  id < nchips));
      assert (id == 0  ||  (0 <= bg  &&  bg < nchips));
      assert (id == 0  ||  (0 <= x  &&  x*chipw+chipw <= chp->img.width));
      assert (id == 0  ||  (0 <= y  &&  y*chiph+chiph <= chp->img.height));

      /* get pointer to chip */
      cp = get_chip (chp, id);
      assert (cp != NULL);
      assert (cp->id == id);
      
      /* save chip info */
      cp->x  = x;
      cp->y  = y;
      cp->z  = z;
      cp->bg = bg;
      for (j = 0; j < 9; j++) {
        cp->terrain[j] = parse_terrain (terrain[j]);
      }
      
    } else {
      DEBUG (("expecting 14 fields on a chip, got %d", fields));
      fprintf (stderr, "file format error\n");
      return 0;
    }
  }

  /* all OK */
  return 1;
}
 

/*
 * Read workspace from map file.  If the workspace refers to a known map,
 * store the map to a global variable.  Otherwise, skip the workspace.
 *
 * Variable buffer is assumed to contain the header line on function entry.
 * Moreover, buffer is assumed to hold at least MAX_HEADER bytes.
 */
static int
read_workspace(
    struct chipset *chp, char *buffer, FILE *FP)
{
  int w = 0;
  int h = 0;
  int x, y;
  int i;
  char *name = NULL;
  char *p;
  enum workspace_t type = Palette;
  struct map *mp;
  
  /* extract !work from buffer */
  p = strtok (buffer, " \t");
  assert (p != NULL);
  
  /* get width of workspace */
  p = strtok (NULL, " \t");
  if (p != NULL) {
    w = atoi (p);
  } else {
    DEBUG (("missing workspace width"));
    return 0;
  }
  
  /* get height of workspace */
  p = strtok (NULL, " \t");
  if (p != NULL) {
    h = atoi (p);
  } else {
    DEBUG (("missing workspace height"));
    return 0;
  }
  
  /* get type of workspace */
  p = strtok (NULL, " \t");
  if (p != NULL) {
    if (strcmp (p, "map") == 0) {
       type = Map;
    } else if (strcmp (p, "palette") == 0) {
      /* ignore palettes */
      type = Palette;
    } else {
      DEBUG (("invalid workspace type: %s", p));
      return 0;
    }
  } else {
    DEBUG (("missing workspace type"));
    return 0;
  }

  /* get name of workspace */
  p = strtok (NULL, " \t");
  if (p != NULL) {
    name = p;

    /* remove line-feed at the end of buffer */
    p = strchr (p, '\n');
    if (p != NULL) {
      *p = '\0';
    }
  } else {
    DEBUG (("missing workspace name"));
    return 0;
  }

  /* check input */
  assert (w > 0);
  assert (h > 0);
  assert (name != NULL);

  /* is the map interesting */
  if (type == Map  &&  strcmp (name, "mine") == 0) {
    assert (mine == NULL);
    mine = &_mine;
    mp = mine;
  } else if (type == Map  &&  strcmp (name, "hometown") == 0) {
    assert (hometown == NULL);
    hometown = &_hometown;
    mp = hometown;
  } else {
    /* not interesting */
    mp = NULL;
  }

  /* initialize map */
  if (mp != NULL) {
    init_map (mp, w, h, chp);
  }
  
  /* read map blocks */
  x = 0;
  y = 0;
  for (i = 0; i < w*h; i++) {
    int id;
    if (fscanf (FP, "%d", &id) == 1) {
      assert (0 <= id  &&  id < chp->nchips);

      /* assign block if needed */
      if (mp != NULL) {
        /* store info on block */
        struct map_block *mbp = get_block (mp, x, y);
        mbp->id      = id;
        mbp->region  = 0;
        mbp->visible = 0;

        /* fix coordinate for next block */
        x++;
        if (x >= mp->w) {
          x = 0;
          y++;
        }
      }
      
    } else {
      fprintf (stderr, "error while reading %s\n", name);
      return 0;
    }
  }

  /* read to end of line */
  i = fgetc (FP);
  assert (i == '\n'  ||  i == ' '  ||  i == EOF);
  while (i != '\n'  &&  i != EOF) {
    i = fgetc (FP);
    assert (i == '\n'  ||  i == ' '  ||  i == EOF);
  }

  /* all OK */
  return 1;
}
