/*
 * script.c - conversation scripts
 * 
 * Copyright (C) 2005  Toni Ronkko
 * 
 * Peach Gardens is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * Peach Gardens is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with Peach Gardens; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * 
 * Apr  4 2005, Toni Ronkko <toni at users.playstation2-linux.com>
 *
 * $Id: script.c,v 1.1 2005/06/22 13:11:08 tronkko Exp $
 *
 * $Log: script.c,v $
 * Revision 1.1  2005/06/22 13:11:08  tronkko
 * Initial revision
 *
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>

#include "script.h"
#include "locate.h"
#include "debug.h"


/* maximum number of scenes in script file */
#define MAX_SCRIPTS 100

/* max length of phrase in conversation file */
#define BUFFER_LEN 1024

/* name of a file containing conversation scripts */
#define SCRIPT_FILE "script.txt"

/* token types for various read commands */
#define ERROR  256
#define NUM    257
#define END    258
#define GOTO   259
#define TEXT   260
#define YES    261
#define NO     262
#define IF     263
#define SCRIPT 264
#define PERSON 265


/* read multiple scripts from file */
static int read_file (void);

/* read any token */
static int read_token (void);

/* read number and sets variable number on success */
static int read_number (void);

/* read quoted text */
static int read_text (void);

/* read preserved word from file */
static int read_preserved (void);

/* ignore characters until newline */
static void skip_nl (void);

/* read and put back a character */
static int read_char (void);
static void unget_char (int c);

/* report general error at line */
static void parse_error (const char *msg);


/* person returned by read_preserved() */
static enum person_t person = Nobody;

/* string returned by read_text() */
static char buffer[BUFFER_LEN];

/* number read by read_number() */
static int number;

/* current file */
static FILE *FP;

/* line number in currect file */
static int lineno;

/* set to 0 if there were unrecoverable errors in script file */
static int conversation_ok;

/* pointers to scripts */
typedef struct script *script_ptr;
static script_ptr scripts[MAX_SCRIPTS];

/* has the module been initialized? */
static int initialized = 0;


/*
 * Insert say or ask phrase to script.  The type of phrase is detected on
 * the final letter of the message; if the message ends in question mark, it
 * is an Ask phrase.  Otherwise, it is a regular Say phrase.
 */
void
insert_say(
    struct script *sp, int curlabel, enum person_t who, const char *what)
{
  char *msg;
  struct phrase *pp;
  const char *p;
  enum phrase_t type = InvalidPhrase;
  assert (sp != NULL  &&  sp->no >= 0);

  /* detect message type by scanning message from back to front */
  p = strchr (what, '\0');
  while (what < p  &&  type == InvalidPhrase) {
    --p;

    /* inspect character */
    switch (*p) {
    /* ignore white space */
    case ' ': case '\t': case '\r': case '\n':
      break;

    /* question mark means Ask phrase */
    case '?':
      type = Ask;
      break;

    /* any other character marks Say phrase */
    default:
      type = Say;
    }
  }
  assert (type != InvalidPhrase);
  
  /* duplicate string */
  msg = strdup (what);
  DEBUG (("%p: alloc phrase text", msg));
  if (msg != NULL) {

    /* initialize phrase */
    pp = alloc_phrase ();
    pp->type = type;
    pp->label = curlabel;
    pp->who = who;    
    pp->what = msg;

    /* insert phrase to script */
    insert_phrase (sp, pp);
    
  } else {
    out_of_memory();
  }
}


/*
 * Append phrase to end of conversation script.
 */
void
insert_phrase(
    struct script *sp, struct phrase *pp)
{
  assert (sp != NULL  &&  sp->no >= 0);
  assert (pp != NULL);

  /* terminate list */
  pp->next = NULL;

  /* join pp to list of phrases */
  if (sp->last != NULL) {
    /* insert to end */
    sp->last->next = pp;
    sp->last = pp;
  } else {
    /* insert to beginning */
    sp->last = pp;
    sp->phrases = pp;
  }
}


/*
 * Store script sp to scripts array.
 */
void
store_script(
    struct script *sp)
{
  assert (sp != NULL);
  assert (0 < sp->no  &&  sp->no < MAX_SCRIPTS);
  assert (scripts[sp->no] == NULL);
  scripts[sp->no] = sp;
}


/*
 * Locate a script from pointer table by identifier.
 */
struct script *
find_script(
    int no)
{
  if (0 < no  &&  no < MAX_SCRIPTS) {
    assert (scripts[no] != NULL  &&  scripts[no]->no == no);
    return scripts[no];
  } else {
    abort_msg ("script id out of range");
  }
}


/*
 * Check script for validity.  That is, ensure that all jump labels
 * are defined.
 */
int 
check_script(
    struct script *sp)
{
  int ok = 1;
  struct phrase *pp;
  struct phrase *target;
  
  assert (sp != NULL  &&  sp->no >= 0);

  /* go through phrases */
  pp = sp->phrases;
  while (pp != NULL) {

    switch (pp->type) {      
    case IfYes: case IfNo:case Goto:
      /* target phrases of IfYes, IfNo and Goto must exist */
      target = find_phrase (sp, pp->target);
      if (target == NULL) {
        /* no label found */
        fprintf (stderr,
            "undefined reference to label %d in script %d\n",
            pp->target, sp->no);
        ok = 0;
      }
      break;

    case Ask: case Say:
      /* who and what must exist */
      assert (pp->who != Nobody);
      assert (pp->what != NULL);
      break;
      
    case Exit:
      assert (pp->who == Nobody);
      assert (pp->what == NULL);
      break;
      
    case InvalidPhrase:
    default:
      /* invalid phrase */
      abort ();
    }
    
    pp = pp->next;
  }

  /* do not allow game to begin with an error in conversation file */
  if (!ok) {
    conversation_ok = 0;
  }
  
  return ok;
}


/*
 * Allocate script from heap.  Does not return on error.
 */
struct script*
alloc_script(
    int no)
{
  struct script *sp;

  /* allocate memory */
  sp = malloc (sizeof *sp);
  DEBUG (("%p: alloc script %d", sp, no));
  if (sp != NULL) {
    /* initialize structure */
    sp->no = no;
    sp->phrases = NULL;
    sp->last = NULL;
  } else {
    out_of_memory();
  }
  return sp;
}


/*
 * Release script, and free all memory associated with it.
 */
void
free_script(
    struct script *sp)
{
  struct phrase *pp;

  assert (sp != NULL);
  assert (sp->no >= 0);

  /* release phrases */
  pp = sp->phrases;
  while (pp != NULL) {
    struct phrase *next = pp->next;
    free_phrase (pp);
    pp = next;
  }

  /* invalidate structure to spot errors */
#ifndef NDEBUG
  sp->no = -1;
  sp->phrases = (struct phrase*) ~0;  
  sp->last = (struct phrase*) ~0;  
#endif
  
  /* release script */
  DEBUG (("%p: free script", sp));
  free (sp);
}


/*
 * Find phrase by identifier.  Returns NULL if no such phrase is found from
 * given script.
 */
struct phrase *
find_phrase(
    struct script *sp, int target)
{
  struct phrase *pp;
  
  assert (sp != NULL  &&  sp->no != 0);
  assert (target >= 0);

  pp = sp->phrases;
  while (pp != NULL) {
    if (pp->label == target) {
      /* label found */
      return pp;
    }
    pp = pp->next;
  }

  /* label not found */
  DEBUG (("label %d not found in script %d", target, sp->no));
  return NULL;
}


/*
 * Allocate phrase structure from heap.  The function initializes the
 * structure with zero.  Does not return on failure.
 */
struct phrase*
alloc_phrase (void)
{
  struct phrase *pp;
  
  /* allocate memory for structure */
  pp = malloc (sizeof *pp);
  DEBUG (("%p: alloc phrase", pp));
  if (pp != NULL) {
    
    /* initialize structure */
    memset (pp, 0, sizeof *pp);
    
  } else {
    out_of_memory();
  }
  
  return pp;
}


/*
 * Remove piece of conversation from memory.
 */
void
free_phrase(
    struct phrase *pp)
{
  assert (pp != NULL);
  
  /* erase text */
  if (pp->what != NULL) {
    DEBUG (("%p: free phrase text", pp->what));
    free (pp->what);
  }
  
  /* fill structure with garbage to spot dangling pointers */
#ifndef NDEBUG
  memset (pp, 0xcc, sizeof *pp);
#endif
  
  /* erase structure from memory */
  DEBUG (("%p: free phrase", pp));
  free (pp);
}


/*
 * Read conversation scripts from file.  The file must have been opened
 * prior to calling read_file(), and module-local variable FP must have been
 * set accordingly.
 */
static int
read_file (void)
{
  int done = 0;
  int mode = 0;
  int token;
  int curlabel = -1;
  enum phrase_t type = InvalidPhrase;
  struct script *curscript = NULL;
  struct phrase *pp;

  /* get first token from file */
  token = read_token ();

  /* continue reading tokens until EOF */
  while (!done) {

    /* parse token according to current reading mode */
    switch (mode) {
    /* waiting script */
    case 0:
      switch (token) {
        
      /* start of a new script */
      case SCRIPT:       
        token = read_number ();
        if (token == NUM  &&  0 < number  &&  number < MAX_SCRIPTS) {
          if (scripts[number] == NULL) {
            /* script not initialized yet */
            struct script *sp = alloc_script (number);
            
            /* store current script somewhere */
            if (curscript != NULL) {
              check_script (curscript);
              store_script (curscript);
            }
            
            /* start using new script */
            curscript = sp;
            mode = 1;
          } else {
            /* duplicate script */
            parse_error ("duplicate script");
          }
          
        } else if (token != ERROR) {
          parse_error ("expecting script number");
        }
        break;

      case EOF:
        parse_error ("missing scripts");
        done = 1;
        break;
        
      case ERROR:
        done = 1;
        break;
        
      default:
        /* stop processing further lines */
        parse_error ("expecting script");
        done = 1;
      }
      break;

    /* reading phrases */
    case 1:
      switch (token) {
        
      /* label */
      case NUM:
        curlabel = number;
        mode = 2;
        break;

      /* phrase without label */
      case IF: case GOTO: case END: case PERSON:
        curlabel = -1;
        mode = 2;
        continue;

      /* end of script */
      case EOF:
        done = 1;
        break;        
        
      /* start of a new script */
      case SCRIPT:
        mode = 0;
        continue;
        
      case ERROR:
        break;
        
      default:
        parse_error ("syntax error");
      }
      break;

    /* phrase without label */
    case 2:
      switch (token) {
        
      case IF:
        mode = 3;
        break;
        
      case GOTO:
        mode = 5;
        break;
        
      case END:
        mode = 6;
        break;
        
      case PERSON:
        mode = 7;
        break;
        
      /* end of script */
      case EOF:
        parse_error ("unexpected end-of-file");
        done = 1;
        break;
        
      case ERROR:
        done = 1;
        break;
        
      default:
        parse_error ("syntax error");
        mode = 1;
      }
      break;

    /* if */
    case 3:
      switch (token) {
        
      case YES:
        type = IfYes;
        mode = 4;
        break;
        
      case NO:
        type = IfNo;
        mode = 4;
        break;
        
      /* end of script */
      case EOF:
        parse_error ("unexpected end-of-file");
        done = 1;
        break;
        
      case ERROR:
        done = 1;
        break;
        
      default:
        parse_error ("syntax error");
        mode = 1;
      }
      break;

    /* if yes/no */
    case 4:
      switch (token) {

      /* conditional jump to label */
      case NUM:
        assert (type == IfNo  ||  type == IfYes);
        assert (number >= 0);
        pp = alloc_phrase ();
        pp->type = type;
        pp->label = curlabel;
        pp->target = number;
        insert_phrase (curscript, pp);
        mode = 1;
        break;
        
      /* end of script */
      case EOF:
        parse_error ("unexpected end-of-file");
        done = 1;
        break;
        
      case ERROR:
        done = 1;
        break;
        
      default:
        parse_error ("syntax error");
        mode = 1;
      }
      break;

    /* goto */
    case 5:
      switch (token) {

      /* jump to label */
      case NUM:
        assert (number >= 0);
        pp = alloc_phrase ();
        pp->type = Goto;
        pp->label = curlabel;
        pp->target = number;
        insert_phrase (curscript, pp);
        mode = 1;
        break;
        
      /* end of script */
      case EOF:
        parse_error ("unexpected end-of-file");
        done = 1;
        break;
        
      case ERROR:
        done = 1;
        break;
        
      default:
        parse_error ("syntax error");
        mode = 1;
      }
      break;

    /* end */
    case 6:
      switch (token) {

      /* end expression */
      case NUM:
        pp = alloc_phrase ();
        pp->type = Exit;
        pp->label = curlabel;
        pp->target = number;
        insert_phrase (curscript, pp);
        mode = 1;
        break;
        
      /* end of script */
      case EOF:
        parse_error ("unexpected end-of-file");
        done = 1;
        break;
        
      case ERROR:
        done = 1;
        break;
        
      default:
        parse_error ("syntax error");
        mode = 1;
      }
      break;

    /* colon following name */
    case 7:
      switch (token) {
      case ':':
        mode = 8;
        break;

      /* end of script */
      case EOF:
        parse_error ("unexpected end-of-file");
        done = 1;
        break;
        
      case ERROR:
        done = 1;
        break;
        
      default:
        parse_error ("syntax error");
        mode = 1;
      }
      break;

    /* text phrase */
    case 8:
      switch (token) {

      /* establish say */
      case TEXT:
        insert_say (curscript, curlabel, person, buffer);
        mode = 1;
        break;
        
      /* end of script */
      case EOF:
        parse_error ("unexpected end-of-file");
        done = 1;
        break;
        
      case ERROR:
        done = 1;
        break;
        
      default:
        parse_error ("syntax error");
        mode = 1;
      }
      break;

    default:
      /* invalid mode */
      abort ();
    }

    /*
     * Read next token from file.  Note that this very statement is place in
     * purpose at the end of the loop, although at first sight it would be
     * better placed at the beginning of the loop.  The reason for this
     * behaviour is that by placing the token reader here it will be easy to
     * change modes and have a token re-processed.  Just change mode
     * variable and "continue" expression will re-evaluate current token.
     * There is no need to push back a token.
     */
    if (!done) {
      token = read_token ();
    }
  }

  /* store final script */
  if (curscript != NULL) {
    check_script (curscript);
    store_script (curscript);
  } else {
    /* no scripts defined in script file */
    fprintf (stderr, "missing scripts\n");
    conversation_ok = 0;
  }
  
  return conversation_ok;
}


/*
 * Read any token from conversation file.
 */
static int
read_token( void)
{
  do{
    /* read and parse one character */
    int c = read_char ();
    switch (c) {
    /* update current line number */
    case '\n':
      ++lineno;
      break;
      
    case ' ': case '\r': case '\t': 
      /* ignore white space */
      break;

    /* end of input */
    case EOF:
      return EOF;

    /* read number */
    case '0': case '1': case '2': case '3': case '4':
    case '5': case '6': case '7': case '8': case '9':
      unget_char (c);
      return read_number ();
      
    /* read someone's line */
    case '"':
      return read_text ();

    /* read character name or command */
    case 'a': case 'A':
    case 'b': case 'B':
    case 'c': case 'C':
    case 'd': case 'D':
    case 'e': case 'E':
    case 'f': case 'F':
    case 'g': case 'G':
    case 'h': case 'H':
    case 'i': case 'I':
    case 'j': case 'J':
    case 'k': case 'K':
    case 'l': case 'L':
    case 'm': case 'M':
    case 'n': case 'N':
    case 'o': case 'O':
    case 'p': case 'P':
    case 'q': case 'Q':
    case 'r': case 'R':
    case 's': case 'S':
    case 't': case 'T':
    case 'u': case 'U':
    case 'v': case 'V':
    case 'w': case 'W':
    case 'x': case 'X':
    case 'y': case 'Y':
    case 'z': case 'Z':
      unget_char (c);
      return read_preserved ();
      
    case ':':
      return c;

    /* ignore comments */
    case '#':
      skip_nl ();
      break;
    
    default:
      /* invalid token */
      parse_error ("invalid character");
      return ERROR;
    }
  } while (1);
  /*NOTREACHED*/
}


/*
 * Read quoted text.  When the function is entered, an opening double quote
 * has already been read away.
 */
static int
read_text (void)
{
  int c;
  int done = 0;
  char *p = buffer;
  char *end = p + sizeof (buffer);

  while (!done) {
    /* read one character to buffer */
    c = read_char ();
    switch (c) {
      
    /* not terminated */
    case EOF:
      parse_error ("unterminated character string");
      return ERROR;

    /* replace tabs and line feeds by spaces */
    case '\r': case '\n': case '\t':
      if (p != buffer  &&  p[-1] != ' ') {
        if (p < end) {
          *p++ = ' ';
        } else {
          parse_error ("buffer too small");
          return ERROR;
        }
      }
      break;

    /* ignore nulls */
    case '\0':
      break;

    /* end of string */
    case '"':
      done = 1;
      break;

    /* save other characters to buffer as is */
    default:
      if (p < end) {
        *p++ = c;
      } else {
        parse_error ("buffer too small");
        return ERROR;
      }
    }
  }
  
  /* terminate string */
  if (p < end) {
    *p = '\0';
  } else {
    parse_error ("buffer too small");
    return ERROR;
  }
  
  /* ensure that string is properly terminated */
  c = read_char ();
  if (c == EOF  ||  isspace(c)) {
    /* string ended properly */
    return TEXT;
  } else {
    /* number not terminated properly */
    parse_error ("invalid number");
    return ERROR;
  }
}


/*
 * Read positive number from file.  Returns NUM and sets module-local
 * variable number accordingly on success.  Otherwise, returns ERROR.
 */
static int
read_number (void)
{
  int i = 0;
  int c;
  
  /* read number */
  c = read_char ();
  while ('0' <= c  &&  c <= '9') {
    i = i * 10 + (c - '0');
    c = read_char ();
  }

  /* must be properly terminated */
  if (c == EOF  ||  isspace(c)) {
    /* number terminated by valid separator */
    number = i;
    return NUM;
  } else {
    /* number not terminated properly */
    unget_char (c);
    parse_error ("expecting number");
    return ERROR;
  }
}


/* shorthand for switching modes */
#define IFCHAR(what,goto) \
  if (c == what) { \
    mode = goto; \
  } else { \
    DEBUG(("syntax error in mode %d", mode)); \
    parse_error ("syntax error"); \
    return ERROR; \
  }

/* exit with error */
#define SYNTAX_ERROR() \
  DEBUG(("syntax error in mode %d", mode)); \
  parse_error ("syntax error"); \
  return ERROR;



/*
 * Read preserved word from file.  Returns IF, YES, NO, GOTO, END, SCRIPT,
 * ERROR or PERSON.  Sets variable person to name of person when PERSON is
 * returned.
 */
static int
read_preserved (void)
{
  int mode = 0;

  do{
    /* read low case character */
    int c = read_char ();
    if (isupper (c)) {
      c = tolower (c);
    }

    /* parse character according to mode */
    switch (mode) {
    case 0:
      switch (c) {
      case 'a':
        mode = 29;
        break;
        
      case 'e':
        mode = 12;
        break;
        
      case 'g':
        mode = 8;
        break;
        
      case 'i':
        mode = 22;
        break;
        
      case 'n':
        mode = 20;
        break;

      case 'r':
        mode = 33;
        break;
        
      case 's':
        mode = 1;
        break;
        
      case 'y':
        mode = 5;
        break;
        
      default:
        SYNTAX_ERROR();
      }
      break;

    /* read so far: s */
    case 1:
      if (c == 't') {
        mode = 2;
      } else if (c == 'c') {
        mode = 24;
      } else {
        SYNTAX_ERROR();
      }
      break;
      
    /* read so far: st */
    case 2:
      IFCHAR ('a', 3);
      break;
      
    /* read so far: sta */
    case 3:
      IFCHAR ('n', 4);
      break;
  
    /* read so far: stan */
    case 4:
      if (c == EOF  ||  (!isalpha(c)  &&  !isdigit(c))) {
        unget_char (c);
        person = Stan;
        return PERSON;
      } else {
        SYNTAX_ERROR();
      }
      break;
  
    /* read so far: y */
    case 5:
      if (c == 'o') {
        mode = 6;
      } else if (c == 'e') {
        mode = 18;
      } else {
        SYNTAX_ERROR();
      }
      break;
      
    /* read so far: yo */
    case 6:
      IFCHAR ('u', 7);
      break;
      
    /* read so far: you */
    case 7:
      if (c == EOF  ||  (!isalpha(c)  &&  !isdigit(c))) {
        unget_char (c);
        person = Stan;
        return PERSON;
      } else {
        SYNTAX_ERROR();
      }
      break;
  
    /* read so far: g */
    case 8:
      IFCHAR ('o', 9);
      break;
      
    /* read so far: go */
    case 9:
      IFCHAR ('t', 10);
      break;
      
    /* read so far: got */
    case 10:
      IFCHAR ('o', 11);
      break;
  
    /* read so far: goto */
    case 11:
      if (c == EOF  ||  (!isalpha(c)  &&  !isdigit(c))) {
        return GOTO;
      } else {
        SYNTAX_ERROR();
      }
      break;
  
    /* read so far: e */
    case 12:
      if (c == 'n') {
        mode = 13;
      } else if (c == 'x') {
        mode = 15;
      } else if (c == 'l') {
        mode = 41;
      } else {
        SYNTAX_ERROR();
      }
      break;
      
    /* read so far: en */
    case 13:
      IFCHAR ('d', 14);
      break;
      
    /* read so far: end */
    case 14:
      if (c == EOF  ||  (!isalpha(c)  &&  !isdigit(c))) {
        unget_char (c);
        return END;
      } else {
        SYNTAX_ERROR();
      }
      break;

    /* read so far: ex */
    case 15:
      IFCHAR ('i', 16);
      break;
      
    /* read so far: exi */
    case 16:
      IFCHAR ('t', 17);
      break;
      
    /* read so far: exit */
    case 17:
      if (c == EOF  ||  (!isalpha(c)  &&  !isdigit(c))) {
        unget_char (c);
        return END;
      } else {
        SYNTAX_ERROR();
      }
      break;

    /* read so far: ye */
    case 18:
      IFCHAR ('s', 19);
      break;
      
    /* read so far: yes */
    case 19:
      if (c == EOF  ||  (!isalpha(c)  &&  !isdigit(c))) {
        unget_char (c);
        return YES;
      } else {
        SYNTAX_ERROR();
      }
      break;

    /* read so far: n */
    case 20:
      IFCHAR ('o', 21);
      break;
      
    /* read so far: no */
    case 21:
      if (c == EOF  ||  (!isalpha(c)  &&  !isdigit(c))) {
        unget_char (c);
        return NO;
      } else {
        SYNTAX_ERROR();
      }
      break;

    /* read so far: i */
    case 22:
      IFCHAR ('f', 23);
      break;
      
    /* read so far: if */
    case 23:
      if (c == EOF  ||  (!isalpha(c)  &&  !isdigit(c))) {
        unget_char (c);
        return IF;
      } else {
        SYNTAX_ERROR();
      }
      break;
      
    /* read so far: sc */
    case 24:
      IFCHAR ('r', 25);
      break;
      
    /* read so far: scr */
    case 25:
      IFCHAR ('i', 26);
      break;
      
    /* read so far: scri */
    case 26:
      IFCHAR ('p', 27);
      break;
      
    /* read so far: scrip */
    case 27:
      IFCHAR ('t', 28);
      break;
      
    /* read so far: script */
    case 28:
      if (c == EOF  ||  (!isalpha(c)  &&  !isdigit(c))) {
        unget_char (c);
        return SCRIPT;
      } else {
        SYNTAX_ERROR();
      }
      break;
      
    /* read so far: a */
    case 29:
      IFCHAR ('n', 30);
      break;
      
    /* read so far: an */
    case 30:
      IFCHAR ('n', 31);
      break;
      
    /* read so far: ann */
    case 31:
      IFCHAR ('a', 32);
      break;
      
    /* read so far: anna */
    case 32:
      if (c == EOF  ||  (!isalpha(c)  &&  !isdigit(c))) {
        unget_char (c);
        person = Anna;
        return PERSON;
      } else {
        SYNTAX_ERROR();
      }
      break;
      
    /* read so far: r */
    case 33:
      IFCHAR ('o', 34);
      break;
      
    /* read so far: ro */
    case 34:
      IFCHAR ('s', 35);
      break;
      
    /* read so far: ros */
    case 35:
      IFCHAR ('a', 36);
      break;
      
    /* read so far: rosa */
    case 36:
      if (c == EOF  ||  (!isalpha(c)  &&  !isdigit(c))) {
        unget_char (c);
        person = Rosalila;
        return PERSON;
      } else if (c == 'l') {
        mode = 37;
      } else {
        SYNTAX_ERROR();
      }
      break;
      
    /* read so far: rosal */
    case 37:
      IFCHAR ('i', 38);
      break;
      
    /* read so far: rosali */
    case 38:
      IFCHAR ('l', 39);
      break;
      
    /* read so far: rosalil */
    case 39:
      IFCHAR ('a', 40);
      break;
      
    /* read so far: rosalila */
    case 40:
      if (c == EOF  ||  (!isalpha(c)  &&  !isdigit(c))) {
        unget_char (c);
        person = Rosalila;
        return PERSON;
      } else {
        SYNTAX_ERROR();
      }
      break;
      
    /* read so far: el */
    case 41:
      IFCHAR ('s', 42);
      break;
      
    /* read so far: els */
    case 42:
      IFCHAR ('a', 43);
      break;
      
    /* read so far: elsa */
    case 43:
      if (c == EOF  ||  (!isalpha(c)  &&  !isdigit(c))) {
        unget_char (c);
        person = Elsa;
        return PERSON;
      } else {
        SYNTAX_ERROR();
      }
      break;
      
    default:
      /* invalid mode */
      abort();
    }
  } while (1);
  /*NOTREACHED*/
}


/*
 * Read away characters until newline.  The function is used for ignoring
 * comments and reading away extra characters after an error.
 */
static void
skip_nl (void)
{
  int c = read_char ();
  while (c != '\r'  &&  c != '\n'  &&  c != EOF) {
    c = read_char ();
  }
}


/*
 * Read character from conversation file.
 */
static int
read_char (void)
{
  int c;

  /* get character */
  c = fgetc (FP); 
  assert (FP != NULL);

  /* convert ms-dos line feeds to unix line feeds */
  if (c == '\r') {
    c = fgetc (FP);
    if (c != EOF  &&  c != '\n') {
      /* not a ms-dos line-feed, put back char following \r */
      ungetc (c, FP);

      /* treat \r as unix line-feed */
      c = '\n';
    }
  }

  /* update line number */
  if (c == '\n') {
    ++lineno;
  }

  return c;
}


/*
 * Put back previously read character.  White space is not put back to
 * buffer as it would be skipped anyway.
 */
static void
unget_char(
    int c)
{
  switch (c) {
  case '\n': case '\r': case '\t': case ' ':
    break;
    
  default:
    assert (FP != NULL);
    ungetc (c, FP);
  }
}


/*
 * Print parse error on line, and try to deal with an error.
 */
static void
parse_error(
    const char *msg)
{
  /* show message */
  fprintf (stderr, "%s:%d: %s\n", SCRIPT_FILE, lineno, msg);

  /* ignore remaining line to bypass the error */
  skip_nl ();

  /* record the fact that an error has happened */
  conversation_ok = 0;
}


/*
 * Load scripts for Peach Gardens.
 */
void
init_script (void)
{
  int ok;
  assert (!initialized);

  /* reset script pointers */
  memset (scripts, 0, sizeof (scripts));

  /* load scripts from file */
  FP = fopen (locate(SCRIPT_FILE), "r");
  if (FP != NULL) {

    /* read file */
    lineno = 1;
    conversation_ok = 1;
    ok = read_file ();
    fclose (FP);

    /* stop now if errors were found */
    if (!ok) {
      abort_msg ("error(s) in script file");
    }
    
  } else {
    abort_msg ("cannot load file %s", SCRIPT_FILE);
  }
  
  initialized = 1;
}


/*
 * Remove scripts from memory.
 */
void
done_script (void)
{
  struct script *sp;
  int i;
  assert (initialized);

  /* release all scripts */
  for (i = 0; i < MAX_SCRIPTS; i++) {
    sp = scripts[i];
    if (sp != NULL) {

      /* erase single script */
      free_script (sp);
      scripts[i] = NULL;
      
    }
  }
  
  initialized = 0;
}
