/*
 * split.cpp - split string into array
 * 
 * Copyright (C) 2005  Toni Ronkko
 * 
 * Peach Gardens is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * Peach Gardens is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with Peach Gardens; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * 
 * Dec 29 2003, Toni Ronkko <toni at users.playstation2-linux.com>
 *
 * $Id: split.c,v 1.2 2005/01/09 21:07:20 tronkko Exp $
 *
 * $Log: split.c,v $
 * Revision 1.2  2005/01/09 21:07:20  tronkko
 * using debug.h
 *
 * Revision 1.1  2004/11/20 10:59:29  tronkko
 * Initial revision
 *
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>


#include "debug.h"
#include "split.h"


/* convert escape sequence to an ASCII character code */
static char esc (char c);


/*
 * Split string `opts' into string table `argv', returning the number of
 * strings in the table to variable argc.  On a succesful return the
 * function initializes pointers argv and argc.  The memory area held by the
 * argv pointer must be freed using free() when the string table is not
 * needed anymore.
 *
 * The function can be used for preparing command line options on Windows,
 * or for reading default command line options from the environment.
 * The function returns false if there was an unterminated argument in the
 * string.
 */
int 
split(
    int *argc, char ***argv,
    const char *opts, const char *separ,
    int skip_multi)
{ 
  int opts_needed;                             /* number of ptrs in table */
  const char *ss;
  size_t bytes;                                /* size of ptrs + strings */
  char **tbl;                                  /* pointer table */
  
  /* skip separators at the beginning of the string */
  if (skip_multi) {
    while (*opts != '\0'  &&  strchr (separ, *opts) != NULL) {
      opts++;
    }
  }
  
  /*
   * Compute the number of separators in the string.  The separators will
   * most likely mark the arguments, and thus tell us the size of the
   * pointer table that must be allocated to hold the strings.  However, due
   * to the quoted strings possibly containing separators, the number of
   * arguments computed may be greater than the actual number of arguments.
   * The number is always greater than the actual numebr, however.
   */
  opts_needed = 1;
  ss = opts;
  while (*ss != '\0') {
    /* find next separator character */
    if (strchr (separ, *ss) != NULL) {
      opts_needed++;
      /* skip repeated separator characters */
      if (skip_multi) {
        do {
          ss++;
        } while (*ss != '\0'  &&  strchr (separ, *ss) != NULL);
      } else {
        ss++;
      }
    } else {
      ss++;
    }
  }
  
  /*
   * Compute number of bytes needed to store the pointer table and the
   * strings.  We will allocate the memory in one block that holds both the
   * pointer table and the option strings to make the memory freeable by
   * standard free().
   */
  bytes = opts_needed * sizeof (char*) + (ss - opts + 2);
  
  /* allocate the memory for the pointer table and string */
  tbl = (char**) malloc (bytes);
  if (tbl != NULL) {
    int i = 0;
    
    /* compute start of the strings in the memory block just allocated */
    char *p = (char*)(tbl + opts_needed);

    /* save the very first argument */
    if (*opts != '\0') {
      tbl[i++] = p;
    }
    
    /* extract arguments to the string area following the pointer table */
    while (*opts != '\0') {
      switch (*opts) {
      /* string in double quotes */
      case '"':
        /* skip the first double quote */
        opts++;
        
        /* copy up to the very next double quote */
        while (*opts != '\0'  &&  *opts != '\"') {
          if (*opts != '\\') {
            /* copy as is */
            *p++ = *opts++;
          } else {
            /* convert escape sequence */
            if (*++opts != '\0') {
              *p++ = esc(*opts++);
            } else {
              /* backslash at the end of the string */
              *p++ = '\\';
            }
          }
        }
        if (*opts != '\0') {
          /* skip ending double quote */
          opts++;
        } else {
          /* unterminated string, release memory allocated so far */
          free (tbl);
          return 0;
        }
        break;
        
      /* quoted string */
      case '\'':
        /* skip the very first quote */
        opts++;
        
        /* copy up to the very next quote */
        while (*opts != '\0'  &&  *opts != '\'') {
          if (*opts != '\\') {
            /* copy as is */
            *p++ = *opts++;
          } else {
            /* convert escape sequence */
            if (*++opts != '\0') {
              *p++ = esc(*opts++);
            } else {
              /* backslash at the end of the string */
              *p++ = '\\';
            }
          }
        }      
        if (*opts != '\0') {
          /* skip ending quote */
          opts++;
        } else {
          /* unterminated string, release memory allocated so far */
          free (tbl);
          return 0;
        }
        break;
        
      /* read escape sequence */
      case '\\':
        if (*++opts != '\0') {
          *p++ = esc(*opts++);
        } else {
          /* backslash at the end of the string */
          *p++ = '\\';
        }
        break;
        
      /* plain character or separator */
      default:
        if (strchr (separ, *opts) != NULL) {
          /* separators terminates the argument */
          *p++ = '\0';
          
          /* skip duplicate separators */
          opts++;
          if (skip_multi) {
            while (*opts != '\0'  &&  strchr (separ, *opts) != NULL) {
              opts++;
            }
          }
          
          /* start new argument if there is some characters left in string*/
          if (!skip_multi  ||  *opts != '\0') {
            tbl[i++] = p;
            assert (i <= opts_needed);
          }
        } else {
          /* plain character */
          *p++ = *opts++;
        }
      }
      assert (p < (char*)tbl + bytes);
    }
    
    /* terminate final argument */
    if (p != NULL) *p = '\0';
    
    /* return the string table and the count */
    *argv = tbl;
    *argc = i;
    return 1;
  } else {
    /* not enough memory for the pointer table */
    out_of_memory ();
    /*NOTREACHED*/
  }
}


/*
 * Get ASCII code associated with the second letter of the two-letter
 * escape sequence.  For example, return '\a' for "\\a" or '\n' for
 * "\\n".
 */
static char
esc(
    char c)
{
  switch (c) {
  case 'a': return '\a';
  case 'b': return '\b';
  case 'r': return '\r';
  case 'n': return '\n';
  case 'f': return '\f';
  case 't': return '\t';
  case 'v': return '\v';
    
  case '\0':
    /* invalid escape sequence */
    assert (c != '\0');
    invalid_escape ();
    /*NOTREACHED*/
    
  default:
    return c;
  }
}
