/*
 * spotlight.c - highlight circular area
 * 
 * Peach Gardens is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * Peach Gardens is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with Peach Gardens; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * 
 * Jan 21 2005, Toni Ronkko <toni at users.playstation2-linux.com>
 *
 * $Id: spotlight.c,v 1.2 2005/06/22 13:11:40 tronkko Exp $
 *
 * $Log: spotlight.c,v $
 * Revision 1.2  2005/06/22 13:11:40  tronkko
 * use sinf() and cosf()
 *
 * Revision 1.1  2005/04/03 16:25:32  tronkko
 * Initial revision
 *
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>

#include "spotlight.h"
#include "graphics.h"


/* Maximum step size in radians.  Decrease the value to make ellipses
 * smooth and increase the value to speed up drawing.  This value affects
 * especially small ellipses. */
#define MAX_STEP ((M_PI*2/360)*20)

/* Smoothness factor of ellipses.  Increase the value to speed up drawing or
 * decrease the value to make ellipses smoother.  This value affects
 * especially large ellipses. */
#define SMOOTHNESS 25

/* number of pre-computed sine and cosine values */
#define MAX_ANGLES 128

/* max ellipse radius for which to compute suitable step size */
#define MAX_RADIUS 500

/* interval on which to compute step sizes */
#define RADIUS_INTERVAL 20


/* is module initialized? */
static int initialized = 0;

/* sine and cosine tables */
static float _sin[MAX_ANGLES];
static float _cos[MAX_ANGLES];

/* for determinating suitable step sizes based on radius */
static int steps[MAX_RADIUS / RADIUS_INTERVAL];


/* color everything except ellipse at (x,y) */
void
draw_spotlight(
    int x, int y, int dx, int dy)
{
  assert (initialized);

  if (dx > 0  &&  dy > 0) {
    int max;
    int step;
    int i;
    int x1, y1, x2, y2;
    
    /* compute suitable step size */
    max = dx;
    if (dy > max) max = dy;
    if (max < MAX_RADIUS) {
      step = steps[max / RADIUS_INTERVAL];
    } else {
      step = 1;
    }
    assert (step > 0);
    assert (step < MAX_ANGLES / 4);

    /* lower right quadrant */
    i = 0;
    x1 = x + dx;
    y1 = y;
    do{
      /* compute next vertex */
      assert (i < MAX_ANGLES);
      x2 = (int) (x + _cos[i] * dx);
      y2 = (int) (y + _sin[i] * dy);
      i += step;

      triangle (screenwidth(), screenheight(), x1, y1, x2, y2);

      /* remember previous vertex */
      x1 = x2;
      y1 = y2;
    }while (i < MAX_ANGLES / 4);

    /* connect quadrants */
    triangle (x1, y1, screenwidth(), screenheight(), 0, screenheight());
    
    /* lower right quadrant */
    do{
      assert (i < MAX_ANGLES);
      x2 = (int) (x + _cos[i] * dx);
      y2 = (int) (y + _sin[i] * dy);
      i += step;
      
      triangle (0, screenheight(), x1, y1, x2, y2);
      
      x1 = x2;
      y1 = y2;
    }while (i < MAX_ANGLES / 2);
    
    triangle (x1, y1, 0, screenheight(), 0, 0);
    
    /* upper left quadrant */
    do{
      assert (i < MAX_ANGLES);
      x2 = (int) (x + _cos[i] * dx);
      y2 = (int) (y + _sin[i] * dy);
      i += step;
      
      triangle (0, 0, x1, y1, x2, y2);
      
      x1 = x2;
      y1 = y2;
    }while (i < MAX_ANGLES / 2 + MAX_ANGLES / 4);
    
    triangle (x1, y1, 0, 0, screenwidth(), 0);
    
    /* upper right quadrant */
    do{
      assert (i < MAX_ANGLES);
      x2 = (int) (x + _cos[i] * dx);
      y2 = (int) (y + _sin[i] * dy);
      i += step;
      
      triangle (screenwidth(), 0, x1, y1, x2, y2);
      
      x1 = x2;
      y1 = y2;
    }while (i < MAX_ANGLES);
    
    /* close the gap between first and last triangle */
    triangle (screenwidth(), 0, x1, y1, x + dx, y);
    
    /* connect first and last quadrant */
    triangle (x + dx, y, screenwidth(), 0, screenwidth(), screenheight());
  }  
}


/*
 * Recompute sine and cosine table so enhance speed of draw_spotlight().
 * This is mostly for PlayStation 2 which does not seem to execute sin() or
 * cos() in hardware, or at least GCC 2.x does optimize that.  Without
 * precomputed cosine tables drawing a single ellipse can consume about 30%
 * of clock cycles.  With precompute tables the effect is almost zero.
 */
void
init_spotlight (void)
{
  float a;
  int i, radius;
  
  assert (!initialized);

  /* precompute sine and cosine angles */
  a = 0;
  i = 0;
  while (a < M_PI*2) {
    assert (i < MAX_ANGLES);
    _sin[i] = sinf (a);
    _cos[i] = cosf (a);
    i++;
    a += M_PI*2 / MAX_ANGLES;
  }

  /* compute suitable step sizes */
  i = 0;
  radius = RADIUS_INTERVAL;
  while (radius < MAX_RADIUS) {
    int step;
    
    a = asin (((float)SMOOTHNESS) / radius);
    if (a > MAX_STEP) {
      a = MAX_STEP;
    }

    /* convert radians to match sine and cosine tables */
    step = (int) (a * MAX_ANGLES / (M_PI * 2));
    if (step < 1) {
      step = 1;
    }

    /* save step size */
    assert (i < MAX_RADIUS / RADIUS_INTERVAL);
    assert (step < MAX_ANGLES);
    steps[i] = step;

    /* next radius */
    radius += RADIUS_INTERVAL;
    i++;
  }
  
  initialized = 1;
}


/* de-initialize modules */
void
done_spotlight (void)
{
  assert (initialized);
  initialized = 0;
}
