/*
 * text.c - text printing
 * 
 * Copyright (C) 2005  Toni Ronkko
 * 
 * Peach Gardens is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * Peach Gardens is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with Peach Gardens; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * 
 * Nov 16 2004, Toni Ronkko <toni at users.playstation2-linux.com>
 *
 * $Id: text.c,v 1.4 2005/06/22 13:12:10 tronkko Exp $
 *
 * $Log: text.c,v $
 * Revision 1.4  2005/06/22 13:12:10  tronkko
 * support for non-bitmap fonts
 *
 * Revision 1.3  2005/04/03 16:25:58  tronkko
 * textn(), wrap_text(), settextwrap()
 *
 * Revision 1.2  2005/01/09 21:07:43  tronkko
 * moved font handling to separate module
 *
 * Revision 1.1  2004/11/20 10:58:54  tronkko
 * Initial revision
 *
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "text.h"
#include "image.h"
#include "graphics.h"
#include "debug.h"


/* number of pixels to leave between lines of text in wrap_text() */
#define LINE_SPACING 2


/* current text position */
static int cur_textx = 0;
static int cur_texty = 0;

/* wrap boundaries */
static int wrap_x1 = 0;
static int wrap_y1 = 0;
static int wrap_x2 = -9999;
static int wrap_y2 = -9999;


/*
 * Print string s to current text position with current font and drawing
 * color.  All letters will be drawn below and left of current text
 * position, despite of the font being used.  The function updates the
 * current text position such that the very next string will be printed
 * right after the previous one, unless settextpos() is called in between.
 * The function refuses to deal with control characters such as tabs and
 * line-feeds.
 */
void
textn(
    const char *s, unsigned int n)
{
  int x2;
  struct font *fnt = getfont();
  assert (fnt != NULL);
  assert (fnt->draw != NULL);
  
  /* compute max coordinate */
  x2 = screenwidth();
  
  /* draw letters one at the time */
  while (n > 0  &&  *s != '\0'  &&  cur_textx < x2) {

    /* paint glyhp */
    cur_textx += fnt->draw (*s);

    n--;
    s++;
  }
}


/* print string s */
void
text(
    const char *s)
{
  textn (s, strlen (s));
}


/*
 * Print multi-line string s to current text position with current font and
 * drawing color while wrapping lines automatically to wrap region set by
 * setwrap().  Nothing is printed if the current text position is outside
 * wrap region.
 */
void
wrap_text(
    const char *s)
{
  int x = textx();
  int y = texty();
  int h = textheight ("M") + LINE_SPACING;

  /* ensure that text pos is within wrap region */
  if (wrap_x1<=x && x<wrap_x2  &&  wrap_y1<=y && y+h<wrap_y2) {

    /* print text line by line */
    while (*s != '\0'  &&  y + h < wrap_y2) {
      const char *p = s;
      const char *start = s;
      int soft_break = 1;
      int w = x;

      /*
       * Find wrap position.  Note that nth position is also investigated.
       * This is useful when a line completely fills available space.
       */      
      while (*s != '\0') {
        const struct glyph *gp;
        char c = *s++;

        /* record possble break points */
        if (c == ' '  ||  c == '\t') {
          /* any space can serve as a wrap point */
          p = s;
        } else if (c == '\n') {
          /* explicit end of line */
          p = s;
          soft_break = 0;
          break;
        }
        
        /* add character width */
        gp = getglyph (c);
        if (gp != NULL) {
          w += gp->w + gp->lead + gp->follow;
          if (w >= wrap_x2) {
            /* char does not fit */
            s--;
            break;
          }
        }
      }

      /* include final word */
      if (*s == '\0') {
        p = s;
      }

      /* print lineful of text */
      if (p != start) {
        /* wrap pos found */
        textn (start, p - start);

        /* move pointer to beginning of a word that did not fit to line */
        s = p;
      } else {
        /* cut line forcefully */
        assert (start < s + 1);
        s--;
        textn (start, s - start);
      }

      /* skip spaces following implicit wrap point */
      if (soft_break) {
        while (*s == ' '  ||  *s == '\t') {
          s++;
        }
      }
      
      /* move cursor to new line */
      if (*s != '\0') {
        x = wrap_x1;
        y = y + h;
        settextpos (x, y);
      }
    }
  }
}


/*
 * Set current text position.  Note that functions clear() or swap_buffers()
 * will not reset text position automatically.  Hence, be sure to set the
 * text position at least once per frame.
 */
void 
settextpos(
    int x, int y)
{
  cur_textx = x;
  cur_texty = y;
}


/*
 * Get x-coordinate of current text position.
 */
int 
textx (void)
{
  return cur_textx;
}


/*
 * Get y-coordinate of current text position.  This can be used to simulate
 * line feeds.  For example, settextpos(0, texty()+textheight()) will set
 * text cursor to the beginning of the very next row.
 */
int 
texty (void)
{
  return cur_texty;
}


/*
 * Set boundaries for automatic line wrapping.
 */
void
setwrap(
    int x1, int y1, int x2, int y2)
{
  wrap_x1 = x1;
  wrap_y1 = y1;
  wrap_x2 = x2;
  wrap_y2 = y2;
}
