/*
 * ticks.c - get/set elapsed time
 * 
 * Copyright (C) 2005  Toni Ronkko
 * 
 * Peach Gardens is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * Peach Gardens is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with Peach Gardens; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * 
 * Nov 21 2004, Toni Ronkko <toni at users.playstation2-linux.com>
 *
 * $Id: ticks.c,v 1.1 2005/01/09 21:07:46 tronkko Exp $
 *
 * $Log: ticks.c,v $
 * Revision 1.1  2005/01/09 21:07:46  tronkko
 * Initial revision
 *
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <time.h>

#include "config.h"
#include "ticks.h"
#include "debug.h"

#ifdef HAVE_SYS_TIME_H
# include <sys/time.h>
#endif

#ifdef HAVE_GETTIMEOFDAY

/* has the timer interface been initialized */
static int initialized = 0;

/* number of seconds since epoch at initialization time */
static struct timeval start;


/*
 * Get number of ticks (milliseconds) since program start.  Beware: the
 * counter wraps around after running the application for about 50 days
 * straight.
 *
 * Note that time spend in suspended execution mode stops the tick counter
 * from increasing.  Thus, if the user suspends the application with ctrl+z
 * while the application is waiting for some amount of ticks to pass, then
 * the application does see the tick count to increase and hence does not
 * appear to continue processing while sleeping.
 */
ticks_t
ticks (void)
{
  int error;
  struct timeval now;
  
  /* get current time */
  error = gettimeofday (&now, NULL);
  assert (error == 0);

  /* record starting time */
  if (!initialized) {
    start = now;
    initialized = 1;
    return 0;
  }
    
  /* convert time to milliseconds */
  if (start.tv_usec <= now.tv_usec) {
    return (now.tv_sec - start.tv_sec) * 1000
        + (now.tv_usec - start.tv_usec) / 1000;
  } else {
    assert (now.tv_usec <= 1000000);
    return (now.tv_sec - start.tv_sec - 1) * 1000
        + (1000000 - start.tv_usec + now.tv_usec) / 1000;
  }
}


/*
 * Specify number of ticks passed since the program start.  This is
 * sometimes useful for "forgetting" time from being ever passed.  For
 * example, time spent in suspended mode does not really help us do
 * anything, and it is easier to base delay loops to tick count if the
 * suspended mode does not accumulate the tick count.
 */
void
set_ticks(
    ticks_t n)
{
  int error;
  struct timeval now;
  
  /* get current time */
  error = gettimeofday (&now, NULL);
  assert (error == 0);

  /* fix starting time such that ticks() returns n the next time */
  if (now.tv_usec >= (int) ((n % 1000) * 1000)) {
    start.tv_sec = now.tv_sec - n / 1000;
    start.tv_usec = now.tv_usec - (n % 1000) * 1000;
  } else {
    start.tv_sec = now.tv_sec - n / 1000 - 1;
    start.tv_usec = 1000000 - (n % 1000) * 1000 + now.tv_usec;
  }
  initialized = 1;
}

#else
# error "missing gettimeofday()"
#endif
