/*
 * x11.c - graphics interface to x11 
 * 
 * Copyright (C) 2005  Toni Ronkko
 * 
 * Peach Gardens is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * Peach Gardens is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with Peach Gardens; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * 
 * Nov  6 2004, Toni Ronkko <toni at users.playstation2-linux.com>
 *
 * $Id: x11.c,v 1.4 2005/06/22 13:12:50 tronkko Exp tronkko $
 *
 * $Log: x11.c,v $
 * Revision 1.4  2005/06/22 13:12:50  tronkko
 * added stub for error dialog, compiles cleanly under cygwin
 *
 * Revision 1.3  2005/04/03 16:28:15  tronkko
 * fade(), specify joystick device on command line, flush_gfx()
 *
 * Revision 1.2  2005/01/09 21:10:16  tronkko
 * several small fixes, use of textures is controlled by USE_TEXTURES
 *
 * Revision 1.1  2004/11/20 10:58:56  tronkko
 * Initial revision
 *
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "config.h"
#if defined(X11) && defined(OPENGL) && !defined(PS2DEV)
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>
#include <X11/Xlib.h>

#ifdef HAVE_SIGNAL_H
# include <signal.h>
#endif
#if HAVE_UNISTD_H
# include <unistd.h>
#endif

#include "graphics.h"
#include "options.h"
#include "image.h"
#include "debug.h"
#include "input.h"
#include "ticks.h"


/* program title */
#define APP_TITLE "Peach Gardens"

/* set to 1 to enable full-screen modes */
#define ENABLE_FULLSCREEN 1

/* set to 1 to use full screen modes by default */
#define PREFER_FULLSCREEN 1

/* Duration of display cycle in millisecond.  Value of 20 is equal to 50
 * frames per sec.  Undefine for dynamic frame rate. */  
#define TICKS_PER_FRAME 20

/* Set to 1 to implement bitblt() with textures.  Textures are fast and
 * work OK with majority of OpenGL cards, but some cards with little
 * on-board memory may miss some textures.  Moreover, textures may flash
 * when used for the very first time.  If enabled, some cards may need a
 * couple of minutes of use before the textures show up. */
#define USE_TEXTURES 0

/* swap integers x and y */
#define SWAP(x,y) do{int _swap_tmp=(x); (x)=(y); (y)=_swap_tmp; }while(0)


/* include support for XFree86 video mode extension */
#if defined(XF86VMODE) && ENABLE_FULLSCREEN
# define XF86_FULLSCREEN
#endif
#ifdef XF86_FULLSCREEN
#include <X11/extensions/xf86vmode.h>
#define V_PHSYNC        0x001 
#define V_NHSYNC        0x002
#define V_PVSYNC        0x004
#define V_NVSYNC        0x008
#define V_INTERLACE     0x010 
#define V_DBLSCAN       0x020
#define V_CSYNC         0x040
#define V_PCSYNC        0x080
#define V_NCSYNC        0x100
#endif


/*
 * Properties of registered image.  The structure maintains information on
 * a texture.  The image is linked to generic image structure with custom
 * data field.
 */
struct reg_image {
  int used;                                    /* for allocated structs */
  int transparent;                             /* has transparent pixels */
  unsigned int texname;                        /* texture name */
  struct pixel *pixels;                        /* image data */
  int size;                                    /* width & height of tex */
};

/* store pixel in a format that is suitable for OpenGL */
struct pixel {
  unsigned char r;
  unsigned char g;
  unsigned char b;
  unsigned char a;
};

/* type of signal handler function */
typedef void sigaction_handler (int, siginfo_t*, void *);


/* load image to texture memory */
static void load_texture (struct reg_image *rip);

/* purge image from OpenGL context */
static void purge_image (struct reg_image *rip);

/* purge all images */
static void purge_images (void);

/* handle sudden termination */
static void sudden_death (int sig, int code, void *ctx);


/* current display */
static Display *dpy = NULL;

/* handle of our window */
static Window win;

/* has the graphics window been created? */
static int window_open = 0;

/* position, width and height of graphics window */
static int winx = 0;
static int winy = 0;
static int winwidth = 640;
static int winheight = 512;

/* screen update frequency */
#ifdef XF86_FULLSCREEN
static int desired_frequency = 50;
#endif

/* has the video mode been changed? */
#ifdef XF86_FULLSCREEN
static int video_changed = 0;
#endif

/* list of available videw modes */
#ifdef XF86_FULLSCREEN
static XF86VidModeModeInfo **modes = NULL;
#endif

/* original display offset (for switching back to original video mode) */
#ifdef XF86_FULLSCREEN
static int oldx, oldy;
#endif

/* prefer full screen video? */
static int prefer_fullscreen = 
#if PREFER_FULLSCREEN
  1
#else
  0
#endif
;

/* use fullscreen mode? */
static int fullscreen = 0;

/* current OpenGL context */
static GLXContext glcontext = NULL;

/* graphics libraries initialized? */
static int initialized = 0;

/* all image tags */
static struct reg_image images[MAX_IMAGES];

/* colors */
static struct color bgcolor = { 0, 0, 0, 0xff};
static struct color fgcolor = { 0xff, 0xff, 0xff, 0xff};


/*
 * Load image data to texture memory.  This is required for using an image
 * as a texture.
 */
static void 
load_texture(
    struct reg_image *rip)
{
  assert (rip != NULL);
  assert (&images[0] <= rip  &&  rip < &images[MAX_IMAGES]);
  assert (rip->used);
  assert (rip->texname == 0);
  
  /* reset unpacking modes */
  glPixelStorei (GL_UNPACK_ALIGNMENT, 1);
  glPixelStorei (GL_UNPACK_SKIP_ROWS, 0);
  glPixelStorei (GL_UNPACK_SKIP_PIXELS, 0);
  glPixelStorei (GL_UNPACK_ROW_LENGTH, rip->size);
  
  /* generate name for the texture */
  glGenTextures (1, &rip->texname);
  glBindTexture (GL_TEXTURE_2D, rip->texname);
  
  /*
   * Pass texture to OpenGL.  After this, texture should be now ready for
   * use.  However, at least on my Voodoo 3 Banshee and NVidia Ti4200 cards
   * the texture only becomes visible at the very next frame.  Thus, the
   * texture is not actually available for drawing right now, but only a
   * single-color rectangle is produced.  This is most likely a feature in
   * OpenGL drivers.  Probably something to preserve reliable frame rate.
   */
  glTexImage2D(
      GL_TEXTURE_2D,
      0,                                       /* mipmap level */
      GL_RGBA,                                 /* internal storage format */
      rip->size,                               /* width (2^x) */
      rip->size,                               /* height (2^x) */
      0,                                       /* border */
      GL_RGBA,                                 /* input format */
      GL_UNSIGNED_BYTE,                        /* data type */
      rip->pixels);                            /* ptr to pixel data */
}


/*
 * Blend rectangular sub-image (sx,sy)-(sx+w,sy+h) with foreground color to
 * frame buffer at (dx,dy).  This is typically used for painting fonts.
 */
void
textblt(
    struct image *imp,
    int sx, int sy, int w, int h,
    int dx, int dy)
{
  float w1;
  struct reg_image *rip;
  
  assert (window_open);
  assert (imp != NULL);
  assert (imp->width > 0);
  assert (w > 0  &&  h > 0);
  assert (0 <= sx  &&  sx+w <= imp->width);
  assert (0 <= sy  &&  sy+h <= imp->height);
  
  /* get pointer to registered image */
  rip = (struct reg_image*) imp->data;
  assert (rip != NULL);
  assert (rip->used);
  assert (&images[0] <= rip  &&  rip < &images[MAX_IMAGES]);

  /*
   * Load texture to OpenGL memory if not already done so.  Note that
   * depending on the image and the graphics cards used, the texture may not
   * be actually loaded until the very next frame.  At least Nvidia OpenGL
   * driver seems to load the texture at next glClear() call.
   */
  if (rip->texname == 0) {
    load_texture (rip);
  }
  assert (rip->texname != 0);

  /* for converting pixel coordinates to (s,t) coordinates */
  w1 = 1.0f / rip->size;

  /* transparency */
  glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glEnable (GL_BLEND);
  
  /* enable blending to foreground color */
  glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_BLEND);
  
  /* draw pixels as a texture */
  glBindTexture (GL_TEXTURE_2D, rip->texname);
  glEnable (GL_TEXTURE_2D);
  glBegin (GL_TRIANGLE_STRIP);
  
  glTexCoord2f (((float) sx) * w1, ((float) sy) * w1);
  glVertex2f (dx, dy);
  
  glTexCoord2f (((float) sx) * w1, ((float) sy+h) * w1);
  glVertex2f (dx, dy+h);
  
  glTexCoord2f (((float) sx+w) * w1, ((float) sy) * w1);
  glVertex2f (dx+w, dy);
  
  glTexCoord2f (((float) sx+w) * w1, ((float) sy+h) * w1);
  glVertex2f (dx+w, dy+h);
  
  glEnd ();
  glDisable (GL_TEXTURE_2D);  
  glDisable (GL_BLEND);
}


/*
 * Draw part (sx1,sy1)-(sx2,sy2) of multi-color image imp to screen at
 * position (dx1,dy1)-(dx2,dy2).
 */
void
bitblt(
    struct image *imp,
    int sx1, int sy1, int sx2, int sy2,
    int dx1, int dy1, int dx2, int dy2)
{
  struct reg_image *rip;
  
  assert (imp != NULL);
  assert (window_open);
  assert (dx1 != dx2);
  assert (dy1 != dy2);
  assert (0 <= sx1  &&  sx1 <= imp->width);
  assert (0 <= sy1  &&  sy1 <= imp->height);
  assert (0 <= sx2  &&  sx2 <= imp->width);
  assert (0 <= sy2  &&  sy2 <= imp->height);
  
  /* get pointer to registered image */
  rip = (struct reg_image*) imp->data;
  assert (rip != NULL);
  assert (&images[0] <= rip  &&  rip < &images[MAX_IMAGES]);
  assert (rip->used);
  assert (imp->width > 0);

  /* enable blending to achieve transparency */
  if (rip->transparent) {
    glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable (GL_BLEND);
  }

  /* paint pixels to screen */
#if USE_TEXTURES
  {
    float w1;
    
    /*
     * Load texture to OpenGL memory if not already done so.  Note that
     * depending on the image and the graphics cards used, the texture may
     * not be actually loaded until the very next frame.  At least Nvidia
     * OpenGL driver seems to load the texture at next glClear() call.
     */
    if (rip->texname == 0) {
      load_texture (rip);
    }
    assert (rip->texname != 0);
    
    /* for converting pixel coordinates to (s,t) coordinates */
    w1 = 1.0f / rip->size;

    /* disable blending to foreground color */
    glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    
    /* draw pixels as a texture */
    glBindTexture (GL_TEXTURE_2D, rip->texname);
    glEnable (GL_TEXTURE_2D);
    glBegin (GL_TRIANGLE_STRIP);
    
    glTexCoord2f (((float) sx1) * w1, ((float) sy1) * w1);
    glVertex2f (dx1, dy1);
    
    glTexCoord2f (((float) sx1) * w1, ((float) sy2) * w1);
    glVertex2f (dx1, dy2);
    
    glTexCoord2f (((float) sx2) * w1, ((float) sy1) * w1);
    glVertex2f (dx2, dy1);
    
    glTexCoord2f (((float) sx2) * w1, ((float) sy2) * w1);
    glVertex2f (dx2, dy2);
    
    glEnd ();
    glDisable (GL_TEXTURE_2D);
  }
#else
  {
    float w, h;
    
    /*
     * Ensure that source rectangle is processed from left to right and from
     * top to bottom.  This is needed as glDrawPixels does not accept
     * negative widths or heights.
     */
    if (sx2 < sx1) {
      SWAP (sx1, sx2);
      SWAP (dx1, dx2);
    }
    if (sy2 < sy1) {
      SWAP (sy1, sy2);
      SWAP (dy1, dy2);
    }
    
    /*
     * Clip image to overcome bugs/features in OpenGL drivers.  (Some
     * drivers leave the whole image unpainted, should clipping occur.)
     * Note that the clipping code below is not perfect if the image is
     * magnified at the same time.  You can sometimes see the image growing
     * and shrinking when clipping takes place.
     *
     * FIXME: the code below does not work if dx1 is greater than dx2 or
     * dy1 greater than dy2.
     */
    w = ((float)(sx2 - sx1)) / (dx2 - dx1);
    h = ((float)(sy2 - sy1)) / (dy2 - dy1);
    if (dx1 < 0) {
      sx1 += -dx1 * w;
      dx1 = 0;
    }
    if (dy1 < 0) {
      sy1 += -dy1 * h;
      dy1 = 0;
    }
    if (dx2 > winwidth) {
      sx2 -= (dx2 - winwidth) * w;
      dx2 = winwidth;
    }
    if (dy2 > winheight) {
      sy2 -= (dy2 - winheight) * h;
      dy2 = winheight;
    }
    
    /*
     * Specify location of output rectangle.  This is a bit complicated as
     * glDrawPixels does not have parameters for changing these settings.
     */
    glRasterPos2i (dx1, dy1);
    glPixelStorei (GL_UNPACK_ALIGNMENT, 1);
    glPixelStorei (GL_UNPACK_SKIP_ROWS, sy1);
    glPixelStorei (GL_UNPACK_SKIP_PIXELS, sx1);
    glPixelStorei (GL_UNPACK_ROW_LENGTH, rip->size);
    
    /*
     * With glDrawPixels we cannot directly specify the size of the output
     * rectangle.  However, there is a separate function for enlarging the
     * pixmap.  To use that, we must specify the location (dx2,dy2) in
     * relation to (dx1,dy1).
     */
    w = ((float)(dx2 - dx1)) / (sx2 - sx1);
    h = ((float)(dy2 - dy1)) / (sy2 - sy1);
    glPixelZoom (w, -h);
    
    /*
     * Paint pixels to screen.  Note that this does not seem to work on
     * Voodoo 3 Banshee if a pixel falls outside of the screen boundaries.
     * However, on NVidia Ti4200 this works sometimes despite of the
     * clipping, but not always!
     */
    glDrawPixels(
        sx2 - sx1,                             /* width */
        sy2 - sy1,                             /* height */
        GL_RGBA,                               /* input format */
        GL_UNSIGNED_BYTE,                      /* data type */
        rip->pixels);
  }
#endif
  
  /* disable blending of successive primitives */
  if (rip->transparent) {      
    glDisable (GL_BLEND);
  }
}


/*
 * Clear graphics window while resetting any drawing modes.  As window is
 * only cleared once per frame, this is a good place to set up default
 * drawing modes.
 */
void
clear (void)
{
  assert (window_open);

  /* clear graphics window */
  glClear (GL_COLOR_BUFFER_BIT);

  /* initialize texture modes */
  glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
}


/*
 * Set background color used by clear().
 */
void
setbg(
    int r, int g, int b)
{ 
  float f = 1.0 / 255.0;
  assert (window_open);

  /* save the color for future */
  bgcolor.r = r;
  bgcolor.g = g;
  bgcolor.b = b;
  
  /* start using the color */
  glClearColor (r*f, g*f, b*f, 1.0);
}


/*
 * Set foregroung color.
 */
void
setfg(
    int r, int g, int b)
{
  float texcolor[4];
  float f = 1.0 / 255.0;
  assert (window_open);
  
  /* save the color for future */
  fgcolor.r = r;
  fgcolor.g = g;
  fgcolor.b = b;

  /* convert the color component between 0 and 0xff to OpenGL color */
  texcolor[0] = r*f;
  texcolor[1] = g*f;
  texcolor[2] = b*f;
  texcolor[3] = 1.0;
  
  /* start using the color */
  glColor3f (texcolor[0], texcolor[1], texcolor[2]);

  /* set font color */
  glTexEnvfv (GL_TEXTURE_ENV, GL_TEXTURE_ENV_COLOR, &texcolor[0]);
}


/*
 * Draw line (x1,y1)-(x2,y2) using current pen.
 */
void
line(
    int x1, int y1, int x2, int y2)
{
  assert (window_open);
  glBegin (GL_LINES);
  glVertex2i (x1, y1);
  glVertex2i (x2, y2);
  glEnd ();
}


/*
 * Draw triangle (x1,y1)-(x2,y2),-(x3,y3) using current pen.
 */
void
triangle(
    int x1, int y1, int x2, int y2, int x3, int y3)
{
  assert (window_open);
  glBegin (GL_TRIANGLES);
  glVertex2i (x1, y1);
  glVertex2i (x2, y2);
  glVertex2i (x3, y3);
  glEnd ();
}


/*
 * Draw box (x1,y1) - (x2,y2) with a semi-transparent color.  Argument
 * alpha ranges from 0 to 255, and it defines how much the original color
 * shows through.  If alpha is 200, intensity of foreground color is 80%
 * while 20% of background color shows through.
 */
void
fade(
    int x1, int y1, int x2, int y2, int alpha)
{
  float f = 1.0 / 255.0;
  assert (window_open);
  assert (0 <= alpha  &&  alpha < 256);
    
  /* specify alpha */
  glColor4f (fgcolor.r * f, fgcolor.g * f, fgcolor.b * f, alpha * f);
  
  /* enable blending with background color */
  glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glEnable (GL_BLEND);

  /* paint a rectable */
  glBegin (GL_POLYGON);
  glVertex2i (x1, y1);
  glVertex2i (x2, y1);
  glVertex2i (x2, y2);
  glVertex2i (x1, y2);
  glEnd ();

  /* restore OpenGL state */
  glDisable (GL_BLEND);
  glColor3f (fgcolor.r * f, fgcolor.g * f, fgcolor.b * f);
}


/*
 * Swap double buffers to show hidden buffer to user.
 */
void
swap_buffers (void)
{
#ifdef TICKS_PER_FRAME
  static ticks_t last = 0;
  ticks_t now;
#endif
  
  assert (window_open);
  assert (dpy != NULL);  
  if (window_open) {

    /* flush drawing queues */
    glFinish ();
    
    /*
     * Wait for vertical retrace interrupt and swap display buffers to show
     * the recent drawings to the user.  Note that many popular OpenGL
     * drivers do not really wait for vertical retrace here, but execute the
     * buffer swap function as soon as possible.  This sometimes makes
     * frame rates unreasonably high, which can be seen as wildly running
     * animation and excessively short delays.  Moreover, with NVidia OpenGL
     * driver this sometimes confuses the display address, and the screen
     * flashes.  A small amount of time must be spent after glXSwapBuffers
     * before any more OpenGL requests can be issued.
     *
     * If defined, feature TICKS_PER_FRAME sets the minimum duration of a
     * display cycle.  This can be used for both slowing down frame rate and
     * making frame rate more reliable on varying CPU load.  The feature can
     * also be used for decreasing CPU load caused by the application on a
     * multi-user environments.
     */
#ifdef TICKS_PER_FRAME
    do{      
      now = ticks ();
      if (now - last < TICKS_PER_FRAME) {
        /*
         * Wait some micro seconds to lenghten the display cycle.  Note that
         * usleep can return shortly upon receiving a signal.  Some OpenGL
         * drivers use signals for registering vertical retrace interrupts,
         * so we must keep looping until enough time has passed.
         */
        usleep ((TICKS_PER_FRAME - (now - last)) * 1000);
        now = ticks ();
      }
    } while (now - last < TICKS_PER_FRAME);
    last = now;
#endif
    glXSwapBuffers(dpy, win);

    /*
     * Wait for buffer switch.  This is extremely important for NVidia
     * OpenGL driver.  Calling usleep() also seems to reduce tearing.
     */
    glFinish ();
    usleep (1);
  }
}


/*
 * Flush OpenGL drawing queues and return immediately.
 */
void
flush_gfx (void)
{
  assert (window_open);
  glFlush ();
}


/*
 * Get width of graphics screen.  For Peach Gardens, the width is usually
 * 640, but you should not depend on this.
 */
int
screenwidth (void)
{
  assert (window_open);
  return winwidth;
}


/*
 * Get height of graphics window.
 */
int
screenheight (void)
{
  assert (window_open);
  return winheight;
}


/*
 * Open graphics window and change video mode if on full-screen mode.
 */
int
open_screen (void)
{
  int ok = 0;
  
  assert (initialized);
  assert (!window_open);

  /* open display if needed */
  if (dpy == NULL) {
    dpy = XOpenDisplay(NULL); 
  }
  if (dpy != NULL) {
    int screen;
    unsigned long attrMask;
    XSetWindowAttributes attr;  
    int visAttributes[] = {
      GLX_RGBA,
      GLX_RED_SIZE, 1,
      GLX_GREEN_SIZE, 1,
      GLX_BLUE_SIZE, 1,
      GLX_DOUBLEBUFFER,
      None
    };
  
    Window root;
    XVisualInfo *visinfo;

    /* get handle of root window */
    root = RootWindow (dpy, 0);
    screen = DefaultScreen (dpy);
  
    /* choose GLX visual and pixel format */
    visinfo = glXChooseVisual(dpy, 0, visAttributes);
    if (visinfo) {
      
      /* switch to full screen mode */
#ifdef XF86_FULLSCREEN
      int nmodes;
      int mode;
      int bestmode = -1;
      int diff = 999999999;
      if (prefer_fullscreen) {
        
        /* find out current gfx mode */
        XF86VidModeGetViewPort (dpy, screen, &oldx, &oldy);
      
        /* get list of available video modes */
        XF86VidModeGetAllModeLines (dpy, screen, &nmodes, &modes);
      
        /* select closest mode */
        for (mode = 0; mode < nmodes; mode++) {
          int framerate;
          XF86VidModeModeInfo *p = modes[mode];
          
          /* compute frame rate for current mode */
          framerate = p->dotclock*1000/(p->htotal*p->vtotal);
          if ((p->flags & (V_INTERLACE|V_DBLSCAN)) != 0) framerate /= 2;
          
          /*
           * Compute distance to the desired mode.  Note that the dimension
           * of the selected mode cannot be smaller than that requested.
           */
          int a, b, c, d;
          a = ((int) p->hdisplay) - winwidth;
          b = ((int) p->vdisplay) - winheight;
          c = framerate - desired_frequency;
          d = a*a + b*b + c*c;

          /* remember the best mode so far */
          if (d < diff  &&  a >= 0  &&  b >= 0) {
            diff = d;
            bestmode = mode;
          }
        }
        
        /* change resolution */
        XF86VidModeSwitchToMode (dpy, screen, modes[bestmode]);
        XF86VidModeSetViewPort (dpy, screen, 0, 0);
        video_changed = 1;

        /* fullscreen video */
        fullscreen = 1;
      } else {
        /* windowed mode */
        fullscreen = 0;
      }
#else /*!XF86_FULLSCREEN*/

      /* full screen mode not available */
      fullscreen = 0;
      
#endif /*!XF86_FULLSCREEN*/

      /* create suitable window for full-screen/windowed mode*/
      attr.background_pixel = 0;    
      attr.border_pixel = 0;
      attr.colormap = XCreateColormap(dpy, root, visinfo->visual, AllocNone);
      if (fullscreen) {
        /* create the window that fills the complete screen */
#ifdef XF86_FULLSCREEN
        attr.override_redirect = True;
        attrMask = CWBackPixel|CWBorderPixel | CWColormap | CWOverrideRedirect;
        win = XCreateWindow(
            dpy, root, 0, 0,
            modes[bestmode]->hdisplay, modes[bestmode]->vdisplay,
            0, visinfo->depth, InputOutput,
            visinfo->visual, attrMask, &attr);
#else
        abort ();
#endif
      } else {
        /* create standard window */
        attrMask = CWBackPixel | CWBorderPixel | CWColormap;
        win = XCreateWindow(
            dpy, root, winx, winy, winwidth, winheight,
            0, visinfo->depth, InputOutput,
            visinfo->visual, attrMask, &attr);
      }
      
      if (win) {
        int x1, y1;
        XSizeHints winsize;
        XClassHint winclass;
        char title[] = APP_TITLE;
        typedef char *char_ptr;
        char_ptr list[] = { title, NULL };
        XTextProperty textprop;
        char resname[] = "pgardens";
        char classname[] = "PGardens";
        
        /* initialize window properties */
        winsize.flags = PMinSize | PMaxSize;
        winsize.min_width = winwidth;          /* min & max width */
        winsize.max_width = winwidth;
        winsize.min_height = winheight;
        winsize.max_height = winheight;        
        winclass.res_name = resname;           /* resource name */
        winclass.res_class = classname;        /* resource class */

        /* make window title */
        XStringListToTextProperty (list, 1, &textprop);
        
        /* specify window manager hints */
        XSetWMProperties (dpy, win,
            &textprop, &textprop, NULL, 0, &winsize, NULL, &winclass);

        /* release window name */
        XFree (textprop.value);
        
        /* window created */
        window_open = 1;
        
        /* move window to upper left corner of the virtual screen */
        if (fullscreen) {
          XMoveWindow (dpy, win, 0, 0);
        }
        
        /* display the window */
        XMapWindow(dpy, win);
        
        /* create GLX rendering context */
        glcontext = glXCreateContext(dpy, visinfo, NULL, True);
        if (glcontext != NULL) {                 
          /* bind the rendering context to the window window */
          glXMakeCurrent(dpy, win, glcontext);
          
          /* make a blank cursor */
          if (fullscreen) {
            Cursor cursor;
            XColor dummy;
            Pixmap blank;
            
            /* create empty cursor */
            char data[1] = { 0 };
            
            blank = XCreateBitmapFromData (dpy, win, data, 1, 1);
            cursor = XCreatePixmapCursor(dpy,blank,blank,&dummy,&dummy,0,0);
            XFreePixmap (dpy, blank);
            
            /* confine mouse pointer to the window */
            XWarpPointer (dpy, None, win, 0, 0, 0, 0, 0, 0);
            XGrabPointer (dpy, win, True, ButtonPressMask,
                GrabModeAsync, GrabModeAsync, win, cursor, CurrentTime);
          }

          /* define upper left corner of clipping region */
          if (fullscreen) {
            /* full screen */
#ifdef XF86_FULLSCREEN
            x1 = modes[bestmode]->hdisplay/2 - winwidth/2;
            y1 = modes[bestmode]->vdisplay/2 - winheight/2;
#else
            abort ();
#endif
          } else {
            /* regular window */
            x1 = 0;
            y1 = 0;
          }

          /* specify clipping region */
          glViewport (x1, y1, winwidth, winheight);
          
          /* set up coordinate system (with relation to clip region) */
          glMatrixMode (GL_PROJECTION);
          glLoadIdentity ();
          glOrtho (0, winwidth, winheight, 0, 0, 1);
          glMatrixMode (GL_MODELVIEW);
          glLoadIdentity ();
          
          /* set up drawing modes */
          glShadeModel (GL_FLAT);
          glDisable (GL_DEPTH_TEST);
          glDisable (GL_LIGHTING);

          /* reset drawing colors */
          setbg (bgcolor.r, bgcolor.g, bgcolor.b);
          setfg (fgcolor.r, fgcolor.g, fgcolor.b);

          ok = 1;

        } else {
          fprintf (stderr, "cannot create OpenGL context\n");
        }
      } else {
        fprintf (stderr, "cannot create window\n");
      }
    } else {
      fprintf (stderr, "no matching visual\n");
    }    
  } else {
    fprintf (stderr, "cannot open display\n");
  }  
  return ok;
}


/*
 * Close graphics window and switch back to original video mode.  Note that
 * we also need to close our connection to screen at the same time.  XFree86
 * will not change the video mode unless we also close the display with
 * XCloseDisplay().
 */
void
close_screen (void)
{
  if (dpy != NULL) {

    /* un-initialize fullscreen video */
    if (fullscreen) {
      /* let the mouse pointer move freely again */
      XUngrabPointer (dpy, CurrentTime);
      fullscreen = 0;
    }

    /* remove images from OpenGL texture memory */
    purge_images ();
    
    /* restore original resolution (takes place after XCloseDisplay) */
#ifdef XF86_FULLSCREEN
    if (video_changed) {
      int screen = DefaultScreen (dpy);
      XF86VidModeSwitchToMode (dpy, screen, modes[0]);  
      XF86VidModeSetViewPort (dpy, screen, oldx, oldy);
      video_changed = 0;
    }
#endif /*XF86_FULLSCREEN*/
    
    /* release list of resolutions */
#ifdef XF86_FULLSCREEN
    if (modes != NULL) {
      XFree (modes);
    }
#endif /*XF86_FULLSCREEN*/

    /*
     * Release OpenGL resources allocated for the window.  Note that calling
     * glXMakeCurrent detaches glx context from display.  I am not certain
     * if this is necessary, but I have seen X11 library crash at program
     * exit without an apparent reason.  It might be something with Nvidia
     * drivers, but releasing context should not make thing any worse.
     */
    if (glcontext != NULL) {
      glXMakeCurrent (dpy, None, NULL);
      glXDestroyContext (dpy, glcontext);
      glcontext = NULL;
    }
    
    /* remove window from screen */
    if (window_open) {
      XUnmapWindow (dpy, win);
      XDestroyWindow (dpy, win);
      window_open = 0;
    }

    /* close x11 connection */
    XCloseDisplay (dpy);
    dpy = NULL;
  }
}


/*
 * Initialize X11 graphics libraries.  On X11 the sudden death of a process
 * is not much of a problem, as X11 always seems to leave the graphics
 * console in an usable state even though we forgot to reset the graphics
 * mode.  Hence, we do not need to handle SIGTSTP and SIGCONT signal those
 * are raised as a result of user pressing ctrl+z.
 */
void
init_graphics (void)
{
  struct sigaction handler;
  
  assert (!initialized);

  /* reset registered images */
  memset (&images[0], 0, sizeof (images));

  /* set signal handlers to de-initialize graphics mode at sudden exit */
  memset (&handler, 0, sizeof (handler));
  handler.sa_sigaction = (sigaction_handler*) sudden_death;
  sigaction (SIGINT,  &handler, NULL);
  sigaction (SIGABRT, &handler, NULL);
  sigaction (SIGTERM, &handler, NULL);
  sigaction (SIGSEGV, &handler, NULL);
  
  /* ignore suspend and resume */
  handler.sa_sigaction = (sigaction_handler*) SIG_IGN;
  sigaction (SIGTSTP, &handler, NULL);
  sigaction (SIGCONT, &handler, NULL);
  
  initialized = 1;
}


/*
 * De-initialize graphics libraries.  That is, close the graphics window
 * return the graphics mode to the original one.  Note that the function
 * may be called multiple times, if the an error occurs during ctrl+c
 * handler.  Thus, be sure not to de-initialize things twice. 
 */
void
done_graphics (void)
{
  /* return to text mode/close window */
  close_screen ();

  /* close display connection */
  if (dpy != NULL) {
    XCloseDisplay (dpy);
    dpy = NULL;
  }

  initialized = 0;
}


/*
 * Print an error message to graphics window.
 */
void
error(
    const char *msg)
{
  /* FIXME: create a dialog window */
  fprintf (stderr, "%s\n", msg);
  abort();
}


/*
 * Purge loaded texture from OpenGL context.  The image will be again
 * uplodaded to OpenGL driver the next time the image is used.
 */
static void
purge_image(
    struct reg_image *rip)
{
  assert (rip != NULL);
  assert (&images[0] <= rip  &&  rip < &images[MAX_IMAGES]);
  assert (rip->used);

  /* release texture */
  if (rip->texname != 0) {
    glDeleteTextures (1, &rip->texname);
    rip->texname = 0;
  }
}


/*
 * Purge all images from OpenGL context.  This is used prior to closing the
 * graphics windows.
 */
static void
purge_images (void)
{
  int i;
  struct reg_image *rip;

  for (i = 0; i < MAX_IMAGES; i++) {
    rip = &images[i];
    if (rip->used) {
      purge_image (rip);
    }
  }
}


/*
 * Register image for drawing.  Note that we must not use any OpenGL
 * resources at this time as images can be registered before opening
 * graphics devices.
 *
 * Note that colors will be copied from palette to a new pixmap by the
 * function, and any changes made to the palette has no effect on the
 * registered image.  If you need to change palette, please be sure to
 * change it before registering an image.
 */
int
register_image(
    struct image *imp, const char *buffer, int len)
{
  int ok = 0;
  struct reg_image *rip = NULL;
  int i;
  int w;

  assert (initialized);
  assert (imp != NULL);
  assert (buffer != NULL);
  assert (imp->data == NULL);
  assert (imp->width * imp->height <= len);

  /* compute size of corresponding texture */
  w = imp->width;
  if (imp->height > w) w = imp->height;
  if (w <= 32)        w = 32;
  else if (w <= 64)   w = 64;
  else if (w <= 128)  w = 128;
  else if (w <= 256)  w = 256;
  else if (w <= 512)  w = 512;
  else if (w <= 1024) w = 1024;
  else {
    abort_msg ("texture too large");
  }

  /* find a free reg_image structure */
  for (i = 0; i < MAX_IMAGES; i++) {
    if (images[i].used == 0) {
      /* found an unused image */
      rip = &images[i];
      rip->used = 1;
      break;
    }
  }
  DEBUG (("%p: alloc reg_image", rip));
  if (rip != NULL) {
    /* initialize structure */
    rip->size = w;
    rip->texname = 0;

    /*
     * See whether alpha blending is needed for this image.  That is,
     * whether there is at least one color with alpha component other than
     * 0xff.  Alpha blending is implemented in software with some 3D cards,
     * and avoiding the blending improves performance on such hardware.
     */
    rip->transparent = 0;
    for (i = 0; i < imp->ncolors; i++) {
      if (imp->palette[i].a != 0xff) {
        /* requires blending */
        rip->transparent = 1;
        break;
      }
    }

    /* associate hw dependent structure with the generic image struct */
    imp->data = rip;
    
    /* allocate memory for pixels */
    rip->pixels = malloc (w * w * sizeof (struct pixel));
    DEBUG (("%p: alloc pixels", rip->pixels));
    if (rip->pixels != NULL) {
      int x, y;
      unsigned char *p = (unsigned char*) buffer;
      struct pixel *pix;

      /*
       * Copy pixels from buffer to a dedicated pixel array that usable
       * as a texture.  Note that the memory layoyt of the image may change
       * as a result of copying as we pad the image to 2^m size.  Hence,
       * memcpy() or plain copy cannot be used.
       *
       * Also note that we convert pixel values from color table indices to
       * rgba at the same time.  This simplifies further operations on the
       * image as we do not need to present the palette to OpenGL 
       * with glColorTable() function.  However, the rgba format requires 4
       * bytes per pixel whereas the color index would only require one byte
       * per pixel.  Under unix we generally have plenty of memory available
       * so we do not desperately need to save memory, and I dp not
       * consider the extra memory consumption a problem.  Moreover, OpenGL
       * support for color tables is an extension that may or may not be
       * available.
       */
      for (y = 0; y < imp->height  &&  y < w; y++) {
        for (x = 0; x < imp->width  &&  x < w; x++) {
          int idx;

          /* get pixel from buffer */
          assert ((void*) p < (void*) (buffer + len));
          idx = *p++;

          /* compute address of the pixel in new array */
          pix = rip->pixels + y * w + x;
              
          /* fetch corresponding color from palette and store pixel */
          assert (0 <= idx  &&  idx < imp->ncolors);
          pix->r = imp->palette[idx].r;
          pix->g = imp->palette[idx].g;
          pix->b = imp->palette[idx].b;
          pix->a = imp->palette[idx].a;
        }
      }

      /*
       * At least my Voodoo 3 Banshee and NVidia Ti4200 cards do not really
       * upload the texture into the graphics card until the image has been
       * used at least once.  To circumvent this bug, draw the image outside
       * the screen space to force uploading the texture.  Note, however,
       * that even this does not help if several new textures are brought in
       * at the very same frame.
       */
      if (window_open) {
        bitblt (imp, 0,0, 1,1, -5001,-5001, -5000,-5000);
      }

      /* all done */
      ok = 1;

    } else {
      /* cannot allocate memory for pixel data */
      out_of_memory ();
    }  
  
  } else {
    /* no slots left image array */
    abort_msg ("too many images");
  }
  
  if (ok == 0) {
    unregister_image (imp);
  }
  return ok;
}


/*
 * Unregister image to release all hardware specific resources allocated for
 * it.
 *
 * Note that this very function can be called without init_graphics()
 * function being invoked first.  This may happen in batch programs that
 * only modify images, without any intent on showing them.
 */
void
unregister_image(
    struct image *imp)
{
  struct reg_image *rip;
  assert (imp != NULL);
  
  /* get pointer to reg_image */
  rip = (struct reg_image*) imp->data;  
  if (rip != NULL) {
    /* has some hardware specific data */
    assert (initialized);
    assert (&images[0] <= rip  &&  rip < &images[MAX_IMAGES]);
    assert (rip->used);
    
    /* release OpenGL resources */
    purge_image (rip);
    
    /* release pixmap */
    if (rip->pixels != NULL) {
      /* erase image to notice unappropriate use */
#ifndef NDEBUG
      memset (rip->pixels, 0xcc, rip->size*rip->size*sizeof(struct pixel));
#endif
      
      /* free pixmap */
      DEBUG (("%p: free pixels", rip->pixels));
      free (rip->pixels);
    }

    /* destroy structure to detect dangling pointers */
#ifndef NDEBUG
    memset (rip, 0xcc, sizeof (*rip));
#endif

    /* free image structure */
    DEBUG (("%p: free reg_image", rip));
    rip->used = 0;
  }
}


/*
 * Signal handler for exiting the application in a hurry.  The handler will
 * be called if user aborts the program with ctrl+c, a segmentation fault
 * occurs or the application aborts due to assert() or abort().
 */
static void
sudden_death(
    int sig, int code, void *ctx)
{
  (void) sig; (void) code; (void) ctx;
  
  DEBUG (("aborting due to signal %d, code %d", sig, code));
  if (initialized) {
    done_graphics ();
  }
  _exit (3);
}


/*
 * Initialize keyboard interface.  With X11 interface the keyboard handling
 * is tightly linked with the graphics surface and no special initialization
 * is needed.  However, this very function must be defined as other
 * hardware may need special initialization.
 */
void
init_key (void)
{
  /*NOP*/
}


/*
 * De-initialize keyboard interface.
 */
void
done_key (void)
{
  /*NOP*/
}


/*
 * Get raw hardware specific keyboard event.
 */
int
key_event (void)
{
  /* FIXME: do something */
  return NOEVENT;
}


/*
 * Print help on X11 specific options.
 */
void
print_graphics_options (void)
{
  printf(
      "Graphics options:\n"
      "  -w               Windowed mode.\n"
      "  -f               Full-screen mode.\n"
      "  -j dev           Specify joystick device. [/dev/js0]\n"
/*    "  --geometry WxH   Specify screen size.\n" */
  );
}


/*
 * Parse and read away argv[i] if the switch is a graphics related option.
 * The function should return the number of options it reads away.
 * 
 * The function will be called before reading the standard options so that
 * you can re-define standard options such as -geometry here.  However, 
 * the contents of the arguments array may be destroyed soon; do not store
 * references to the arguments.
 *
 * The graphics library must have been initialized prior to using this very
 * function.
 */
int
read_graphics_options(
    int *argc, char **argv, int i)
{
  int opts = 0;
  const char *p;

  assert (initialized);
  assert (argv != NULL);
  assert (argc != NULL);
  assert (i < *argc);

  p = argv[i];

  /* parse arguments */
  if (strcmp (p,"-w") == 0  ||  strcmp (p,"--window") == 0) {

    /* prefer windowed modes */
    prefer_fullscreen = 0;
    opts = 1;
    
  } else if (strcmp (p,"-f") == 0  ||  strcmp (p,"--full") == 0) {

    /* prefer fullscreen mode */
    prefer_fullscreen = 1;
    opts = 1;
    
  } else if (strncmp (p, "-j", strlen("-j")) == 0) {
    
    /* use joystick device */
    const char *dev;
    if (strlen (p) > strlen("-j")) {
      /* joystick device specified with -j */
      opts = 1;
      dev = p + strlen("-j");
    } else {
      /* joystick device given as separate argument */
      if (i + 1 < *argc) {
        opts = 2;
        dev = argv[i+1];
      } else {
        fprintf (stderr, "missing joystick device\n");
        exit (EXIT_FAILURE);
      }
    }
    joy_device (dev);
    
  } else if (strcmp(p,"--geometry") == 0
      ||  strcmp(p,"-geometry") == 0
      ||  strcmp(p,"-g") == 0) {

    /* specify screen geometry */
    if (i + 1 < *argc) {
      int w, h, x, y;
      char c;

      /* get option */
      const char *curarg = argv[i+1];

      /* parse geometry string */
      if (sscanf (curarg,"%ix%i%c",&w,&h,&c) == 2) {
        /* got width and height */
        winwidth = w;
        winheight = h;
        opts = 2;
      } else if (sscanf (curarg,"%ix%i%d%d%c", &w, &h, &x, &y, &c) == 4) {
        /* got x, y, width and height */
        winx = x;
        winy = y;
        winwidth = w;
        winheight = h;
        opts = 2;
      } else {
        fprintf (stderr, "invalid geometry specification: %s\n", curarg);
        exit (EXIT_FAILURE);
      }
    } else {
      fprintf (stderr, "missing geometry specification\n");
      exit (EXIT_FAILURE);
    }
  }

  /* read away options that we processed */
  if (opts != 0) {
    int j;
    
    /* remove arguments from the argv array */
    for (j = i + opts; j < *argc; j++) {
      argv[j-opts] = argv[j];
    }

    /* fix argc */
    *argc -= opts;
    assert (*argc > 0);
  }

  /* return number of options read */
  return opts;
}
#endif /*X11*/
